Require Import ssreflect ssrbool ssrnat eqtype fintype tuple finfun ssrfun seq.
Require Import pred prelude idynamic ordtype finmap pcm coding unionmap heap.
Require Import feap state zigzag world inject rely atomic autosplit.
Require Import privstate model lemmas stsep histories getter array spcm.
Require Import Eqdep alloc.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(* The implementation of a generic flat combiner *)
(* We can instantiate it with the PCM of histories, get a stack implementation *)
(* Or with natural numbers PCM, to get an incrementor *)

Local Notation nat_set := (union_map nat_ordType unitSet).
Local Notation mutex := (lift mutexUnlifted).

Local Notation "[ g1 <<= g2 ]" := (exists g, g1 \+ g = g2).

(* flat combiner requires methods that work over private heaps only *)
(* so they should be specified by specs over heaps only *)
(* But then I need the interpretations which live in more general states *)
(* so I need some lifting 
Definition lift_pre pv (P : heap -> Prop) := fun s => P (self (pvg pv) s).
Definition lift_post pv B (Q : heap -> B -> heap -> Prop) := 
  fun s1 v s2 => Q (self (pvg pv) s1) v (self (pvg pv) s2). 
*)

Structure flat_combineT := FlatCombineT {
  fc_orig : ptr; (* origin of the side-chanel array *)
  fc_lk : ptr; (* the lock for grabing the resource *)
  fc_n : nat; (* size of the array; i.e., number of admitted threads *)
  fc_label : nat; (* label of the concurroid *)
  fc_pcm : encoded_pcm; (* resource pcm *)
  fc_inv : fc_pcm -> Pred heap; (* resource invariant *)
  fc_precise : forall g, precise (fc_inv g);
  fc_code : Type;  (* code for each method *)
  fc_inT : fc_code -> Type; (* input type of a coded method *)
  fc_outT : fc_code -> Type; (* output type of a coded method *)
  (* method c specification over result value and input/output ghosts *)
  fc_R : forall c, fc_inT c -> fc_outT c -> fc_pcm -> fc_pcm -> Prop; 
  (* program of the coded method *)
  fc_interp : forall pv al asent alk c (v : fc_inT c), 
    {g}, STsep [entangle (Priv pv) (AllocCAS al asent alk)] 
      (fun i => fc_inv g (self (pvg pv) i), 
       fun r m => exists g', 
         [/\ valid (g \+ g'), fc_inv (g \+ g') (self (pvg pv) m) &
             fc_R v r g g']);
  (* the ghost to be returned by help is uniquely determined *)
  (* i.e. fc_R is a partial function *)
  fc_uniq : forall m (v : fc_inT m) (w : fc_outT m) g g1' g2',
    valid (g \+ g1') -> valid (g \+ g2') -> 
    fc_R v w g g1' -> fc_R v w g g2' -> g1' = g2'}. 

(* a structure for threads to make requests *)
Structure req (T : flat_combineT) := 
  NoReq                             (* thread is not making a request *)
| Req (c : fc_code T) of fc_inT c   (* thread wants to run method c, with input iT c *)
| Ans (c : fc_code T) of fc_outT c. (* thread got back result of c *)
                                    (* probably Ans should be parametrized by v as well *)

Definition fc_V (T : flat_combineT) := 
  [encoded_pcm of nat_set * mutex * fc_pcm T].

Definition fc_J (T : flat_combineT) := 
  [encoded_type of prod heap {ffun 'I_(fc_n T) -> fc_pcm T}].

Definition get_tid T (x : fc_V T) : nat_set := x.1.1.
Definition get_mtx T (x : fc_V T) : mutex := x.1.2.
Definition getU T (x : fc_V T) : fc_pcm T := x.2.

Local Notation geth gc s := ((joint gc s).1).
Local Notation getg gc s := ((joint gc s).2).

Definition sum_up (U : pcm) (n : nat) (g : {ffun 'I_n -> U}) := 
  foldr (fun r t => r \+ t) Unit (fgraph g).

(* Lemmas about arrays; to be moved to a separate file *)

Definition is_array T A p f := @shape [finType of 'I_(fc_n T)] A (Array p) f.

Lemma array_footeqH (I : finType) (A1 A2 : Type) (a : ptr) h1 h2 
                    (r1 : {ffun I -> A1}) (r2 : {ffun I -> A2}) : 
        shape (Array a) r1 h1 ->
        shape (Array a) r2 h2 -> footeqH h1 h2.
Proof.
case=>-> V1 [-> V2] /=; split; first by rewrite V1 V2.
move=>z; case: updiP=>[[H][m][->]|X]. 
- by rewrite updimV H /= !fgraph_codom /= !codomE !size_map.
case: updiP=>//; case=>H [m][E S]; elim: X; split=>//; exists m.
by rewrite !fgraph_codom /= !codomE !size_map /= in S *. 
Qed.


Module FlatCombineCoh.
Section FlatCombineCoh.
Variable T : flat_combineT.
Notation a := (fc_orig T).
Notation lk := (fc_lk T).
Notation n := (@fc_n T).
Notation fc := (@fc_label T). 
Notation U := (@fc_pcm T).
Notation I := (@fc_inv T).
Notation V := (fc_V T).
Notation J := (fc_J T).

(* The self component is tickets for thread ids, plus mutex to
symbolize who owns the lock, plus a user ghost. The joint component is
a pair. The first part of the joint pair is a heap storing the
side-channel array.  The second part of the joint pair is the finite
function of uncollected constributions. *)

Definition fcg := @Getter fc V J (Unit, [ffun => Unit]). 

Lemma fc_flatsE (xs xo : V) (xj : J) : flats (fc ->> [xs, xj, xo]) = xj.1.
Proof.
case: xj=>xj1 xj2; rewrite fsE flatsE /= !pcm_flatPt tp_flatPtPairE.
by case: decP=>//= _; rewrite unitL unitR. 
Qed.

Notation getso s := (getso fcg s).
Notation self s := (self fcg s).
Notation other s := (other fcg s). 
Notation joint s := (joint fcg s).
Notation geth s := (geth fcg s).
Notation getg s := (getg fcg s). 

(* Synchronization protocol between side-chanel array and the array of
ghosts. If there is no request for help, or request is pending for
thread tid, then the ghost field for tid is null. The ghost can be
non-null only if the thread has been helped, but hasn't yet collected
the result. *)

Definition syncrg (r : {ffun 'I_n -> req T}) (g : {ffun 'I_n -> U}) :=
  forall t : 'I_n, if r t is NoReq | Req _ _ then g t = Unit else True.

(* Definition sum_up (g : {ffun 'I_n -> U}) := 
  foldr (fun r t => r \+ t) Unit (fgraph g).
*)

Section ValidRes.
Variable P : mutex -> U -> bool -> heap -> Prop.

Definition validRes s := 
  let: H := self s \+ other s in
  exists (r : {ffun 'I_n -> req T}) (b : bool) (ha hr : heap), 
    [/\ s = fc ->> [self s, joint s, other s], 
        geth s = lk :-> b \+ ha \+ hr, is_array a r ha /\ syncrg r (getg s),
        forall tid, tid \in dom (get_tid H) -> tid < n, 
        valid (getU H \+ sum_up (getg s)) &
        P (get_mtx H) (getU H \+ sum_up (getg s)) b hr].

End ValidRes.

Definition cohRes P := 
  [Pred s | [/\ validH s, validSO s & validRes P s]].

Definition locked m (g : U) (b : bool) (h : heap) := 
  [/\ m = up own, b = true & h = Unit].

Definition unlocked m (g : U) (b : bool) (h : heap) := 
  [/\ m = up nown, b = false & I g h].

Definition either m g b h := locked m g b h \/ unlocked m g b h.

Definition coh := cohRes either.

Lemma locked_either s : s \In cohRes locked -> s \In coh.
Proof. 
case=>Vh Vso [r][n][ha][hr][Es Eh Ha S Tid L]; split=>//.  
by exists r, n, ha, hr; split=>//; left. 
Qed.

Lemma unlocked_either s : s \In cohRes unlocked -> s \In coh.
Proof.
case=>Vh Vso [r][b][ha][hr][Es Eh Ha S Tid L]; split=>//.  
by exists r, b, ha, hr; split=>//; right. 
Qed.

Lemma fc_cohE P s : s \In cohRes P -> s = fc ->> [self s, joint s, other s].
Proof. by case=>_ _ [r][b][ha][hr][]. Qed.

Lemma fc_cohSO P s : s \In cohRes P -> validSO s. 
Proof. by case. Qed.

Lemma fc_cohH P s : s \In cohRes P -> validH s. 
Proof. by case. Qed.

Lemma fc_cohfV P s : s \In cohRes P -> footV_eq s (mem_seq' [:: fc]).
Proof. by case=>_ _ [r][b][ha][hr][->]; rewrite fsE. Qed. 

Lemma fc_coh_weq P s s' : 
        s \In cohRes P -> s' \In cohRes P -> weq typeCode (shr s) (shr s').
Proof.
case=>_ _ [r1][b1][ha1][hr1][-> _ _ _ _ _][_ _][r2][b2][ha2][hr2][-> _ _ _ _ _].
by rewrite !fsE; apply/eqP/feapE/fmapE. 
Qed.

Lemma fc_coh_sig P s : 
        s \In cohRes P -> fc \-> (Unit : V) = [zunit loc s].
Proof. by case=>_ _ [r][b][ha][hr][-> _ _ _ _ _]; rewrite fsE zunitPt. Qed.

Lemma fc_coh_zigzag P s p : (s \< p) \In cohRes P <-> (s \> p) \In cohRes P.
Proof.
split=>C; case: C (C)=>Vh Vso [r][b][ha][hr][H1 H2 H3 H4 H5 H6] C.
- rewrite validH_zigzag // in Vh; rewrite zagzigVE // in H4 H5 H6.
  rewrite validSO_zigzag in Vso.
  move: H1; rewrite [joint _]getE zagzigE => H1; split=>//. 
  by exists r, b, ha, hr; rewrite !getE in H1 H2 H3 H4 H5 H6 *. 
rewrite -validSO_zigzag in Vso; rewrite -validH_zigzag // in Vh.
rewrite -zagzigVE // in H4 H5 H6.
move: H1; rewrite [joint _]getE -zagzigE => H1; split=>//. 
by exists r, b, ha, hr; rewrite !getE in H1 H2 H3 H4 H5 H6 *. 
Qed.

Definition flat_combine_CohMixin :=
  CohPredMixin (@fc_cohSO either) (@fc_cohH either) (@fc_cohfV either) 
               (@fc_coh_weq either) (@fc_coh_sig either) 
               (@fc_coh_zigzag either).
Definition fc_coh := Eval hnf in CohPred coh flat_combine_CohMixin.

Definition flat_combine_GCMixin := @GCMixin _ fc_coh (@fc_cohE either).
Canonical fc_getcoh := Eval hnf in GC fcg fc_coh flat_combine_GCMixin.
     
(* Localizing the existential hell *)

Lemma size_fgraph X (r1 r2 : {ffun 'I_n -> X}) : 
        size (fgraph r1) = size (fgraph r2).
Proof. by rewrite !fgraph_codom /= !codomE !size_map. Qed.

Lemma fgraphE X (r1 r2 : {ffun 'I_n -> X}) : 
        fgraph r1 = fgraph r2 -> r1 = r2.
Proof. by case: r1; case=>r1 H1; case: r2; case=>r2 H2 /= ->. Qed.

Lemma fc_exE (r1 r2 : {ffun 'I_n -> req T}) 
             (b1 b2 : bool) ha1 ha2 hr1 hr2 : 
        valid (lk :-> b1 \+ ha1 \+ hr1) ->
        lk :-> b1 \+ ha1 \+ hr1 = lk :-> b2 \+ ha2 \+ hr2 ->
        is_array a r1 ha1 -> is_array a r2 ha2 ->
        [/\ b1 = b2, r1 = r2, ha1 = ha2 & hr1 = hr2].
Proof.
rewrite -!joinA=>V E Ha1 Ha2; case: Ha1 Ha2 V E => /= -> _ [-> _] V. 
case/(hcancelV V)=><- {V} V; case/(updi_iinv (size_fgraph _ _) V).
by move/val_inj/fgraphE=><-<-.
Qed.

Lemma fc_cohJ P s : s \In cohRes P -> valid (geth s).
Proof. 
by move=>C; move: (fc_cohH C); rewrite {1}(fc_cohE C) /validH fc_flatsE. 
Qed.

Lemma fc_cohP s (b : bool) (r : {ffun 'I_n -> req T}) ha hr P : 
        let: H := self s \+ other s in
        s \In cohRes P ->
        geth s = lk :-> b \+ ha \+ hr ->
        is_array a r ha ->
        [/\ syncrg r (getg s),
            forall tid, tid \in dom (get_tid H) -> tid < n, 
            valid (getU H \+ sum_up (getg s)) &
            P (get_mtx H) (getU H \+ sum_up (getg s)) b hr].
Proof.
move=>C E Ha; case: (C)=>_ _ [r'][b'][ha'][hr'][H1 H2 [Ha' H3] H4 H5 H6].
move: (fc_cohJ C) H2; rewrite E => V H2.
by case: (fc_exE V H2 Ha Ha')=>->-> _ ->. 
Qed.

Lemma fc_cohPX s ha : 
        s \In coh ->
        geth s = lk :-> true \+ ha ->
        exists (r : {ffun 'I_n -> req T}), is_array a r ha.
Proof.
move=>C E; case: (C)=>_ _ [r'][b'][ha'][hr'][H1 H2 [Ha' H3] H4 H5 H6].
move: (fc_cohJ C) H2; rewrite E -!joinA => V H2.
case/(hcancelV V): H2=>? _ ?; subst b' ha.
case: H6; case=>// Tx _ ?; subst hr'.
by exists r'; rewrite unitR.
Qed.

Lemma fc_cohPY s ha hr : 
        s \In coh ->
        geth s = lk :-> false \+ ha \+ hr ->
        I (getU (self s \+ other s) \+ sum_up (getg s)) hr -> 
        exists (r : {ffun 'I_n -> req T}), is_array a r ha. 
Proof.
move=>C E; case: (C)=>_ _ [r'][b'][ha'][hr'][H1 H2 [Ha' H3] H4 H5 H6].
move: (fc_cohJ C) H2; rewrite E -!joinA => V H2.
case/(hcancelV V): H2=>? {V} V H2 inv; subst b'.
rewrite -(joinC hr) -(joinC hr') -joinA in V H2 H6. 
case: H6; case=>// Tx _ inv2; move: (fc_precise V H2 inv inv2)=>?; subst hr'.
by rewrite (joinfK V H2); exists r'.  
Qed.

Lemma sumE (r1 r2 : {ffun 'I_n -> U}) : r1 =1 r2 -> sum_up r1 = sum_up r2.
Proof. by move=>E; congr sum_up; apply/ffunP. Qed.

Lemma sum0R (r : {ffun 'I_n -> U}) x : 
        sum_up r = sum_up [ffun z => if z == x then Unit else r z] \+ r x.
Proof.
rewrite /sum_up !fgraph_codom /= !codomE /=.
have: x \in enum 'I_n by rewrite mem_enum. 
have: uniq (enum 'I_n) by apply: enum_uniq. 
elim: (enum 'I_n)=>[|y ys IH] //=. 
rewrite (ffunE _ y) eq_sym inE.
case/andP; case: eqP=>[<- N _ _|_ N Un]; last first. 
- by move/(IH Un)=>->; rewrite joinA. 
rewrite joinAC unitL; congr (_ \+ foldr _ _ _). 
apply/eq_in_map=>z; rewrite ffunE.
by case: eqP=>// ->; rewrite (negbTE N).  
Qed.

Lemma sum0L (r : {ffun 'I_n -> U}) x : 
        sum_up r = r x \+ sum_up [ffun z => if z == x then Unit else r z].
Proof. by rewrite joinC; apply: sum0R. Qed.

Lemma sumIR (r : {ffun 'I_n -> U}) x u : 
        sum_up r \+ u = sum_up [ffun z => if z == x then r x \+ u else r z].
Proof.
rewrite (sum0R _ x) -joinA [in X in _ = X](sum0R _ x) (ffunE _ x) eq_refl.
by congr (_ \+ _); apply: sumE=>z; rewrite !ffunE; case: eqP.
Qed.

End FlatCombineCoh.

Module Exports.
Notation fc_coh := fc_coh.
Notation fc_flatsE := fc_flatsE.
Notation fcg := fcg.
Notation fc_getcoh := fc_getcoh.
Notation syncrg := syncrg.
Notation sum_up := sum_up.
Canonical fc_getcoh.

(* additional lemma about FlatCombine joint *)

Notation size_fgraph := size_fgraph.
Notation fgraphE := fgraphE.
Notation fc_exE := fc_exE.
Notation fc_cohJ := fc_cohJ.
Notation fc_cohP := fc_cohP.
Notation fc_cohPX := fc_cohPX.
Notation fc_cohPY := fc_cohPY. 

Notation sumE := sumE.
Notation sum0R := sum0R.
Notation sum0L := sum0L.
Notation sumIR := sumIR.

(* How would this look with big ops, just for fun? *)

Section BigOps.
Variable T : flat_combineT.
Notation U := (fc_pcm T).
Notation n := (fc_n T).

Import bigop Monoid.

Canonical U_monoid := Law (@joinA U) (@unitL U) (@unitR U).
Canonical U_comoid := ComLaw (@joinC U).

Lemma sum0R' (r : {ffun 'I_n -> U}) (x : 'I_n) : 
        \big [PCM.join/Unit]_(i < n) r i = 
        \big [PCM.join/Unit]_(i < n | i != x) r i \+ r x.
Proof. by rewrite (bigD1 x) //= joinC. Qed.

Lemma sumIR' (r : {ffun 'I_n -> U}) (x : 'I_n) (u : U) : 
        \big [PCM.join/Unit]_(i < n) r i \+ u = 
        \big [PCM.join/Unit]_(i < n) (if i == x then r i \+ u else r i).
Proof.
rewrite (bigD1 x) //= [X in _ = X](bigD1 x) //= eq_refl joinAC.
by congr (_ \+ _); apply: eq_bigr=>i /negbTE=>->. 
Qed.

End BigOps.

End Exports.

End FlatCombineCoh.

Export FlatCombineCoh.Exports.


Module Transitions.
Section Transitions. 
Variable T : flat_combineT.
Notation a := (fc_orig T).
Notation lk := (fc_lk T).
Notation n := (@fc_n T).
Notation fc := (@fc_label T). 
Notation U := (@fc_pcm T).
Notation I := (@fc_inv T).
Notation V := (fc_V T).
Notation J := (fc_J T).
Notation coh := (fc_coh T).
Notation getso s := (getso (fcg T) s).
Notation self s := (self (fcg T) s).
Notation other s := (other (fcg T) s).
Notation joint s := (joint (fcg T) s).
Notation geth s := (geth (fcg T) s).
Notation getg s := (getg (fcg T) s). 

(*
Internal transitions for flat combiner:
- transition for requesting help
- transition for helping (only the locking thread can do it)
- transition for collecting the help results
*)

Section ReqHelpTrans.
Variable tid : 'I_n.    (* the thread id *)
Variable m : fc_code T. (* the method that tid request help with *)
Variable v : fc_inT m.  (* the arguments to the method m *)

Definition reqhelp_rel s s' :=
  exists (r : {ffun 'I_n -> req T}) (b : bool) ha hr ha',
    [/\ nat_of_ord tid \in dom (get_tid (self s)),
        geth s = lk :-> b \+ ha \+ hr /\ is_array a r ha /\ r tid = NoReq T,
        self s' = self s,
        joint s' = (lk :-> b \+ ha' \+ hr, getg s) /\ 
        is_array a [ffun z => if z == tid then Req v else r z] ha',
        other s' = other s & 
        s \In coh /\ s' \In coh].

Lemma reqhelp_coh s s' : reqhelp_rel s s' -> s \In coh /\ s' \In coh.
Proof. by case=>r [b][ha][hr][ha'][]. Qed.

Lemma reqhelp_env s s' : reqhelp_rel s s' -> env s = env s'.
Proof. 
case=>r [b][ha][hr][ha'][_ _ _ _ E][C C']. 

by rewrite (cohE C) (cohE C') !fsE /= E. 
Qed.

Lemma reqhelp_zagzig s s' p : 
        env s = env s' ->
        reqhelp_rel (s \> p) (s' \> p) -> reqhelp_rel (s \< p) (s' \< p).
Proof.
move=>E [r][b][ha][hr][ha'][Tid][Hh][Ha Hr] S [J Ha'] _.
case=>/coh_zigzag C /coh_zigzag C'; exists r, b, ha, hr, ha'.
rewrite !getE in Hh S J Tid *; rewrite !getzC //= S E; split=>//.
rewrite domUn /= inE Tid andbT; move/validL: (cohVSO C).
by rewrite getzC //= ; case/andP; case/andP. 
Qed.

Lemma reqhelp_footH s s' : reqhelp_rel s s' -> footeqH (flats s) (flats s').
Proof.
case=>r [b][ha][hr][ha'][Tid][E1][H1 _] S [E2 H2] O [C C']. 
rewrite (cohE C) (cohE C') /= !fc_flatsE E1 E2 /=.
by apply: footeqHUnL; apply: footeqHUnR; apply: array_footeqH H1 H2.
Qed.

End ReqHelpTrans.


Section DoHelpTrans.
Variable tid : 'I_n.    (* the thread id to whom we're helping *)
Variable m : fc_code T. (* the method that tid requests help with *)
Variable v : fc_inT m.  (* the arguments to the method m *)
Variable w : fc_outT m. (* the result is determined by the caller *)

Definition dohelp_rel s s' := 
  let: g := getU (self s \+ other s) \+ sum_up (getg s) in
  exists (r : {ffun 'I_n -> req T}) (ha ha' : heap) g',
    [/\ get_mtx (self s) = up own,
        geth s = lk :-> true \+ ha /\ is_array a r ha /\ r tid = Req v,
        self s' = self s, 
        joint s' = (lk :-> true \+ ha',
                    [ffun z => if z == tid then g' else getg s z]) /\
        is_array a [ffun z => if z == tid then Ans w else r z] ha',
        other s' = other s & valid (g \+ g') /\ fc_R v w g g' /\
        s \In coh /\ s' \In coh].

Lemma dohelp_coh s s' : dohelp_rel s s' -> s \In coh /\ s' \In coh.
Proof. by case=>r [ha][ha'][g'][_ _ _ _ _][_][]. Qed.

Lemma dohelp_env s s' : dohelp_rel s s' -> env s = env s'.
Proof. 
case=>r [ha][ha'][g'][_ _ _ _ E][_][_][C C'].
by rewrite (cohE C) (cohE C') !fsE E. 
Qed.

Lemma dohelp_zagzig s s' p : 
        env s = env s' ->
        dohelp_rel (s \> p) (s' \> p) -> dohelp_rel (s \< p) (s' \< p).
Proof.
move=>E [r][ha][ha'][g'][Mx Ha] S [E1 E2] O [V][R]. 
case=>/coh_zigzag C /coh_zigzag C'; exists r, ha, ha', g'. 
rewrite !getE !getzC //= in Mx Ha S E1 E2 V R *.
rewrite S {1}E; rewrite -!joinA in V R *; split=>//.
move/validL: (cohVSO C); rewrite getzC //=; case/andP; case/andP=>/= _ W _. 
rewrite /get_mtx /= in Mx *; rewrite Mx in W *.
by case: ((getso p).1.2) W=>//; case. 
Qed.

Lemma dohelp_footH s s' : dohelp_rel s s' -> footeqH (flats s) (flats s').
Proof.
case=>r [ha][ha'][g'][_ [E1 [H1 _]]] S [E2 H2] _ [_][_][C C']. 
rewrite (cohE C) (cohE C') /= !fc_flatsE E1 E2 /=.
by apply: footeqHUnR; apply: array_footeqH H1 H2. 
Qed.

End DoHelpTrans.


Section CollectHelpTrans.
Variable tid : 'I_n.    (* the thread id which is collecting the help *)
Variable m : fc_code T. (* the method for which tid obtained help (probably unnecessary) *)
Variable w : fc_outT m. (* the result of the help *)

Definition collect_rel (s s' : state) := 
  exists (r : {ffun 'I_n -> req T}) (b : bool) ha hr ha',
    [/\ nat_of_ord tid \in dom (get_tid (self s)),
        geth s = lk :-> b \+ ha \+ hr /\ is_array a r ha /\ r tid = Ans w, 
        self s' = (get_tid (self s), get_mtx (self s), 
                   getU (self s) \+ getg s tid),
        geth s' = lk :-> b \+ ha' \+ hr /\ 
        is_array a [ffun z => if z == tid then NoReq T else r z] ha' /\
        getg s' = [ffun z => if z == tid then Unit else getg s z],
        other s' = other s & 
        s \In coh /\ s' \In coh].

Lemma collect_coh s s' : collect_rel s s' -> s \In coh /\ s' \In coh.
Proof. by case=>r [b][ha][hr][ha'][]. Qed.

Lemma collect_env s s' : collect_rel s s' -> env s = env s'.
Proof. 
case=>r [b][ha][hr][ha'][_ _ _ _ E][C C'].
by rewrite (cohE C) (cohE C') !fsE E. 
Qed.

Lemma collect_zagzig s s' p : 
        env s = env s' ->
        collect_rel (s \> p) (s' \> p) -> collect_rel (s \< p) (s' \< p).
Proof.
move=>E [r][b][ha][hr][ha'][Tid G1 S G2 _]. 
case=>/coh_zigzag C /coh_zigzag C'; exists r, b, ha, hr, ha'.
rewrite !getE !getzC //= in Tid G1 S G2 *; rewrite E. 
split=>//; last by rewrite S pcmE /= pcmE /= joinAC. 
rewrite domUn /= inE Tid /= andbT; move/validL: (cohVSO C). 
by rewrite getzC //; case/andP; case/andP.
Qed.

Lemma collect_footH s s' : collect_rel s s' -> footeqH (flats s) (flats s').
Proof.
case=>r [b][ha][hr][ha'][Tid [E1 [H1 _]] S [E2 [H2 _]] O][C C'].
rewrite (cohE C) (cohE C') /= !fc_flatsE E1 E2. 
by apply: footeqHUnL; apply: footeqHUnR; apply: array_footeqH H1 H2. 
Qed.

End CollectHelpTrans.

(* External transitions for the flat combiner *)

Section AcquireTrans.
Variable hr : heap.

Definition acq_rel s s' := 
  let: g := getU (self s \+ other s) \+ sum_up (getg s) in
  exists ha,
    [/\ get_mtx (self s) = up own /\ geth s = lk :-> true \+ ha, 
        self s' = (get_tid (self s), up nown, getU (self s)), 
        joint s' = (lk :-> false \+ ha \+ hr, getg s), 
        other s' = other s, I g hr & 
        s \In coh /\ s' \In coh].
  
Lemma acq_coh s s' : acq_rel s s' -> s \In coh /\ s' \In coh.
Proof. by case=>ha []. Qed.

Lemma acq_env s s' : acq_rel s s' -> env s = env s'.
Proof. by case=>ha [_ _ _ E _][C C']; rewrite (cohE C) (cohE C') !fsE E. Qed.

Lemma acq_zagzig s s' p : 
        env s = env s' ->
        acq_rel (s \> p) (s' \> p) -> acq_rel (s \< p) (s' \< p).
Proof.
move=>E [ha][[Mx H] S J _ inv][/coh_zigzag C /coh_zigzag C'].
exists ha; rewrite !getE !getzC //= in Mx H S J inv *; rewrite {1}E.
move/validL: (cohVSO C); rewrite getzC //=; case/andP; case/andP=>/= _ V _. 
rewrite /get_mtx /= in Mx *; rewrite Mx in V *. 
split=>//; last by [rewrite -!joinA in inv *].
- by case: ((getso p).1.2) V=>//; case. 
rewrite S pcmE /= pcmE /=. 
by case: ((getso p).1.2) V=>//; case.
Qed.

Lemma acq_footH s s' : acq_rel s s' -> footeqH (flats s \+ hr) (flats s').
Proof.
case=>ha [[_ E1] _ E2 _ _][C C'].
rewrite (cohE C) (cohE C') !fc_flatsE E1 E2 /= -!joinA. 
by apply: footeqHUnL; apply: footeqHPt. 
Qed.

End AcquireTrans.


Section ReleaseTrans.
Variable hr : heap.

Definition rel_rel s s' := 
  let: g := getU (self s \+ other s) \+ sum_up (getg s) in
  exists ha,
    [/\ geth s = lk :-> false \+ ha \+ hr, I g hr, 
        self s' = (get_tid (self s), up own, getU (self s)), 
        joint s' = (lk :-> true \+ ha, getg s), 
        other s' = other s &
        s \In coh /\ s' \In coh].

Lemma rel_coh s s' : rel_rel s s' -> s \In coh /\ s' \In coh.
Proof. by case=>ha []. Qed.

Lemma rel_env s s' : rel_rel s s' -> env s = env s'.
Proof. by case=>ha [_ _ _ _ E][C C']; rewrite (cohE C) (cohE C') !fsE E. Qed.

Lemma rel_zagzig s s' p : 
        env s = env s' ->
        rel_rel (s \> p) (s' \> p) -> rel_rel (s \< p) (s' \< p).
Proof.
move=>E [ha][H inv S J _][/coh_zigzag C /coh_zigzag C'].
exists ha; rewrite !getE !getzC //= in H inv S J *; rewrite {2}E.
suff Mx : get_mtx (self s \+ getso p) = up nown.
- split=>//; first by rewrite -!joinA in H inv *. 
  by rewrite S pcmE /= pcmE /=; case/mtx00E: Mx=>_ ->; rewrite unitR. 
have V : valid (geth s) by move: (fc_cohJ C); rewrite getE. 
case: (C)=>_ _ [r][b][ha1][hr1][_ E'] _ _ _.
rewrite getE H -!joinA in E' V; case/(hcancelV V): E'=><-{b} _ _. 
by case; case=>//; rewrite getE getzC //=; case/mtx00E.
Qed.

Lemma rel_footH s s' : rel_rel s s' -> footeqH (flats s) (hr \+ flats s').
Proof.
case=>ha [E1 _ _ E2 _][C C'].
rewrite (cohE C) (cohE C') !fc_flatsE E1 E2 /= joinCA (joinC hr) -joinA.
by apply: footeqHUnL; apply: footeqHPt. 
Qed.

End ReleaseTrans.

End Transitions.

Module Exports. 
Section Exports. 
Variable T : flat_combineT.

(* request help internal transition *)
Definition ReqHelpGuarMixin tid m v := 
  GuarMixin (@reqhelp_coh T tid m v) (@reqhelp_env T tid m v) 
            (@reqhelp_zagzig T tid m v). 
Canonical reqhelp_guar tid m v := 
  Eval hnf in Guar (@reqhelp_rel T tid m v) (@ReqHelpGuarMixin tid m v). 
Definition ReqHelpIntGuarMixin tid m v := 
  IntGuarMixin (@reqhelp_footH T tid m v).
Canonical reqhelp_intguar tid m v := 
  Eval hnf in IntGuar (@reqhelp_guar tid m v) (@ReqHelpIntGuarMixin tid m v). 

(* help internal transition *)
Definition DoHelpGuarMixin tid m v w := 
  GuarMixin (@dohelp_coh T tid m v w) (@dohelp_env T tid m v w) 
            (@dohelp_zagzig T tid m v w). 
Canonical dohelp_guar tid m v w := 
  Eval hnf in Guar (@dohelp_rel T tid m v w) (@DoHelpGuarMixin tid m v w). 
Definition DoHelpIntGuarMixin tid m v w := 
  IntGuarMixin (@dohelp_footH T tid m v w).
Canonical dohelp_intguar tid m v w := 
  Eval hnf in IntGuar (@dohelp_guar tid m v w) (@DoHelpIntGuarMixin tid m v w). 

(* collect internal transition *)
Definition CollectGuarMixin tid m w := 
  GuarMixin (@collect_coh T tid m w) (@collect_env T tid m w) 
            (@collect_zagzig T tid m w). 
Canonical collect_guar tid m w := 
  Eval hnf in Guar (@collect_rel T tid m w) (@CollectGuarMixin tid m w). 
Definition CollectIntGuarMixin tid m w := 
  IntGuarMixin (@collect_footH T tid m w).
Canonical collect_intguar tid m w := 
  Eval hnf in IntGuar (@collect_guar tid m w) (@CollectIntGuarMixin tid m w). 

(* acquire transition (i.e. unlock) *)
Definition baseFCAcqGuarMixin hr := 
  GuarMixin (@acq_coh T hr) (@acq_env T hr) (@acq_zagzig T hr). 
Canonical fc_acq_guar hr := 
  Eval hnf in Guar (acq_rel T hr) (@baseFCAcqGuarMixin hr). 
Definition FCAcqGuarMixin := AcqGuarMixin (@acq_footH T).
Canonical FCAcqGuar := Eval hnf in AcqGuar fc_acq_guar FCAcqGuarMixin.

(* release transition (i.e. cas) *)
Definition baseFCRelGuarMixin hr := 
  GuarMixin (@rel_coh T hr) (@rel_env T hr) (@rel_zagzig T hr). 
Canonical fc_rel_guar hr := 
  Eval hnf in Guar (rel_rel T hr) (@baseFCRelGuarMixin hr). 
Definition FCRelGuarMixin := RelGuarMixin (@rel_footH T).
Canonical FCRelGuar := Eval hnf in RelGuar fc_rel_guar FCRelGuarMixin.

(* the concurroid *)
Canonical FlatCombineCore : mod_core := 
  Mod.Core [PredU [Pred r | exists tid m v, r = @reqhelp_intguar tid m v] & 
           [PredU [Pred r | exists tid m v w, r = @dohelp_intguar tid m v w] &
                  [Pred r | exists tid m w, r = @collect_intguar tid m w]]].
Definition FCChanCore := ChanCore FCAcqGuar FCRelGuar.
Canonical FlatCombine : mod := @Mod.Make FlatCombineCore (Pred1 FCChanCore).

End Exports. 
End Exports.

End Transitions.

Export Transitions.Exports. 


(***********)
(* Actions *)
(***********)

Module Actions.
Section Actions. 
Variable T : flat_combineT.
Notation a := (fc_orig T).
Notation lk := (fc_lk T).
Notation n := (@fc_n T).
Notation fc := (@fc_label T). 
Notation U := (@fc_pcm T).
Notation I := (@fc_inv T).
Notation V := (fc_V T).
Notation J := (fc_J T).

Section Internals.
Notation coh := (fc_coh T).
Notation getso s := (getso (fcg T) s).
Notation self s := (self (fcg T) s).
Notation other s := (other (fcg T) s).
Notation joint s := (joint (fcg T) s).
Notation geth s := (geth (fcg T) s).
Notation getg s := (getg (fcg T) s). 
Notation guar := (Mod.internal (FlatCombine T)). 

Section RequestHelp.
Variables (tid : 'I_n) (m : fc_code T) (v : fc_inT m).

Definition reqhelp_safe s := 
  exists (r : {ffun 'I_n -> req T}) (b : bool) ha hr,
    [/\ nat_of_ord tid \in dom (get_tid (self s)), 
        geth s = lk :-> b \+ ha \+ hr, is_array a r ha,
        r tid = NoReq T & 
        s \In coh].

Definition reqhelp_step s s' (x : unit) := 
  reqhelp_intguar tid v s s'.

Lemma reqhelp_safe_coh : safe_coh coh reqhelp_safe.
Proof. by move=>s [r][b][ha][hr][]. Qed.

Lemma reqhelp_step_safe : step_safe reqhelp_safe reqhelp_step.
Proof. 
move=>s s' [][r][b][ha][hr][ha'][?[?[??]]???][C _].
by exists r, b, ha, hr. 
Qed.

Lemma reqhelp_safe_zagzig : safe_zagzig reqhelp_safe.
Proof. 
move=>s t [r][b][ha][hr][Hd Eh Ha Er] /coh_zigzag C.
exists r, b, ha, hr; rewrite !getE getzC //= in Hd Eh *. 
split=>//; rewrite domUn /= inE Hd andbT; move/validL: (cohVSO C).
by rewrite getzC //= ; case/andP; case/andP. 
Qed.

Lemma reqhelp_step_guar : step_guar guar reqhelp_step.
Proof.
move=>s1 s2 [H]; right; exists (reqhelp_intguar tid v).  
by split=>//; left; exists tid, m, v.  
Qed.

Lemma reqhelp_step_total : step_total reqhelp_safe reqhelp_step.
Proof. 
move=>s [r][b][ha][hr][Ht Eh Ha Er C].
set r' := [ffun z => if z == tid then Req v else r z].
set ha' := updi a (fgraph r'). 
set s' := fc ->> [self s, (lk :-> b \+ ha' \+ hr, getg s), other s].
have Ha' : is_array a r' ha'.
- by split=>//; case: Ha=>->; rewrite /ha' !updiD (size_fgraph r r'). 
suff C' : s' \In coh by exists s', tt, r, b, ha, hr, ha'; rewrite !getE.
split. 
- move: (cohH C); rewrite (cohE C) /s' /=. 
  rewrite /validH !fc_flatsE /= Eh /ha' -!(joinAC _ hr).
  by rewrite (footeqH_validUnR (array_footeqH Ha Ha')).
- by rewrite /validSO /s' fsE /= zipPt allvalidPt; apply: cohVSO C.
case: (fc_cohP C Eh Ha)=>Sx Tx Vx Fx.
exists r', b, ha', hr; rewrite !getE; do !split=>//=.
by move=>t; move: (Sx t); rewrite ffunE; case: eqP=>// ->; rewrite Er.
Qed.

Lemma reqhelp_step_frame : step_frame reqhelp_safe reqhelp_step.
Proof. 
move=>s s' t [] Cx [r][b][ha][hr][ha'][Ht [Eh][Ha] Er S [Ej] Ha' O][C C'].
rewrite !getE getzC //= in  Ht Eh S Ej O.
set s'' := fc ->> [self s, joint s', other s'].
have E : s' = s'' \< t.
- by rewrite (zigSOE C) {1}(cohE C') /s'' !fsE zigE zipPt S joinC.
rewrite E in C' *; exists s''; split=>//.
exists r, b, ha, hr, ha'; rewrite !getE; split=>//.
- by case: Cx=>_ [_][_][_][Tx _ _ _ _]; rewrite getE in Tx.
- by rewrite !getzC //= getE O.
by split; apply/coh_zigzag.  
Qed.

Definition reqhelp_pre_act := 
  Action reqhelp_safe_coh reqhelp_safe_zagzig reqhelp_step_safe 
         reqhelp_step_guar reqhelp_step_frame reqhelp_step_total.

Lemma reqhelp_step_erase : int_step_erase (reqhelp_pre_act).
Proof.
move=>s1 h1 t1 [] s2 h2 t2 [] V E H1 H2; move: H1 H2 V E.
case=>r1 [b1][ha1][hr1][ha1'][Tx1][Eh1][Ha1 Er1] Es1 [Ej1 Ha1'] O1 [C1 C1'].
case=>r2 [b2][ha2][hr2][ha2'][Tx2][Eh2][Ha2 Er2] Es2 [Ej2 Ha2'] O2 [C2 C2'].
rewrite (cohE C1) (cohE C2) /= !fc_flatsE Eh1 Eh2.
rewrite -!(joinA _ hr1) -!(joinA _ hr2) => V E.
rewrite (cohE C1') (cohE C2') /= !fc_flatsE Ej1 Ej2 /=.
rewrite -!joinA; case: (fc_exE V E Ha1 Ha2)=>??? ->; subst b2 r2 ha2.
by case: Ha1' Ha2' => -> _ [-> _].
Qed.

End RequestHelp. 

Definition reqhelp_act tid m v := IntAction (@reqhelp_step_erase tid m v). 


Section DoHelp.
Variables (tid : 'I_n) (m : fc_code T) (w : fc_outT m).

(* we need to know that a result and ghost *)
(* corresponding to m can be found *)
Definition dohelp_safe s := 
  let: g := getU (self s \+ other s) \+ sum_up (getg s) in
  exists (r : {ffun 'I_n -> req T}) (b : bool) ha hr g' v,
    [/\ get_mtx (self s) = up own, 
        geth s = lk :-> b \+ ha \+ hr, is_array a r ha,
        r tid = Req v, valid (g \+ g') /\ fc_R v w g g' & 
        s \In coh].

Definition dohelp_step s s' (x : unit) := 
  exists v, dohelp_intguar tid v w s s'.

Lemma dohelp_safe_coh : safe_coh coh dohelp_safe.
Proof. by move=>s [r][b][ha][hr][g'][v][]. Qed.

Lemma dohelp_step_safe : step_safe dohelp_safe dohelp_step.
Proof. 
move=>s s' [][v][r][ha][ha'][g'][Tx][Eh][Ha] Er _ _ _ [V][R][C _].
by exists r, true, ha, Unit, g', v; rewrite unitR.
Qed.

Lemma dohelp_safe_zagzig : safe_zagzig dohelp_safe.
Proof. 
move=>s t [r][b][ha][hr][g'][v][Em Eh Ha Er VR] /coh_zigzag C.
exists r, b, ha, hr; rewrite !getE !getzC //= in Em Eh VR *.
exists g', v; rewrite -!joinA in Eh VR *; split=>//.
move/validL: (cohVSO C); rewrite getzC //=; case/andP; case/andP. 
rewrite /get_mtx /= in Em *; rewrite Em.
by case: ((getso t).1.2)=>//; case.
Qed.

Lemma dohelp_step_guar : step_guar guar dohelp_step.
Proof.
move=>s1 s2 [][v H]; right; exists (dohelp_intguar tid v w).  
by split=>//; right; left; exists tid, m, v, w.  
Qed.

Lemma dohelp_step_total : step_total dohelp_safe dohelp_step.
Proof. 
move=>s [r][b][ha][hr][gx][vx][Em Eh Ha Er [V R] C]. 
have [??] : b = true /\ hr = Unit.
- case: (C)=>_ _ [r1][b1][ha1][hr1][_ Eh1][Ha1] Sx _ _ F.
  move: (fc_cohJ C); rewrite Eh in Eh1 * => Vx.
  case: (fc_exE Vx Eh1 Ha Ha1)=>????; subst b1 r1 ha1 hr1.
  case: F; case=>//; rewrite /get_mtx /= in Em *; rewrite Em. 
  by case: (other s).1.2=>//; case=>//.
subst b hr; rewrite unitR in Eh.
set r' := [ffun z => if z == tid then Ans w else r z].
set g' := [ffun z => if z == tid then gx else getg s z].
set ha' := updi a (fgraph r').
set s' := fc ->> [self s, (lk :-> true \+ ha', g'), other s].
have Ha' : is_array a r' ha'.
- by split=>//; case: Ha=>->; rewrite /ha' !updiD (size_fgraph r r'). 
suff C' : s' \In coh by exists s', tt, vx, r, ha, ha', gx; rewrite !getE.   
split. 
- move: (cohH C); rewrite (cohE C) /s' /validH !fc_flatsE /= Eh.
  by rewrite (footeqH_validUnR (array_footeqH Ha Ha')).
- by rewrite /validSO /s' fsE /= zipPt allvalidPt; apply: cohVSO C.
rewrite -[_ \+ _]unitR in Eh; case: (fc_cohP C Eh Ha)=>Sx Tx Vx Fx.
exists r', true, ha', Unit; rewrite unitR !getE; do !split=>//=; last 1 first.
- by case: Fx; [left | case].
- by move=>tid'; rewrite !ffunE; case: eqP=>// _; apply: Sx.
suff -> : sum_up g' = sum_up (getg s) \+ gx. by rewrite joinA.
rewrite /g'; rewrite (sumIR (getg s) tid gx).
by move: (Sx tid); rewrite Er => ->; rewrite unitL.
Qed.

Lemma dohelp_step_frame : step_frame dohelp_safe dohelp_step.
Proof. 
move=>s s' t []Cx[v][r][ha][ha'][g'][Em [Eh [Ha Er]] S [J Ha'] O [V][R]][C C']. 
rewrite !getE getzC //= in Em Eh S J O V R.  
set s'' := fc ->> [self s, joint s', other s'].
have E : s' = s'' \< t.
- by rewrite (zigSOE C) {1}(cohE C') /s'' !fsE zigE zipPt S joinC.
rewrite E in C' *; exists s''; split=>//.
exists v, r, ha, ha', g'; rewrite !getE !getzC //; split=>//=.
- by case: Cx=>_ [_][_][_][_][_][Tx] _ _ _ _ _; rewrite getE in Tx.
- by rewrite getE O.
by rewrite -!joinA in V R *; do 3!split=>//; apply/coh_zigzag.  
Qed.

Definition dohelp_pre_act := 
  Action dohelp_safe_coh dohelp_safe_zagzig dohelp_step_safe 
         dohelp_step_guar dohelp_step_frame dohelp_step_total.

Lemma dohelp_step_erase : int_step_erase (dohelp_pre_act).
Proof.
move=>s1 h1 t1 [] s2 h2 t2 [] V E H1 H2; move: H1 H2 V E.
case=>v1 [r1][ha1][ha1'][g1][Em1][Eh1][Ha1 E1] S1 [J1 Ha1'] O1 [V1][R1][C1 C1'].
case=>v2 [r2][ha2][ha2'][g2][Em2][Eh2][Ha2 E2] S2 [J2 Ha2'] O2 [V2][R2][C2 C2'].
rewrite (cohE C1) (cohE C2) /= !fc_flatsE Eh1 Eh2 => V E.
rewrite (cohE C1') (cohE C2') /= !fc_flatsE J1 J2 /=.
case: (fc_exE V E Ha1 Ha2)=>_ ?? ->; subst r2 ha2.
by case: Ha1' Ha2' => -> _ [-> _]. 
Qed.

End DoHelp. 

Definition dohelp_act tid m w := IntAction (@dohelp_step_erase tid m w). 


Section TryCollectHelp.
Variables (tid : 'I_n) (m : fc_code T) (v : fc_inT m).

(* The action directly specifies the method and input values for which
it expects the help. It also issues a precondition "requested" to
ensure that the help is appropriate. Requested is satisfiable by
stability. That is, if a thread requested help for method m with input
v, then by stability, it can be proved that any answer it got must be
appropriate for that method and input. This is so because the helping
transition of the concurroid is defined to respect this property.

Also: the action is of a "try" style. That is, it tests the tid entry
in the side channel array. If it realizes that the help did not yet
arrive, it returns a None. If the help did arrive in the form of a
value w (which, thus has the appropriate type), the action returns
Some w. 

I make v an argument to the method. Ideally, v should be a logical
variable, but that requires nontrivial refactoring to pull all the
invariants through. The chief change will be making Ans take *three*
parameters, i.e. Ans c v w (as opposed to the current Ans c w), in
order to record that w is a suitable helping answer for input
v. Correspondingly, the second case of "requested" below should make
sure that v remains invariant. This isn't too pressing an issue, so I
forego it, and just take v as a proper function argument. *)

Let A := fc_outT m.

Definition cast (X : fc_code T -> Type) m1 m2 (r : m1 = m2) (e : X m1) :=
  match r in (_ = m2) return X m2 with erefl => e end.

Definition eqc X m' (r : m' = m') (e : X m') : cast r e = e.
Proof. by move: r; apply: Streicher_K. Qed.

(* requested has to keep track of the ghosts at the moment of request *)

Definition requested (r : {ffun 'I_n -> req T}) ig (gs : {ffun 'I_n -> U}) g := 
  [/\ forall m' v', r tid = @Req _ m' v' -> 
        exists pf : m = m', cast pf v = v' /\ [ig <<= g \+ sum_up gs]  &
      forall m' w, r tid = @Ans _ m' w -> 
        exists (pf : m' = m) g1, 
          [/\ [ig <<= g1], [g1 <<= g \+ sum_up gs] & 
              fc_R v (cast pf w) g1 (gs tid)]].

Notation reqd r g s := (requested r g (getg s) (getU (self s \+ other s))).

Lemma reqd_zigzag r g s p : 
        s \< p \In coh -> (reqd r g (s \< p) <-> reqd r g (s \> p)).
Proof.
move=>C; split; case=>H1 H2; split=>m' w'; 
rewrite !getE !getzC //= joinA in H1 H2 *;
by [apply: H1 | apply: H2].
Qed.

Implicit Arguments reqd_zigzag [r g s p].
   
Lemma cpf (r : {ffun 'I_n -> req T}) g m' w s : 
        reqd r g s -> r tid = @Ans _ m' w -> m' = m.
Proof. by move=>S R; case: S=>_ /(_ _ _ R) []. Qed.

Definition try_collect_safe s := 
  exists (r : {ffun 'I_n -> req T}) g (pf : reqd r g s) (b : bool) ha hr,
    [/\ nat_of_ord tid \in dom (get_tid (self s)), 
        geth s = lk :-> b \+ ha \+ hr, is_array a r ha & 
        s \In coh].

Definition try_collect_step s s' (x : option A) := 
  exists (r : {ffun 'I_n -> req T}) g (pf : reqd r g s) (b : bool) ha hr ha',
    [/\ nat_of_ord tid \in dom (get_tid (self s)),
        geth s = lk :-> b \+ ha \+ hr /\ is_array a r ha,
        forall g (pf : reqd r g s),
        match r tid as help return r tid = help -> Prop with
          Ans m w => fun pf' =>
          [/\ self s' = (get_tid (self s), get_mtx (self s), 
                         getU (self s) \+ getg s tid),
              geth s' = lk :-> b \+ ha' \+ hr /\ 
              is_array a [ffun z => if z == tid then NoReq T else r z] ha',
              getg s' = [ffun z => if z == tid then Unit else getg s z],
              other s' = other s &
              x = Some (cast (cpf pf pf') w)]
        | _ => fun _ => s' = s /\ x = None
        end (erefl _) & 
        s \In coh /\ s' \In coh].

Lemma trycollect_safe_coh : safe_coh coh try_collect_safe.
Proof. by move=>s [r][g][pf][b][ha][hr][]. Qed.

Lemma trycollect_step_safe : step_safe try_collect_safe try_collect_step.
Proof. 
move=>s s' x [r][g][pf][b][ha][hr][ha'][Tx [Eh Ha] _][C _].  
by exists r, g, pf, b, ha, hr. 
Qed.

Lemma trycollect_safe_zagzig : safe_zagzig try_collect_safe.
Proof. 
move=>s t [r][g][pf][b][ha][hr][Tx Eh Ha] /coh_zigzag C.
move/(reqd_zigzag C): pf=>pf. 
exists r, g, pf, b, ha, hr; rewrite !getE getzC //= in Tx Eh *.
split=>//; rewrite domUn /= inE Tx andbT; move/validL: (cohVSO C).
by rewrite getzC //= ; case/andP; case/andP. 
Qed.

Lemma trycollect_step_guar : step_guar guar try_collect_step.
Proof.
move=>s1 s2 x [r][g][pf][b][ha][hr][ha'][Tx [Eh Ha] X][C C'].
move: (@cpf r) X; case E: (r tid)=>[||m' w].
- by move=>H /(_ _ pf) [-> _]; apply: refl_guar_refl. 
- by move=>H /(_ _ pf) [-> _]; apply: refl_guar_refl.
move=>_ /(_ _ pf) [S][Eh' Ha' Eg O] _.
right; exists (collect_intguar tid w); split. 
- by right; right; exists tid, m', w. 
by exists r, b, ha, hr, ha'.
Qed.

Lemma trycollect_step_total : step_total try_collect_safe try_collect_step.
Proof. 
move=>s [r][g][pf][b][ha][hr][Ht Eh Ha C].
case E: (r tid)=>[||m' w].
- by exists s, None, r, g, pf, b, ha, hr, ha; move: (@cpf r); rewrite E.
- by exists s, None, r, g, pf, b, ha, hr, ha; move: (@cpf r); rewrite E.
set r' := [ffun z => if z == tid then NoReq T else r z].
set ha' := updi a (fgraph r'). 
set g' := [ffun z => if z == tid then Unit else getg s z].
set s' := (get_tid (self s), get_mtx (self s), getU (self s) \+ getg s tid).
set s2 := fc ->> [s', (lk :-> b \+ ha' \+ hr, g'), other s].
have Ha' : is_array a r' ha'.
- by split=>//; case: Ha=>->; rewrite /ha' !updiD (size_fgraph r r'). 
have Es : getU (self s \+ other s) \+ sum_up (getg s) = 
          getU (self s) \+ (getg s) tid \+ getU (other s) \+ sum_up g'.
- by rewrite (sum0L _ tid) -!joinA (joinCA _ (getg s _)). 
suff C2 : s2 \In coh. 
- exists s2, (Some (cast (cpf pf E) w)), r, g, pf, b, ha, hr, ha'. 
  move: (@cpf r) (E); rewrite !getE E /= => pf' E'; split=>// g1 pf1. 
  move: (pf' _ _ _ _ _ _)=>pfx; move: (pf' _ _ _ _ _ _)=>pfy. 
  by subst m'; rewrite !eqc. 
case: (fc_cohP C Eh Ha)=>Sx Tx Vx Fx.
split. 
- move: (cohH C); rewrite (cohE C) /s' /=. 
  rewrite /validH !fc_flatsE /= Eh /ha' -!(joinAC _ hr).
  by rewrite (footeqH_validUnR (array_footeqH Ha Ha')).
- rewrite /validSO /s2 fsE /= zipPt allvalidPt.
  apply/andP; split; first by case/andP: (cohVSO C).
  by rewrite Es in Vx; apply: validL Vx.
exists r', b, ha', hr; rewrite !getE -Es; do !split=>//=.
by move=>t; move: (Sx t); rewrite !ffunE; case: eqP.
Qed.

Lemma trycollect_step_frame : step_frame try_collect_safe try_collect_step.
Proof. 
move=>s s' t x Cx [r][g][pf][b][ha][hr][ha'][Tx [Eh Ha] /(_ _ pf) X][C C'].
rewrite !getE !getzC //= in X Tx Eh. 
have Tx' : nat_of_ord tid \in dom (get_tid (self s)).
- by case: Cx=>? [?][_][?][?][?][]; rewrite getE. 
move/(reqd_zigzag C): (pf)=>pfz.
move: (@cpf r) X; case Er : (r tid)=>[||m' w].
- move=>H [-> ->]; exists s; split=>//. 
  exists r, g, pfz, b, ha, hr, ha'. 
  split=>//; last by split; apply/coh_zigzag.
  - by rewrite !getE. 
  - by rewrite !getE.
  by move: (@cpf r); rewrite Er. 
- move=>H [-> ->]; exists s; split=>//. 
  exists r, g, pfz, b, ha, hr, ha'.
  split=>//; last by split; apply/coh_zigzag.
  - by rewrite !getE.
  - by rewrite !getE.
  by move: (@cpf r); rewrite Er. 
move=>pf' [S][Eh' Ha'] Eg O Ex.
set sx := (get_tid (self s), get_mtx (self s), getU (self s) \+ getg s tid).
set s'' := fc ->> [sx, joint s', other s'].
have E : s' = s'' \< t.
- rewrite (zigSOE C) {1}(cohE C') /s'' !fsE /= zigE zipPt.
  by rewrite S /sx //= -!joinA joinC joinCA.  
rewrite E in C' *; exists s''; split=>//. 
exists r, g, pfz, b, ha, hr, ha'. 
split=>//; try by [rewrite !getE]; last by split; apply/coh_zigzag.
move: (@cpf r); rewrite Er=>pf2; split=>//=.
- by rewrite !getE.
- by rewrite !getE.
- by rewrite !getE Eg. 
- by rewrite !getzC //= getE O.
move: (pf2 _ _ _ _ _ _) (pf' _ _ _ _ _ _) Ex=>pfx pfy.
by subst m'; rewrite !eqc. 
Qed.

Definition trycollect_pre_act := 
  Action trycollect_safe_coh trycollect_safe_zagzig trycollect_step_safe 
         trycollect_step_guar trycollect_step_frame trycollect_step_total.

Lemma trycollect_step_erase : int_step_erase (trycollect_pre_act).
Proof.
move=>s1 h1 t1 x1 s2 h2 t2 x2 V E H1 H2; move: H1 H2 V E.
case=>r1 [g1][pf1][b1][ha1][hr1][ha1'][Tx1][Eh1 Ha1] /(_ _ pf1) X1 [C1 C1'].
case=>r2 [g2][pf2][b2][ha2][hr2][ha2'][Tx2][Eh2 Ha2] /(_ _ pf2) X2 [C2 C2'].
rewrite (cohE C1) (cohE C2) /= !fc_flatsE Eh1 Eh2.
rewrite -!(joinA _ hr1) -!(joinA _ hr2) => V E.
rewrite (cohE C1') (cohE C2') /= !fc_flatsE. 
case: (fc_exE V E Ha1 Ha2)=>??? E'; subst b2 r2 ha2.
move: (@cpf r1) X1 X2; case Er : (r1 tid)=>[||m' w].
- by move=>_ [->->][->->]; rewrite Eh1 Eh2 -!joinA E'. 
- by move=>_ [->->][->->]; rewrite Eh1 Eh2 -!joinA E'. 
move=>pf1' [S1 [->] Ha1' Eg1 O1 ->] [S2][->] Ha2' Eg2 O2 ->.
split; last by rewrite -!joinA E'; case: Ha1' Ha2' => -> _ [-> _].
by move: (pf1' _ _ _ _ _ _)=>pf1''; subst m'; rewrite !eqc.
Qed.

End TryCollectHelp. 

Definition try_collect_act tid m v := 
  IntAction (@trycollect_step_erase tid m v). 


Section ReadRequest.
Variable tid' : 'I_n. (* the index in the array from which to read *)

Definition readreq_safe s := s \In coh. 

Definition readreq_step s s' (x : req T) := 
  exists (r : {ffun 'I_n -> req T}) (b : bool) ha hr,
    [/\ geth s = lk :-> b \+ ha \+ hr, is_array a r ha,
        x = r tid', s' = s & s \In coh /\ s' \In coh].

Lemma readreq_safe_coh : safe_coh coh readreq_safe.
Proof. by []. Qed. 

Lemma readreq_step_safe : step_safe readreq_safe readreq_step.
Proof. by move=>s s' x [r][b][ha][hr][_ _ _ _][]. Qed.

Lemma readreq_safe_zagzig : safe_zagzig readreq_safe.
Proof. by move=>s t /coh_zigzag C. Qed.

Lemma readreq_step_guar : step_guar guar readreq_step.
Proof. by move=>s s' x [r][b][ha][hr][_ _ _ ->][C]; apply: refl_guar_refl. Qed.

Lemma readreq_step_total : step_total readreq_safe readreq_step.
Proof. 
move=>s C; case: (C)=>_ _ [r][b][ha][hr][_ Eh][Ha].
by exists s, (r tid'), r, b, ha, hr.
Qed.

Lemma readreq_step_frame : step_frame readreq_safe readreq_step.
Proof. 
move=>s s' t x Cx [r][b][ha][hr][Eh Ha ->->{x s'}][C C'].
by exists s; split=>//; exists r, b, ha, hr; rewrite !getE in Eh *. 
Qed.

Definition readreq_pre_act := 
  Action readreq_safe_coh readreq_safe_zagzig readreq_step_safe 
         readreq_step_guar readreq_step_frame readreq_step_total.

Lemma readreq_step_erase : int_step_erase (readreq_pre_act).
Proof.
move=>s1 h1 t1 x1 s2 h2 t2 x2 V E H1 H2; move: H1 H2 V E.
case=>r1 [b1][ha1][hr1][Eh1 [/= Ha1 V1] ->->][C1 _]. 
case=>r2 [b2][ha2][hr2][Eh2 [/= Ha2 V2] ->->][C2 _]. 
move=>V E; split=>//; move: E.
rewrite (cohE C1) (cohE C2) /= !fc_flatsE {}Eh1 {}Eh2 in V *.
rewrite -!joinA -(joinCA ha1) -(joinCA ha2) {}Ha1 {}Ha2 in V *. 
by case/(updi_iinv (size_fgraph _ _) V); move/val_inj/fgraphE=>->. 
Qed.

End ReadRequest. 

Definition read_request_act tid := IntAction (@readreq_step_erase tid). 

End Internals.


Section Externals.
Variable pv : nat.
Notation W := (entangle (Priv pv) (FlatCombine T)).
Notation pv_coh := (Mod.coh (Priv pv)).
Notation pt_coh := (Mod.coh (FlatCombine T)).
Notation coh := (Mod.coh W).
Notation guar := (Mod.internal W).
Notation pv_getso s := (getso (pvg pv) s).
Notation pv_self s := (self (pvg pv) s).
Notation pv_other s := (other (pvg pv) s).
Notation pv_joint s := (joint (pvg pv) s).
Notation fc_getso s := (getso (fcg T) s).
Notation fc_getj s := (getj (fcg T) s).
Notation fc_self s := (self (fcg T) s).
Notation fc_other s := (other (fcg T) s).
Notation fc_joint s := (joint (fcg T) s).
Notation fc_geth s := (geth (fcg T) s).
Notation fc_getg s := (getg (fcg T) s). 

Section CAS.

Definition cas_safe s := s \In coh.

Definition cas_step s s' (x : bool) := 
  let: g := getU (fc_self s \+ fc_other s) \+ sum_up (fc_getg s) in
  exists (r : {ffun 'I_n -> req T}) (b : bool) ha hr,
    [/\ fc_geth s = lk :-> b \+ ha \+ hr, is_array a r ha, 
        if b then s' = s /\ x = false 
        else
          [/\ pv_self s' = pv_self s \+ hr, pv_other s' = pv_other s,
              fc_self s' = (get_tid (fc_self s), up own, getU (fc_self s)),
              fc_joint s' = (lk :-> true \+ ha, fc_getg s),
              fc_other s' = fc_other s & x = true] &
        s \In coh /\ s' \In coh].

Lemma cas_safe_coh : safe_coh coh cas_safe.
Proof. by []. Qed.

Lemma cas_step_safe : step_safe cas_safe cas_step.
Proof. by move=>s s' x [r][b][ha][hr][_ _ _][]. Qed.

Lemma cas_safe_zagzig : safe_zagzig cas_safe.
Proof. by move=>s t /coh_zigzag. Qed.

Lemma cas_step_guar : step_guar guar cas_step.
Proof.
move=>s s' x [r][b][ha][hr][Eh Ha X][C C']; case: b X Eh.
- by case=>-> _ _; apply: refl_guar_refl. 
case=>Ep Op Es' Ef' Of _ Eh; right. 
exists (AcqRelIntGuar (PrivAcqGuar pv) (FCRelGuar T)); split.
- right; exists (Chan (PrivChanCore pv)), (Chan (FCChanCore T)).
  by split; last by [left]; apply/Image_mem. 
split=>//; case: C (C) C' (C') Ha Ep Op Es' Ef' Of Eh => _.
case=>s1 [s2][->{s} Cp1 Cf1] C [_][s1'][s2'][->{s'} Cp1' Cf1'] C'.  
rewrite !(getC Cf1 C) !(getC Cf1' C') !(getC Cp1 C) !(getC Cp1' C') /=.
move=>Ha Ep Op Es' Ef' Of Eh; eexists _, _, _, _; split=>//; exists hr.
split; last by exists ha; case: (fc_cohP Cf1 Eh Ha)=>_ _ _; case; case. 
split=>//; exists (pv_self s1), (pv_other s1); split.
- by rewrite {1}(cohE Cp1) pv_cohJ.  
- by rewrite {1}(cohE Cp1') pv_cohJ Ep Op.  
move: (cohV C) (cohH C). 
rewrite {1 2}(cohE Cp1) {1 2}(cohE Cf1) pv_cohJ /= => V. 
rewrite /validH flatsUn // priv_flatsE fc_flatsE Eh. 
by rewrite -!joinA -(joinC hr) -!(joinCA hr) !joinA; move/validL/validL.
Qed.

Lemma cas_step_total : step_total cas_safe cas_step.
Proof. 
move=>s C; case: (star_cohP C)=>/= E Vh /= C1 C2; move: (validHV Vh) => V.
case: (C2) (cohVSO C2); rewrite !getE; rewrite E /= in Vh V. 
rewrite pv_cohJ in E Vh V C1. 
move=>_ _ [r][b][ha][hr][/=]; rewrite !getE => _ Ef [Haf] Sf Tf Vf.
case; first by case=>Mx ???; subst b hr; exists s, false, r, true, ha, Unit.  
case=>Mx ? inv Vx; subst b.
set p2 := pv ->> [pv_self s \+ hr, tt, pv_other s].
set f2 := fc ->> [(get_tid (fc_self s), up own, getU (fc_self s)), 
                  (lk :-> true \+ ha, fc_getg s),
                  fc_other s].
have V2 : valid (p2 \+ f2). 
- by rewrite /p2 /f2 /valid !fsE /= !fvalidPtUn fvalidPt fdomPt in V *.
have Vh2 : validH (p2 \+ f2).
- move: Vh; rewrite /p2 /f2 !validHUn !priv_flatsE !fc_flatsE V V2 /= Ef. 
  by rewrite -!joinA -!(joinCA (lk :-> _)) (joinC hr) -!joinA !hvalidPtUn. 
exists (p2 \+ f2), true, r, false, ha, hr; rewrite !getE //.
split=>//; first by [rewrite !getV]; do !split=>//. 
exists p2, f2; split=>//; first by split; [move/validHL: Vh2 | eexists _, _].
split; first by move/validHR: Vh2.
- rewrite /f2 /validSO fsE /= zipPt allvalidPt. 
  move: (cohVSO C2); rewrite !getE /valid /= {1 3}/valid /=. 
  by case: (mtx00E Mx)=>->->; rewrite !unitR . 
exists r, true, ha, Unit; rewrite !getE /= unitR /get_mtx; split=>//.
by left; split=>//=; case/mtx00E: Mx=>_ ->; rewrite unitR. 
Qed.

Lemma cas_step_frame : step_frame cas_safe cas_step.
Proof. 
move=>s s' t x Cx [r][b][ha][hr][Eh Ha X][C C'].
rewrite !getE !(getzV (cohSO C)) /= in Eh X.
case: b X Eh C'.
- case=>->-> Eh C'; exists s; split=>//. 
  by exists r, true, ha, hr; rewrite getE. 
case=>/= Ep Op Ef Jf Of ->{x} Eh C'.
case: (star_cohP C') (star_zigSOE C) =>/= E' Vh' C1' C2' Et.
rewrite {1 2}E' in C' Vh'. 
set p1 := pv ->> [_, _, _] in Vh' C1' C'.
set f1 := fc ->> [_, _, _] in Vh' C2' C'.
set pt := pv \-> pv_getso t in Et.
set ft := fc \-> fc_getso t in Et.
set p2 := pv ->> [pv_self s \+ hr, tt, pv_other s].
set f2 := fc ->> [(get_tid (fc_self s), up own, getU (fc_self s)), 
                  (lk :-> true \+ ha, fc_getg s),
                  fc_other s].
have t_mtx : (fc_getso t).1.2 = up nown.
- case: (star_cohP Cx); rewrite !getE /= => X _ _ Cf.
  rewrite {}X in Cx; set j := fc ->> [_, _, _] in Cx.
  have H: fc_geth j = lk :-> false \+ ha \+ hr by rewrite getE. 
  case: (fc_cohP Cf H Ha)=>/= _ _ _; rewrite !getE !(getzV (cohSO C)).
  by case; case=>// /mtx00E [_] /mtx00E []. 
have [pE fE] : p2 \< pt = p1 /\ f2 \< ft = f1.
- rewrite /p1 /p2 /f1 /f2 /pt /ft !fsE !zigE !zipPt pv_cohJ Ep Op Ef Of Jf. 
  split; first by rewrite joinCA joinA.
  rewrite pcmE /= pcmE /= t_mtx unitL; apply/stE; split=>//=.
  by congr (fc \-> (_, _, _)); rewrite joinC. 
have pfE : (p2 \+ f2) \< t = (p2 \< pt) \+ (f2 \< ft).
- rewrite Et; apply: zig_plus; rewrite ?pE ?fE;
  by case/and3P: (cohV C1'); case/and3P: (cohV C2'). 
have Vso : validSO ((p2 \+ f2) \< t).
- by rewrite pfE pE fE; apply: cohSO C'. 
have V2 : valid (p2 \+ f2).
- by move: (cohV C'); rewrite -pE -fE -pfE; case/valid_zig. 
exists (p2 \+ f2); rewrite pfE pE fE; split=>//.
exists r, false, ha, hr; rewrite !getE // !getzV ?(cohSO C) // !getV //=.
by do 2!split=>//; apply/coh_zigzag; rewrite pfE pE fE. 
Qed.

Definition cas_pre_act := 
  Action cas_safe_coh cas_safe_zagzig cas_step_safe 
         cas_step_guar cas_step_frame cas_step_total.

Lemma cas_step_erase : int_step_erase (cas_pre_act).
Proof.
move=>s1 h1 t1 x1 s2 h2 t2 x2 V E H1 H2; move: H1 H2 V E.
case=>r1 [b1][ha1][hr1][Eh1 Ha1 X1][C1 C1'].
case=>r2 [b2][ha2][hr2][Eh2 Ha2 X2][C2 C2'].
move: (cohV C1) (cohV C2) (cohV C1') (cohV C2'). 
rewrite (star_cohE C1) (star_cohE C2) /= => V1 V2.
rewrite (star_cohE C1') (star_cohE C2') /= => V1' V2'.
rewrite !flatsUn // !(pv_cohJ,priv_flatsE,fc_flatsE) Eh1 Eh2.
rewrite -!joinA -!(joinCA (lk :-> _)) => V E.
case/(hcancelV V): E=>? {V} V E; subst b2.
case: b1 X1 X2 Eh1 Eh2.
- by case=>->-> [->->] ->->; rewrite -!joinA -!(joinCA (lk :-> _)) E.
case=>Ep1 Op1 Ef1 Jf1 Of1 ->{x1} [Ep2] Op2 Ef2 Jf2 Of2 ->{x2} _ _.
rewrite Jf1 Jf2 /= Ep1 Ep2 Op1 Op2 -!joinA -!(joinCA (lk :-> _)).
by rewrite 2!(joinCA hr1) 2!(joinCA hr2) E.
Qed.

End CAS. 

Definition cas_act := IntAction cas_step_erase. 


Section Unlock.

Definition unlock_safe s := 
  let: g := getU (fc_self s \+ fc_other s) \+ sum_up (fc_getg s) in
  exists (hr hr' : heap), 
   [/\ get_mtx (fc_self s) = up own, 
       pv_self s = hr \+ hr', I g hr & s \In coh]. 

Definition unlock_step s s' (x : unit) := 
  let: g := getU (fc_self s \+ fc_other s) \+ sum_up (fc_getg s) in
  exists (r : {ffun 'I_n -> req T}) ha hr hr',
    [/\ get_mtx (fc_self s) = up own /\ pv_self s = hr \+ hr' /\ I g hr, 
        fc_geth s = lk :-> true \+ ha /\ is_array a r ha, 
        pv_self s' = hr' /\ pv_other s' = pv_other s, 
        fc_self s' = (get_tid (fc_self s), up nown, getU (fc_self s)) /\
        fc_joint s' = (lk :-> false \+ ha \+ hr, fc_getg s),
        fc_other s' = fc_other s & 
        s \In coh /\ s' \In coh].

Lemma unlock_safe_coh : safe_coh coh unlock_safe.
Proof. by move=>s [hr][hr'][]. Qed.

Lemma unlock_step_safe : step_safe unlock_safe unlock_step.
Proof. 
move=>s s' [][r][ha][hr][hr'][[Mx [Es inv]] _ _ _ _][C _].
by exists hr, hr'. 
Qed.

Lemma unlock_safe_zagzig : safe_zagzig unlock_safe.
Proof. 
move=>s t [hr][hr'][Mx Ep inv] /coh_zigzag C.
rewrite !getE !getzV ?(cohSO C) // in Mx Ep inv.
suff My : (fc_getso t).1.2 = up nown.
- exists hr, (hr' \+ pv_getso t).
  rewrite !getE !getzV ?(cohSO C) // /get_mtx /= My unitR Ep.
  by rewrite -!joinA in inv *.
rewrite /get_mtx /= in Mx *; move/proj2: (star_cohVSO C). 
rewrite !getE !getzV ?(cohSO C) //=.
case/andP; case/andP; rewrite /= Mx =>_ /validL V _.
by case: (fc_getso t).1.2 V=>//; case.
Qed.

Lemma unlock_step_guar : step_guar guar unlock_step.
Proof.
move=>s s' [][r][ha][hr][hr'][[Mx][Ep] inv][Eh Ha][Ep' Op'][Ef Jf Of][C C'].
set ag := (FCAcqGuar T); set rg := (PrivRelGuar pv).
right; exists (DoIntGuar (starC _ _) (AcqRelIntGuar ag rg)); split.
- right; exists (Chan (PrivChanCore pv)), (Chan (FCChanCore T)). 
  by split; last by [right]; apply/Image_mem.
split; last by rewrite starC.
rewrite (star_cohE C) (star_cohE C') !(joinC (pv ->> [_, _, _])) /=.
eexists _, _, _, _; split=>//; exists hr; split.
- by exists ha; rewrite !getE; case: (star_cohP C); case (star_cohP C'). 
split; last by case: (star_cohP C).
by exists (pv_self s'), (pv_other s); rewrite Ep Ep' joinC !pv_cohJ Op'. 
Qed.

Lemma unlock_step_total : step_total unlock_safe unlock_step.
Proof. 
move=>s [hr][hr'][Ms Ep inv] C; move: (cohH C)=>Vh; move: (cohV C)=>V.
case: (star_cohP C)=>/= E _ Cp Cf.
rewrite E in V Vh; rewrite pv_cohJ in E Cp V Vh C.
set p1 := pv ->> [_, _, _] in E Cp C.
set f1 := fc ->> [_, _, _] in E Cf C.
case: Cf (Cf)=>_ _ [r][b][ha][hr2][_] Eh [Ha Sg] Tx Vx X Cf.
rewrite !getE /get_mtx in Eh Sg Tx Vx X Ms. 
have Mo : (fc_other s).1.2 = up nown.
- move: (cohVSO Cf); rewrite !getE pcmE /= pcmE /=.
  rewrite Ms; case/andP; case/andP.
  by case: ((fc_other s).1.2)=>//; case.
case: X; last by case; rewrite /get_mtx /= Ms Mo.
case=>// _ ??; subst b hr2; rewrite unitR in Eh. 
set p2 := pv ->> [hr', tt, pv_other s].
set f2 := fc ->> [(get_tid (fc_self s), up nown, getU (fc_self s)), 
                  (lk :-> false \+ ha \+ hr, fc_getg s), fc_other s]. 
have V2 : valid (p2 \+ f2). 
- by rewrite /p2 /f2 /valid !fsE /= !fvalidPtUn fvalidPt fdomPt in V *.
have Vh2 : validH (p2 \+ f2).
- move: Vh; rewrite /p2 /f2 !validHUn !priv_flatsE !fc_flatsE V V2 /= Ep Eh. 
  by rewrite -!joinA -!(joinCA (lk :-> _)) (joinC hr) -!joinA !hvalidPtUn. 
exists (p2 \+ f2), tt, r, ha, hr, hr'; rewrite !getE // !getV //; do !split=>//.
exists p2, f2; split=>//; first by split; [move/validHL: Vh2 | eexists _, _].
split; first by move/validHR: Vh2.
- rewrite /f2 /validSO fsE /= zipPt allvalidPt. 
  by move: (cohVSO Cf); rewrite !getE /valid /= {1 3}/valid /= Ms Mo.  
exists r, false, ha, hr; rewrite !getE; split=>//; right; split=>//=. 
by rewrite /get_mtx /= Mo.
Qed.

Lemma unlock_step_frame : step_frame unlock_safe unlock_step.
Proof. 
move=>s s' t [] Cx [r][ha][hr][hr'][[Ms][Ep inv]][Eh Ha][Ep' Op][Ef J Of][C C'].
move: (cohSO C)=>Vso; rewrite !getE !getzV // in Ms Ep inv Eh Op Ef J Of. 
case: (star_cohP C') (star_zigSOE C) =>/= E' Vh' C1' C2' Et.
rewrite {1 2}E' in C' Vh'; subst hr'; rewrite /get_mtx in Ms.
set p1 := pv ->> [_, _, _] in Vh' C1' C'.
set f1 := fc ->> [_, _, _] in Vh' C2' C'.
set pt := pv \-> pv_getso t in Et.
set ft := fc \-> fc_getso t in Et.
have [t_mtx [hr' [Eps Eps']]] : 
  (fc_getso t).1.2 = up nown /\ exists hr', 
  pv_self s = hr \+ hr' /\ pv_self s' = pv_getso t \+ hr'.
- case: Cx=>hr2 [hr2'][Ms' Ep' inv'] Cx.
  rewrite /get_mtx !getE !getzV //= -!joinA in Ms Ms' Ep' inv inv'.
  split; first by move: Ms; rewrite Ms'; case: (fc_getso t).1.2=>//; case.
  have Vx : valid (pv_self s\+ pv_getso t).
  - by move/proj1/validL: (star_cohVSO C); rewrite getzV //.
  rewrite Ep' -!joinA in Vx Ep *. 
  move: (fc_precise Vx Ep inv' inv)=>?; subst hr2.
  by rewrite -(joinfK Vx Ep) (joinC hr2'); exists hr2'.  
set p2 := pv ->> [hr', tt, pv_other s].
set f2 := fc ->> [(get_tid (fc_self s), up nown, getU (fc_self s)), 
                  (lk :-> false \+ ha \+ hr, fc_getg s),
                  fc_other s].
have [pE fE] : p2 \< pt = p1 /\ f2 \< ft = f1.
- rewrite /p1 /p2 /f1 /f2 /pt /ft !fsE !zigE !zipPt pv_cohJ -Eps' Op Ef Of J.
  split=>//; rewrite pcmE /= pcmE /= t_mtx unitL; apply/stE; split=>//=.
  by congr (fc \-> (_, _, _)); rewrite joinC. 
have pfE : (p2 \+ f2) \< t = (p2 \< pt) \+ (f2 \< ft).
- rewrite Et; apply: zig_plus; rewrite ?pE ?fE;
  by case/and3P: (cohV C1'); case/and3P: (cohV C2'). 
have Vso2 : validSO ((p2 \+ f2) \< t).
- by rewrite pfE pE fE; apply: cohSO C'. 
have V2 : valid (p2 \+ f2).
- by move: (cohV C'); rewrite -pE -fE -pfE; case/valid_zig. 
exists (p2 \+ f2); rewrite pfE pE fE; split=>//.
exists r, ha, hr, hr'; rewrite !getE // !getzV // !getV //=.
split=>//; last by split; apply/coh_zigzag=>//; rewrite pfE pE fE. 
move: Ms; rewrite /get_mtx /= t_mtx unitR; do !split=>//.
by rewrite -!joinA in inv *. 
Qed.

Definition unlock_pre_act := 
  Action unlock_safe_coh unlock_safe_zagzig unlock_step_safe 
         unlock_step_guar unlock_step_frame unlock_step_total.

Lemma unlock_step_erase : int_step_erase (unlock_pre_act).
Proof.
move=>s1 h1 t1 [] s2 h2 t2 [] V E H1 H2; move: H1 H2 V E.
case=>r1 [ha1][hr1][hr1'][[_]][Ep1 _][Eh1 _][Ep1' Op1][_ J1 Of1][C1 C1'].
case=>r2 [ha2][hr2][hr2'][[_]][Ep2 _][Eh2 _][Ep2' Op2][_ J2 Of2][C2 C2'].
move: (cohV C1) (cohV C2) (cohV C1') (cohV C2'). 
rewrite (star_cohE C1) (star_cohE C2) /= => V1 V2.
rewrite (star_cohE C1') (star_cohE C2') /= => V1' V2'.
rewrite !flatsUn // !(pv_cohJ,priv_flatsE,fc_flatsE) Eh1 Eh2 Ep1 Ep2.
rewrite -!joinA -!(joinCA (lk :-> _)) => V E.
case/(hcancelV V): E=>_ {V} V E. 
rewrite J1 J2 /= Ep1' Ep2' Op1 Op2 -!joinA -!(joinCA (lk :-> _)).
by rewrite -3!(joinCA hr1) -3!(joinCA hr2) E. 
Qed.

End Unlock.

Definition unlock_act := IntAction unlock_step_erase. 

End Externals.

End Actions. 


Module Exports.
Notation reqhelp_act := reqhelp_act.
Notation dohelp_act := dohelp_act.
Notation try_collect_act := try_collect_act.
Notation read_request_act := read_request_act.
Notation cas_act := cas_act.
Notation unlock_act := unlock_act.

Notation cast := cast.
Notation eqc := eqc.
Notation cpf := cpf. 
End Exports.

End Actions. 

Export Actions.Exports.


(*************)
(* Stability *)
(*************)

Section Stability.
Variable T : flat_combineT.
Notation a := (fc_orig T).
Notation lk := (fc_lk T).
Notation n := (@fc_n T).
Notation fc := (@fc_label T). 
Notation U := (@fc_pcm T).
Notation I := (@fc_inv T).
Notation V := (fc_V T).
Notation J := (fc_J T).

Section Internals.
Notation coh := (fc_coh T).
Notation getso s := (getso (fcg T) s).
Notation self s := (self (fcg T) s).
Notation other s := (other (fcg T) s).
Notation joint s := (joint (fcg T) s).
Notation geth s := (geth (fcg T) s).
Notation getg s := (getg (fcg T) s). 
Notation W := (FlatCombine T).

Definition specr (P : fc_V T -> fc_V T -> 
                       {ffun 'I_n -> req T} -> {ffun 'I_n -> U} -> Prop) s :=
  exists r (b : bool) ha hr, 
    [/\ geth s = lk :-> b \+ ha \+ hr, is_array a r ha,
        P (self s) (other s) r (getg s) & 
        s \In coh].

(* lemma for removing state futsing from stability proofs *)
(* so that we can focus only on important ghost components *)

Definition specrPQ (P Q : _ -> _ -> _ -> _ -> Prop) := 
 forall gL gE (r : {ffun 'I_n -> req T}) (gs : {ffun 'I_n -> U}), 
   valid (gL \+ gE) ->
   [/\ (* identity transition *)
       P gL gE r gs -> Q gL gE r gs, 
       (* request-help transition *)
       forall tid m v, 
         P gL gE r gs -> nat_of_ord tid \in dom (get_tid gE) -> 
         r tid = NoReq T ->
         Q gL gE [ffun z => if z == tid then @Req _ m v else r z] gs,
       (* do-help transition *)
         forall tid m v w g', 
           P gL gE r gs -> get_mtx gE = up own -> 
           r tid = @Req _ m v -> gs tid = Unit -> 
           valid (getU (gL \+ gE) \+ sum_up gs \+ g') ->
           fc_R v w (getU (gL \+ gE) \+ sum_up gs) g' ->
           Q gL gE [ffun z => if z == tid then Ans w else r z] 
                   [ffun z => if z == tid then g' else gs z],
       (* collect-help transition *)
         forall tid m w, 
           P gL gE r gs -> nat_of_ord tid \in dom (get_tid gE) ->
           r tid = @Ans _ m w -> 
           Q gL (get_tid gE, get_mtx gE, getU gE \+ gs tid) 
             [ffun z => if z == tid then NoReq T else r z]
             [ffun z => if z == tid then Unit else gs z], 
       (* cas transition *)
         P gL gE r gs -> get_mtx gL = up nown -> get_mtx gE = up nown ->
         Q gL (get_tid gE, up own, getU gE) r gs &
       (* unlock transition *)
         P gL gE r gs -> get_mtx gE = up own -> 
         Q gL (get_tid gE, up nown, getU gE) r gs].

Definition stable1 P Q := 
  forall s1 s2, mrely W s1 s2 -> specr P s1 -> specr Q s2.

Definition stable P := 
  forall s1 s2, menv_steps W s1 s2 -> specr P s1 -> specr P s2.

Lemma stablePQ P Q : specrPQ P Q -> stable1 P Q. 
Proof.
move=>H s1 s2 M; case: (mrely_coh M)=>_ C2. 
case=>r [b][ha][hr][Eh Ha Hp] C1; move: (fc_cohJ C1); rewrite Eh => V.
case: (H (self s1) (other s1) r (getg s1) (cohVSO C1))=>
  Tid Treq Tdo Tcollect Tcas Tunlock.
case: M.
- case. 
  - case; move/transp_inj=><-; exists r, b, ha, hr.
    by split=>//; apply: Tid Hp.
  case=>g []; case; [|case].
  - case=>t1 [m1][v1] ->{g}.  
    case=>/= r1 [b1][ha1][hr1][ha1'][Tx1][Eh1][Ha1 Er1] S1 [Jh1 Ha1' O1] _.  
    rewrite Eh in Eh1.
    case: (fc_exE V Eh1 Ha Ha1)=>????; subst b1 r1 ha1 hr1.
    exists [ffun z => if z == t1 then Req v1 else r z], b, ha1', hr. 
    by rewrite O1 S1 Jh1; split=>//; apply: Treq Hp Tx1 Er1. 
  - case=>t1 [m1][v1][w1] ->{g}.
    case=>/= r1 [ha1][ha1'][g'][Tx1][Eh1][Ha1 Er1] S1 [Jh1 Ha1'] O1 [Vx1][R1] _. 
    rewrite -[_ \+ _]unitR in Eh1; rewrite Eh in Eh1.
    case: (fc_exE V Eh1 Ha Ha1)=>????; subst b r1 ha1 hr.
    exists [ffun z => if z == t1 then Ans w1 else r z], true, ha1', Unit. 
    rewrite Jh1 /= unitR O1 S1; split=>//.   
    suff G: getg s1 t1 = Unit.
    - rewrite -!joinA !(joinCA _ (self s1).2) !joinA in Vx1 R1.
      by apply: Tdo Hp Tx1 Er1 G Vx1 R1.
    case: C1=>_ _ [r2][b2][ha2][hr2][_ Eh2 [Ha2 Sx]] _ _ _. 
    rewrite Eh in Eh2; case: (fc_exE V Eh2 Ha Ha2)=>????; subst b2 r2 ha2 hr2.
    by move: (Sx t1); rewrite Er1.
  case=>t1 [m1][v1] ->{g}.
  case=>/= r1 [b1][ha1][hr1][ha1'][Tx1][Eh1][Ha1 Er1] S1 [Jh1][Ha1' Eg1] O1 _. 
  rewrite Eh in Eh1.
  case: (fc_exE V Eh1 Ha Ha1)=>????; subst b1 r1 ha1 hr1.
  exists [ffun z => if z == t1 then NoReq T else r z], b, ha1', hr.
  by rewrite Jh1 O1 S1 Eg1; split=>//; apply: Tcollect Hp Tx1 Er1.
case=>ch [x][/mem_chan][_][-> <-] /=; case. 
- case=>ha1 [/=][Tx1] Eh1 S1 Eh1' O1 inv1 _. 
  case: (fc_cohPX C1 Eh1)=>r1 Ha1.
  rewrite -[_ \+ _]unitR Eh in Eh1. 
  case: (fc_exE V Eh1 Ha Ha1)=>? _ ??; subst b ha1 hr. 
  exists r, false, ha, x; rewrite Eh1' O1 S1.
  by split=>//; apply: Tunlock Hp Tx1.
case=>ha1 [/=] Eh1 inv1 O1 Eh1' S1 _.
rewrite -joinA joinCA joinA in inv1.
case: (fc_cohPY C1 Eh1 inv1)=>r1 Ha1.
rewrite Eh in Eh1; case: (fc_exE V Eh1 Ha Ha1)=>????; subst b r1 ha1 x.
exists r, true, ha, Unit; rewrite Eh1' S1 O1 unitR; split=>//.
case: C1=>_ _ [r2][b2][ha2][hr2][_ Eh2][Ha2] Sx _ _.
rewrite Eh in Eh2; case: (fc_exE V Eh2 Ha Ha2)=>????; subst b2 r2 ha2 hr2.
by case; case=>// /mtx00E [H1 H2] _ _; apply: Tcas Hp H1 H2. 
Qed.

Lemma stableP P : specrPQ P P -> stable P.
Proof.
move=>X s1 s2 [n M].
elim: n s1 s2 M=>[|n IH] s1 s2 /= M S1; first by case: M => <-.
by case: M S1 => s [/(stablePQ X) Y] M /Y; apply: IH M. 
Qed.

Lemma specr_and P Q s : 
        specr P s -> specr Q s -> 
        specr (fun gL gE r gs => P gL gE r gs /\ Q gL gE r gs) s.
Proof.
case=>r1 [b1][ha1][hr1][Eh1 Ha1 Hp] C [r2][b2][ha2][hr2][Eh2 Ha2 Hq] _.
move: (fc_cohJ C); rewrite Eh1 in Eh2 *; move=>V.
case: (fc_exE V Eh2 Ha1 Ha2)=>????; subst b2 r2 ha2 hr2.
by exists r1, b1, ha1, hr1. 
Qed.

Lemma stable_and P Q : 
        stable P -> stable Q -> 
        stable (fun gL gE r gs => P gL gE r gs /\ Q gL gE r gs).
Proof.
move=>H1 H2 s1 s2 M [r][b][ha][hr][H3 H4 [Hp1 Hq1]] C1.
have [Hpx Hqx] : specr P s1 /\ specr Q s1 by split; exists r, b, ha, hr. 
case: (H1 _ _ M Hpx)=>rp [bp][hap][hrp][P3 P4 P5] C2.
case: (H2 _ _ M Hqx)=>rq [bq][haq][hrq][Q3 Q4 Q5] _.
have V : valid (geth s2) by move: (fc_cohJ C2).  
rewrite P3 in Q3 V; case: (fc_exE V Q3 P4 Q4)=>????; subst bq rq haq hrq.
by exists rp, bp, hap, hrp. 
Qed.

(* A number of stability definitions and properties *)

Definition requested tid m v g gL gE r gs := 
  @Actions.requested T tid m v r g gs (getU (gL \+ gE)). 

Section Lemmas.
Variables (tid : 'I_n) (m : fc_code T) (v : fc_inT m) (w : fc_outT m).

(*
Notational convention: the *_req stability lemmas talk about threads
that *request* help. The *_help stability lemmas talk about the
helping thread. If the lemma applies to both kinds of threads, it has neither
_req not _help postfix. All the lemmas have S postfix, to indicate stability.
*)

(* if I have the lock, nobody can take it *)
Lemma lockS : stable (fun gL _ _ _ => get_mtx gL = up own). 
Proof. by apply: stableP. Qed.

(* only the thread itself can make a request on its own behalf *)
Lemma noreq_reqS : 
        stable (fun gL gE r gs => 
          nat_of_ord tid \in dom (get_tid gL) /\ r tid = NoReq T).
Proof.
apply: stableP=>gL gE r gs V; split=>//.
- move=>t m1 v1 [Ttid Etid] Tt Et; split=>//.
  rewrite ffunE; case: eqP Tt=>// <-; rewrite /get_tid in Ttid *.
  by case/andP/proj1/andP: V; case: validUn=>// _ _ /(_ _ Ttid) /negbTE ->. 
- move=>t m1 v1 w1 g' [Ttid Etid] _ Et _ _ _; split=>//.
  by rewrite ffunE; case: eqP Et Etid=>// ->->.
move=>t m1 w1 [Ttid Etid] Mt Et; split=>//.
by rewrite ffunE; case: eqP.
Qed.

Lemma requested_reqS g : 
        stable (fun gL gE r gs => 
          nat_of_ord tid \in dom (get_tid gL) /\ requested tid v g gL gE r gs). 
Proof.
apply: stableP=>gL gE r gs V; split=>//.
- move=>t m1 v1 [Ttid S] Tt Et; split=>//.
  split; rewrite ffunE=>m' w'; last by case: eqP (proj2 S m' w').
  case: eqP Tt=>[<-|_] Tt; last by apply: (proj1 S).
  by case/andP/proj1/andP: V; case: validUn=>// _ _ /(_ _ Ttid); rewrite Tt.
- move=>t m1 v1 w1 g' [Tx S] Mx Et Gt Vx Rx; split=>//.
  split; rewrite !ffunE=>m' w'; case: eqP Et Gt=>[<-|_] Et Gt //.
  - case/(proj1 S)=>pf [Cx [g'' Eg]].
    exists pf; split=>//; exists (g'' \+ g'). 
    by rewrite joinA Eg -joinA (sumIR _ t) Gt unitL.
  - case=>?; subst m' => /(@inj_pair2 _ _ _ _ _) <-.
    case: (proj1 S m1 v1 Et)=>? []; subst m1; rewrite eqc=>? Eg; subst v1.
    exists (erefl _), (getU (gL \+ gE) \+ sum_up gs); split=>//.
    by exists g'; rewrite -joinA (sumIR _ tid) Gt unitL.
  case/(proj2 S)=>pf [g1][Eg1 [g2 Eg2] Rg]; exists pf, g1; split=>//. 
  by exists (g2 \+ g'); rewrite joinA Eg2 -joinA (sumIR _ t) Gt unitL.
move=>t m1 w1 [Ttid S] Tt Et; split=>//.
split; rewrite !ffunE => m' v'; case: eqP=>// _.
- case/(proj1 S)=>pf [Cx][g2 Eg]; exists pf; split=>//.
  by exists g2; rewrite Eg (sum0R _ t) -!joinA (joinC (gs t)). 
case/(proj2 S)=>pf [g1][Eg1 [g2 Eg2] R]; exists pf, g1; split=>//.
by exists g2;rewrite Eg2 (sum0R _ t) -!joinA (joinC (gs t)). 
Qed.

(* helping requests can only be served by the helper thread *)
Lemma req_helpS t : 
        stable (fun gL _ r _ => get_mtx gL = up own /\ r t = Req v). 
Proof.
apply: stableP=>gL gE r gs V; split=>//.
- move=>t1 m1 v1 [Mx Etid] Tx Et; split=>//.
  by rewrite ffunE; case: eqP Etid Et=>// ->->.
- move=>t1 m1 v1 w1 g' [Mx1 _] Mx2 _ _ _ _.
  rewrite /get_mtx in Mx1 Mx2.
  by case/andP: V; case/andP; rewrite /= Mx1 Mx2. 
move=>t1 m1 w1 [Mx Etid] Tx Et; split=>//. 
by rewrite ffunE; case: eqP Etid Et=>// ->->.
Qed.

(* when the helper holds the lock, the threads can shuffle the contributions *)
(* but they can't change the sum total *)
Lemma sumup_helpS g : 
        stable (fun gL gE _ gs => 
          get_mtx gL = up own /\ g = getU (gL \+ gE) \+ sum_up gs). 
Proof.
apply: stableP=>gL gE r gs V; split=>//.
- rewrite /get_mtx => t m1 v1 w1 g' [Mx1 _] Mx2.
  by case/andP: V; case/andP; rewrite /= Mx1 Mx2.
move=>t m1 w1 [Mx Eg] Tt Et; split=>//=.
rewrite Eg -!joinA; congr (_ \+ (_ \+ _)).
by rewrite joinC -(sum0R _ t).
Qed.

(* the above is quite commonly used, and for larger concurroids as well *)
(* so let's give it in a more user-friendly form *)
Lemma sumupCE s1 s2 : 
        menv_steps W s1 s2 -> 
        get_mtx (self s1) = up own -> 
        getU (self s1 \+ other s1) \+ sum_up (getg s1) = 
        getU (self s2 \+ other s2) \+ sum_up (getg s2).
Proof.
move=>M1 Mx; case: (menvs_coh M1)=>C1 C2.
set g := getU (self s1 \+ other s1) \+ sum_up (getg s1).
case/(@sumup_helpS g): M1; last by move=>r [b][ha][hr'][_ _][_] <-. 
by case: (C1)=>_ _ [r][b][ha][hr][_ Eh][Ha]; exists r, b, ha, hr. 
Qed.

Lemma sumupCE_ext V s1 s2 : 
        injects W V ->
        menv_steps V s1 s2 -> 
        get_mtx (self s1) = up own -> 
        getU (self s1 \+ other s1) \+ sum_up (getg s1) = 
        getU (self s2 \+ other s2) \+ sum_up (getg s2).
Proof.
move=>inj M1 Mx; case: (menvs_coh M1)=>C1 C2.
rewrite (Injection.Exports.cohE inj) /= in C1 C2.
case: (C1)=>_ [x1][y1][?] Cx1 _; subst s1.
case: (C2)=>_ [x2][y2][?] Cx2 _; subst s2.
rewrite !(getC Cx1 C1) !(getC Cx2 C2) /= in Mx *.
case/(menvs_split inj Cx1 Cx2): M1=>M1x _. 
by apply: sumupCE M1x Mx.
Qed.

(* the sum of ghosts only grows *)
(* this is like the definition of [- <<= -] in histories, but done for PCMS *)
(* this should be defined in the file for general PCMs *)
Lemma gmonoS g : 
        stable (fun gL gE r gs => [g <<= getU (gL \+ gE) \+ sum_up gs]).
Proof.
apply: stableP=>gL gE r gs V; split=>//.
- move=>t m1 v1 w1 g' [g2 Eg] Mx Et Gt Vx Rx.
  by exists (g2 \+ g'); rewrite !joinA Eg -!joinA (sumIR _ t) Gt unitL.
move=>t m1 w1 [g2 Eg] Tt Et; exists g2.
by rewrite Eg -!joinA (sum0R _ t) (joinC (gs t)). 
Qed.

End Lemmas.

(**********************)
(* req_help stability *)
(**********************)

(* I'm going with largefoot self spec cause it's shorter than smallfoot spec *)
(* Do I need to say that g1 is bigger than the initial g \+ sum_up gs? *)
(* Yes, I do *)
Program Definition req_help tid m (v : fc_inT m) : {s g}, STsep [W] 
    (specr (fun gL gE r gs => 
       [/\ s = gL, nat_of_ord tid \in dom (get_tid gL), r tid = NoReq T & 
           [g <<= getU (gL \+ gE) \+ sum_up gs]]),
     fun (_ : unit) => 
       specr (fun gL gE r gs => 
        [/\ s = gL, [g <<= getU (gL \+ gE) \+ sum_up gs] &
            requested tid v g gL gE r gs])) := 
  Do (act (reqhelp_act tid v)).
Next Obligation.
apply: gh=>i [_ g][r][b][ha][hr][Eh Ha][->] /= Tx Etid Eg C _. 
apply: val_act=>i1 M1; case: (menvs_coh M1)=>_ C1; move: (fc_cohJ C1)=>V.
case/(stable_and (@noreq_reqS tid) (@gmonoS g)): (M1).
- by exists r, b, ha, hr.
move=>/= r1 [b1][ha1][hr1][Eh1 Ha1][[Ttid1 Etid1]] Eg1 _.
split; first by exists r1, b1, ha1, hr1. 
case=>i2 i3 X M23; case: (menvs_coh M23)=>C2 C3 {Eh Ha Tx Etid Eg}.
case: X=>r' [b'][ha'][hr'][ha2][_][Eh'][Ha'] _ S2 [Jh2 Ha2] O2 _.
rewrite Eh1 in V Eh'.
case: (fc_exE V Eh' Ha1 Ha')=>???? {Eh' Ha'}; subst b' r' ha' hr'.
case/(stable_and (@requested_reqS tid m v g) (@gmonoS g)): (M23).
- exists [ffun z => if z == tid then Req v else r1 z], b1, ha2, hr1. 
  rewrite Jh2 S2 O2; do !split=>//; move=>m' v'; rewrite ffunE eq_refl //. 
  case=>?; subst m'=> /(@inj_pair2 _ _ _ _ _) ->.
  by  exists (erefl _); rewrite eqc. 
move=>r3 [b3][ha3][hr3][Eh3 Ha3][[Tx3 Rx3]] Eg3 _.
by exists r3, b3, ha3, hr3; rewrite (menvs_loc M1) -S2 (menvs_loc M23).
Qed. 

(*********************)
(* do_help stability *)
(*********************)

(* I need to specify that the helping doesn't *)
(* remove the request condition of the helper thread *)
(* for other threads, that follows by stability *)
(* but for self thread, it has to be specified directly *)

Section DoHelp.
Variables (tid : 'I_n). (* the thread id of the thread *to be helped *)
Variables (m : fc_code T) (w : fc_outT m). 
(* w is the result that the helper wants to write into the helping array *)
(* m is the method code; it's an argument because it determines the type *)
(* of w *)

Program Definition do_help : {s g g1 g' mytid m' v'}, STsep [W] 
  (specr (fun gL gE r gs => 
       [/\ s = gL /\ nat_of_ord mytid \in dom (get_tid gL) /\ 
           get_mtx gL = up own, g1 = getU (gL \+ gE) \+ sum_up gs,
           @requested mytid m' v' g gL gE r gs &
           exists v, [/\ r tid = Req v, valid (g1 \+ g') & fc_R v w g1 g']]), 
   fun (_ : unit) => 
     specr (fun gL gE r gs => 
       [/\ s = gL, 
           @requested mytid m' v' g gL gE r gs &
           g1 \+ g' = getU (gL \+ gE) \+ sum_up gs])) :=
  Do (act (dohelp_act tid w)).
Next Obligation.
apply: gh=>i [s][g][g1][g'][mytid][m' v'][r][b][ha][hr] /= [Eh Ha][][->][Tx]. 
move=>Mx Eg Rqx [v][Er Vx Rx] C _.
apply: val_act=>i1 M1; case: (menvs_coh M1)=>_ C1; move: (fc_cohJ C1) => V.
case: (stable_and (@req_helpS m v tid) 
      (stable_and (@requested_reqS mytid m' v' g) (@sumup_helpS g1)) M1).
- by exists r, b, ha, hr.  
move=>r1 [b1][ha1][hr1][Eh1 Ha1][[Tx1 Er1]][[Mtx1]] Mrx1 [_ Eg1] _; split.
- by exists r1, b1, ha1, hr1, g', v; rewrite -Eg1.  
have G1 : getg i1 tid = Unit.
- case: (C1)=>_ _ [rx][bx][hax][hrx][_ Ehx [Hax] Sx _] _ _.
  rewrite Eh1 in Ehx V.
  case: (fc_exE V Ehx Ha1 Hax)=>????; subst bx rx hax hrx.
  by move: (Sx tid); rewrite Er1. 
case=>i2 i3 X M23; case: X=>/= v1' [r1'][ha1'][ha2][g1'][_][Eh1'][Ha1' X] S2.
rewrite -[_ \+ _]unitR Eh1 in V Eh1'.
case: (fc_exE V Eh1' Ha1 Ha1')=>???? {Ha1' Eh1'}; subst b1 r1' ha1' hr1.
case=>Jh2 Ha2 O2 [Vx2][Rx2][_ C2]; rewrite -Eg1 in Vx2 Rx2.
rewrite Er1 in X; case: X; move/(@inj_pair2 _ _ _ _ _)=>?; subst v1'.
move: (fc_uniq Vx Vx2 Rx Rx2)=>? {Vx2 Rx2}; subst g1'.
suff : specr (fun gL gE r gs => self i1 = gL /\ 
  (nat_of_ord mytid \in dom (get_tid gL) /\ requested mytid v' g gL gE r gs) /\
  (get_mtx gL = up own /\ g1 \+ g' = getU (gL \+ gE) \+ sum_up gs)) i2. 
- move/(stable_and _ (stable_and (@requested_reqS mytid m' v' g) _)).
  move/(_ _ (@sumup_helpS (g1 \+ g')) _ M23).
  case; first by apply: stableP.
  move=>r3 [b3][ha3][hr3][Eh3 Ha3][S3][[Mtx3 Rx3]][Tx3] Eg3 C3.
  by exists r3, b3, ha3, hr3; rewrite (menvs_loc M1).
exists [ffun z => if z == tid then Ans w else r1 z], true, ha2, Unit.  
rewrite Jh2 /= unitR S2 O2 Eg1 -!joinA (sumIR _ tid).
do 4!split=>//; last by rewrite G1 unitL.
split=>m2 v2; rewrite !ffunE.
- case: eqP=>// _; case/(proj1 Mrx1)=>pf [Cx [g2 Eg2]]. 
  exists pf; split=>//; exists (g2 \+ g'). 
  by rewrite joinA Eg2 -joinA (sumIR _ tid) G1 unitL.
case: eqP=>[?[?]|_]. 
- subst mytid m2; move/(@inj_pair2 _ _ _ _ _)=>?; subst v2.
  case: (proj1 Mrx1 _ _ Er1)=>pf []; subst m'; rewrite !eqc=>? Eg2; subst v'.
  exists (erefl _), g1; rewrite eqc {1 2}Eg1; split=>//; exists g'. 
  by rewrite -!joinA (sumIR _ tid) G1 unitL. 
case/(proj2 Mrx1)=>pf [g2][Eg2 [g3 F3] Rx2]; subst m2.
exists (erefl _), g2; rewrite eqc in Rx2 *; split=>//; exists (g3 \+ g'). 
by rewrite joinA F3 -!joinA (sumIR _ tid) G1 unitL.
Qed.

End DoHelp.

(*************************)
(* try_collect stability *)
(*************************)

(* v should really be a ghost rather than a normal argument *)
(* but that requires non-trivial refactoring *)
(* see the comments at trycollect_act *)

Program Definition try_collect tid m v : {s g}, STsep [W]
  (specr (fun gL gE r gs =>
     [/\ s = gL, nat_of_ord tid \in dom (get_tid s) &
         requested tid v g gL gE r gs]),
   fun (x : option (fc_outT m)) => 
     specr (fun gL gE r gs =>
       if x is Some w then 
         exists g' g1, 
           [/\ gL = (get_tid s, get_mtx s, getU s \+ g'),
               r tid = NoReq T, gs tid = Unit, 
               [g <<= g1], [g1 <<= getU (gL \+ gE) \+ sum_up gs] & 
               fc_R v w g1 g']
       else s = gL /\ requested tid v g gL gE r gs)) :=
  Do (act (@try_collect_act _ tid m v)). 
Next Obligation.
apply: gh=>i [s g][r][b][ha][hr][Eh Ha] /= [-> Tx S] C _.
apply: val_act=>i1 M1; case: (menvs_coh M1)=>_ C1; move: (fc_cohJ C1)=>V.
have: specr (fun gL gE r gs => 
  nat_of_ord tid \in dom (get_tid gL) /\ requested tid v g gL gE r gs) i.
- by exists r, b, ha, hr.
move/(requested_reqS M1)=>S1; rewrite (menvs_loc M1).
case: (S1)=>r1 [b1][ha1][hr1][Eh1 Ha1 [Tx1 Rx1]] _.
split; first by exists r1, g, Rx1, b1, ha1, hr1. 
move=>x i2 i3 X M23; case: (menvs_coh M23)=>C2 C3; move: (fc_cohJ C3)=>V3. 
case: X=>/= r1' [g1][T1][b1'][ha1'][hr1'][ha2][_][Eh1' Ha1'] X _.
rewrite Eh1 /= in Eh1' V.
case: (fc_exE V Eh1' Ha1 Ha1')=>???? {Eh1' Ha1'}; subst b1' r1' ha1' hr1'.
move: (@cpf _ _ _ _ r1) {T1 X} (X _ Rx1); case Er1: (r1 tid)=>[|c f|m' v'].
- move=>_ [??]; subst i2 x.
  case/(requested_reqS M23): S1=>r' [b'][ha'][hr'][Eh' Ha'][Tx' Rx'] C'.
  by exists r', b', ha', hr'; rewrite -{1}(menvs_loc M23). 
- move=>_ [??]; subst i2 x.
  case/(requested_reqS M23): S1=>r' [b'][ha'][hr'][Eh' Ha'][Tx' Rx'] C'.
  by exists r', b', ha', hr'; rewrite -{1}(menvs_loc M23). 
move=>X [S2][Eh2 Ha2] Eg2 O2 ->{x}.
move: (X _ _ _ _ _ _)=>{X} X; subst m'; rewrite eqc.
case: (proj2 Rx1 _ _ Er1)=>pf [g2][Eg Eg']; rewrite eqc => Rg.
case/(stable_and (@noreq_reqS tid) (@gmonoS g2)): (M23).  
- exists [ffun z => if z == tid then NoReq T else r1 z], b1, ha2, hr1.
  rewrite Eg2 S2 O2 /= ffunE eq_refl; do !split=>//. 
  by rewrite -!joinA (joinC (getg i1 tid)) -joinA -(sum0R _ tid) joinA.
move=>r3 [b3][ha3][hr3][Eh3 Ha3][[Tx3 Er3]] Eg3 _.
exists r3, b3, ha3, hr3; split=>//; exists (getg i1 tid), g2.
split=>//.
- by rewrite -(menvs_loc M23) -S2. 
case: (C3)=>_ _ [rx][bx][hax][hrx][_ Ehx][Hax] Sx _ _ _. 
rewrite Eh3 in V3 Ehx.
case: (fc_exE V3 Ehx Ha3 Hax)=>????; subst bx rx hax hrx.
by move: (Sx tid); rewrite Er3. 
Qed.

(**************************)
(* read_request stability *)
(**************************)

Program Definition read_request (tid tid' : 'I_n) : {s g1 g m v}, STsep [W]
  (specr (fun gL gE r gs => 
     [/\ s = gL, get_mtx gL = up own, g1 = getU (gL \+ gE) \+ sum_up gs, 
         nat_of_ord tid \in dom (get_tid gL) & @requested tid m v g gL gE r gs]), 
   fun (x : req T) =>
   specr (fun gL gE r gs => 
     [/\ s = gL, g1 = getU (gL \+ gE) \+ sum_up gs, 
         r tid' = if x is Req _ _ then x else r tid',
         nat_of_ord tid \in dom (get_tid gL) & 
         @requested tid m v g gL gE r gs])) := 
  Do (act (@read_request_act _ tid')).
Next Obligation.
apply: gh=>i [_][g1][g][m v][r][b][ha][hr][Eh Ha] /= [-> Mx Eg Tx Rx] C _.
apply: val_act=>i1 M1; case: (menvs_coh M1)=>_ C1; split=>//.
rewrite (sumupCE M1 Mx) in Eg; rewrite (menvs_loc M1) in Mx *.
case/(@requested_reqS tid m v g): M1; first by exists r, b, ha, hr. 
move=>r1 [b1][ha1][hr1][Eh1 Ha1][Tx1 Rx1] _ {r b ha hr Eh Ha Tx Rx C}.
move=>x i2 i3 X M23; case: (menvs_coh M23)=>C2 C3.
case: X=>r2 [b2][ha2][hr2][Eh2 Ha2 Er2] ? _; subst i2. 
move: (fc_cohJ C1)=>V; rewrite Eh1 in V Eh2.
case: (fc_exE V Eh2 Ha1 Ha2)=>???? {Eh2 Ha2}; subst b2 r2 ha2 hr2.
case: x Er2.
- move=>Er2; case/(@requested_reqS tid m v g): (M23).
  - by exists r1, b1, ha1, hr1. 
  move=>r2 [b2][ha2][hr2][Eh2 Ha2][Tx2 Rx2] _. 
  exists r2, b2, ha2, hr2; split=>//; rewrite (menvs_loc M23). 
  by rewrite (sumupCE M23 Mx) in Eg. 
- move=>m' v' Er2.
  case/(stable_and (@req_helpS m' v' tid') (@requested_reqS tid m v g)): (M23).
  - by exists r1, b1, ha1, hr1. 
  move=>r2 [b2][ha2][hr2][Eh2 Ha2][[Mx2 Er2']][Tx2 Rx2] _. 
  exists r2, b2, ha2, hr2; split=>//; rewrite (menvs_loc M23).
  by rewrite (sumupCE M23 Mx) in Eg. 
move=>Er2; case/(@requested_reqS tid m v g): (M23).
- by exists r1, b1, ha1, hr1. 
move=>r2 [b2][ha2][hr2][Eh2 Ha2][Tx2 Rx2] _. 
exists r2, b2, ha2, hr2; split=>//; rewrite (menvs_loc M23). 
by rewrite (sumupCE M23 Mx) in Eg. 
Qed.
  
End Internals.


Section Externals.
Variable pv : nat.
Notation W := (entangle (Priv pv) (FlatCombine T)).
Notation pv_coh := (Mod.coh (Priv pv)).
Notation pt_coh := (Mod.coh (FlatCombine T)).
Notation coh := (Mod.coh W).
Notation guar := (Mod.internal W).
Notation pv_getso s := (getso (pvg pv) s).
Notation pv_self s := (self (pvg pv) s).
Notation pv_other s := (other (pvg pv) s).
Notation pv_joint s := (joint (pvg pv) s).
Notation fc_getso s := (getso (fcg T) s).
Notation fc_getj s := (getj (fcg T) s).
Notation fc_self s := (self (fcg T) s).
Notation fc_other s := (other (fcg T) s).
Notation fc_joint s := (joint (fcg T) s).
Notation fc_geth s := (geth (fcg T) s).
Notation fc_getg s := (getg (fcg T) s).

(*****************)
(* cas stability *)
(*****************)

(* first we need to lift requested to a larger concurroid *)
Definition made_request tid g m (v : fc_inT m) s := 
  nat_of_ord tid \in dom (get_tid (fc_self s)) /\
  (exists (r : {ffun 'I_n -> req T}) (b : bool) (ha hr : heap),
    [/\ fc_geth s = lk :-> b \+ ha \+ hr, is_array a r ha &
        requested tid v g (fc_self s) (fc_other s) r (fc_getg s)]).

(* we have a stability property *)

(* first for the base concurroid *)
Lemma made_reqS tid g m (v : fc_inT m) s1 s2 :
        menv_steps (FlatCombine T) s1 s2 ->
        made_request tid g v s1 -> made_request tid g v s2.
Proof.
move=>M [Mx]; case: (menvs_coh M)=>C1 C2. 
case=>r [b][ha][hr][Eh Ha Rx]. 
case/(@requested_reqS tid m v g): M.
- by exists r, b, ha, hr. 
move=>r1 [b1][ha1][hr1][Eh1 Ha1][Mx1 Rx1] Cx.
by split=>//; exists r1, b1, ha1, hr1.
Qed.

(* and then for an extension *)
Lemma made_req_extS (W : mod) tid g m (v : fc_inT m) s1 s2 :
        injects (FlatCombine T) W ->
        menv_steps W s1 s2 ->
        made_request tid g v s1 -> made_request tid g v s2.
Proof.
move=>inj M; case: (menvs_coh M)=>C1 C2.
rewrite (Injection.Exports.cohE inj) /= in C1 C2.
case: (C1)=>_ [x1][y1][?] Cx1 Cy1; subst s1.
case: (C2)=>_ [x2][y2][?] Cx2 Cy2; subst s2.
case/(menvs_split inj Cx1 Cx2): M=>/= M M'.
rewrite /made_request !(getC Cx1 C1) !(getC Cx2 C2) /=.
by apply: made_reqS M.
Qed.

Program Definition cas : {h s tid g m v}, STsep [W] 
  (fun i => [/\ h = pv_self i, s = fc_self i &
                @made_request tid g m v i],
   fun (x : bool) j => 
     if x then exists hr,
       [/\ pv_self j = h \+ hr, fc_self j = (get_tid s, up own, getU s),
           I (getU (fc_self j \+ fc_other j) \+ sum_up (fc_getg j)) hr & 
           @made_request tid g m v j]
     else [/\ pv_self j = h, fc_self j = s & @made_request tid g m v j]) := 
  Do (act (cas_act T pv)).
Next Obligation.         
apply: gh=>i [h][s][tid][g][m v][/= Hp Hf][Tx Rx] C.
apply: val_act=>i1 M1; case: (menvs_coh M1)=>_ C1.
case: (C)=>_ [xp][xf][?] Cp Cf; subst i. 
rewrite /made_request !(getC Cp C) !(getC Cf C) /= in Hp Hf Tx Rx *.
case: (C1)=>_ [xp1][xf1][?] Cp1 Cf1; subst i1.
rewrite /= /Actions.cas_safe /Actions.cas_step.
rewrite !(getC Cp1 C1) !(getC Cf1 C1) /=.
case/(menvs_split (injLE _ (erefl _)) Cp Cp1): M1=>/= Mp1 Mf1.
rewrite (menvs_loc Mp1) (menvs_loc Mf1) in Hp Hf.
case/(@requested_reqS tid m v g): Mf1.
- by case: Rx=>r [b][ha][hr][Eh Ha Rx]; exists r, b, ha, hr. 
move=>r [b][ha][hr][Eh Ha]{Tx Rx} [Tx Rx] _ {Cp Cf C Mp1}. 
split=>// y xpf2 xpf3 X M23; case: (menvs_coh M23)=>_ C3.
case: (C3)=>_ [xp3][xf3][?] Cp3 Cf3; subst xpf3.
rewrite !(getC Cp3 C3) !(getC Cf3 C3) /=.
case: X=>r1 [b1][ha1][hr1][Eh1 Ha1] X1 [_ C2].
move: (fc_cohJ Cf1)=>V; rewrite Eh in V Eh1.
case: (fc_exE V Eh1 Ha Ha1)=>???? {Eh1 Ha1}; subst b1 r1 ha1 hr1.
case: b V Eh X1=>V Eh.
- case=>??; subst xpf2 y.
  case/(menvs_split (injLE _ (erefl _)) Cp1 Cp3): M23=>/= Mp23 Mf23.
  rewrite -(menvs_loc Mp23) -{1 2}(menvs_loc Mf23); do !split=>//.
  case/(@requested_reqS tid m v g): (Mf23); first by exists r, true, ha, hr.
  by move=>r1 [b1][ha1][hr1][??[??]]; exists r1, b1, ha1, hr1.  
case=>Hp2 Op Hf2 Jf2 Of ->{y}.
case: (C2)=>_ [xp2][xf2][?] Cp2 Cf2; subst xpf2.
rewrite !(getC Cp2 C2) !(getC Cf2 C2) /= in Hp2 Op Hf2 Jf2 Of.
case/(menvs_split (injLE _ (erefl _)) Cp2 Cp3): M23=>/= Mp23 Mf23.
exists hr; split.
- by rewrite -(menvs_loc Mp23) Hp2 -Hp.
- by rewrite -(menvs_loc Mf23) Hf2 -Hf. 
- set g1 := (getU (fc_self xf2 \+ fc_other xf2) \+ sum_up (fc_getg xf2)). 
  case/(@sumup_helpS g1): Mf23; last first.
  - move=>r3 [b3][ha3][hr3][_ _][_ <-]; rewrite /g1 Hf2 Of Jf2 /=.
    by case: (fc_cohP Cf1 Eh Ha)=>_ _ _; case; case.
  case: (Cf2)=>_ _ [r2][b2][ha2][hr2][_ Eh2][Ha2] _ _ _ F.
  move: (fc_cohJ Cf2)=>V2; rewrite Jf2 -[_ \+ _]unitR /= in V2 Eh2.
  case: (fc_exE V2 Eh2 Ha Ha2)=>????; subst b2 r2 ha2 hr2.
  by exists r, true, ha, Unit; rewrite /g1 Jf2 unitR Hf2.
split; first by rewrite -(menvs_loc Mf23) Hf2.
case/(@requested_reqS tid m v g): (Mf23).
- by exists r, true, ha, Unit; rewrite Jf2 Hf2 Of /= unitR. 
by move=>r1 [b1][ha1][hr1][??[??]]; exists r1, b1, ha1, hr1. 
Qed.

(********************)
(* unlock stability *)
(********************)

Program Definition unlock : {h hr s tid g m v}, STsep [W]
  (fun i => [/\ pv_self i = h \+ hr, fc_self i = s, 
                get_mtx (fc_self i) = up own,
                I (getU (fc_self i \+ fc_other i) \+ sum_up (fc_getg i)) hr & 
                @made_request tid g m v i], 
   fun (_ : unit) j => 
      [/\ pv_self j = h, fc_self j = (get_tid s, up nown, getU s) & 
          @made_request tid g m v j]) := 
  Do (act (unlock_act T pv)).
Next Obligation.
apply: gh=>i [h][hr][s][tid][g][m v][/= Hp Hf Mx inv [Tx Rx]] C.
case: (C)=>_ [xp][xf][?] Cp Cf; subst i.
rewrite !(getC Cp C) !(getC Cf C) /= in Hp Hf Mx Tx Rx inv. 
apply: val_act=>xpf1 M; case: (menvs_coh M)=>_ C1.
case: (C1)=>_ [xp1][xf1][?] Cp1 Cf1; subst xpf1.
case/(menvs_split (injLE _ (erefl _)) Cp Cp1): M=>/= Mp Mf. 
rewrite /= -toPredE /Actions.unlock_safe /= /Actions.unlock_step.
rewrite !(getC Cp1 C1) !(getC Cf1 C1) /=.
rewrite (menvs_loc Mf) (menvs_loc Mp) in Hf Hp.
set g1 := (getU (fc_self xf \+ fc_other xf) \+ sum_up (fc_getg xf)).
case/(stable_and (@sumup_helpS g1) (@requested_reqS tid m v g)): Mf.
- case: (Cf)=>_ _ [r'][b'][ha'][hr'][_ Eh'][Ha']; exists r', b', ha', hr'.
  case: Rx=>r1 [b1][ha1][hr1][Eh1 Ha1].
  move: (fc_cohJ Cf)=>V; rewrite Eh1 in V Eh'.
  by case: (fc_exE V Eh' Ha1 Ha')=>????; subst b' r' ha' hr'. 
move=>r1 [b1][ha1][hr1][Eh1 Ha1][[Mx1 Eg1]][Tx1 Rx1] _ {Mx Tx Rx Cp Cf xp C Mp}.
split; first by exists hr, h; rewrite joinC -Eg1. 
case=>i2 i3 X M23; rewrite -(menvs_loc M23). 
case: X=>r2 [ha2][hr2][ha3][[Mx2]][Hp2 inv2][Eh2 Ha2][Sp2 Op2][Sf2 Jf2] Of2 [_ C2].
move: (fc_cohJ Cf1)=>V; rewrite -[_ \+ _]unitR Eh1 in V Eh2.
case: (fc_exE V Eh2 Ha1 Ha2)=>????; subst b1 r2 ha2 hr1.
rewrite -Eg1 in inv2. 
have <- : ha3 = h. 
- move: Hp Hp2 (cohVSO Cp1); rewrite /= joinC => -> E /validL V'.
  by move: (fc_precise V' E inv inv2) (E)=><- /(joinfK V').
rewrite -Hf; split=>//; apply: made_req_extS (injRE _ (erefl _)) M23 _.
by rewrite /made_request Sf2 Jf2 Of2; split=>//; exists r1, false, ha1, hr2. 
Qed.

End Externals.
End Stability.


(****************)
(* main program *)
(****************)

Section MainProgram.
Variables (T : flat_combineT) (pv al : nat) (asent alk : ptr). 
Notation a := (fc_orig T).
Notation lk := (fc_lk T).
Notation n := (@fc_n T).
Notation fc := (@fc_label T). 
Notation U := (@fc_pcm T).
Notation I := (@fc_inv T).
Notation V := (fc_V T).
Notation J := (fc_J T).
Notation Alloc := (AllocCAS al asent alk).
Notation PA := (entangle (Priv pv) Alloc).
Notation W := (entangle PA (FlatCombine T)).
Notation pv_coh := (priv_coh pv). 
Notation fc_coh := (fc_coh T). 
Notation al_coh := (caslock.LockCoh.Exports.lock_coh 
                    (DummyAllocator.lkT al asent alk)).
Notation coh := (Mod.coh W).
Notation guar := (Mod.internal W).
Notation pv_getso s := (getso (pvg pv) s).
Notation pv_self s := (self (pvg pv) s).
Notation pv_other s := (other (pvg pv) s).
Notation pv_joint s := (joint (pvg pv) s).
Notation fc_getso s := (getso (fcg T) s).
Notation fc_getj s := (getj (fcg T) s).
Notation fc_self s := (self (fcg T) s).
Notation fc_other s := (other (fcg T) s).
Notation fc_joint s := (joint (fcg T) s).
Notation fc_geth s := (geth (fcg T) s).
Notation fc_getg s := (getg (fcg T) s).

(* first we frame the provided methods *)
Program Definition framed_interp c v : 
    {g h}, STsep [entangle (Priv pv) (AllocCAS al asent alk)] 
      (fun i => exists hr, pv_self i = h \+ hr /\ fc_inv g hr, 
       fun r m => exists hr, pv_self m = h \+ hr /\
         exists g', [/\ valid (g \+ g'), fc_inv (g \+ g') hr & 
                        fc_R v r g g']) :=
  Do (@fc_interp T pv al asent alk c v).
Next Obligation.
apply: gh=>i [g h][hr] /= [Hp inv] C.
set gcs := [:: priv_getcoh pv; al_getcoh al asent alk].
have X: sself gcs i = (hr, self (alg al) i) \+ (h, Unit).
- by rewrite pcm_joinE /= joinC unitR Hp. 
apply: vrf_coh_post; apply: frame_do X _=>//.  
apply: (gh_ex g); apply: val_do; first by move/cohV=>V'; rewrite getE.
move=>r m [g'][V H Rx] _ [hr1 b1][Op _] /cohV V'.
by exists (pv_self m); rewrite getV //= joinC; split=>//; exists g'. 
Qed.

Variables (tid : 'I_n) (m : fc_code T) (v : fc_inT m).

Definition no_reqs s := 
  nat_of_ord tid \in dom (get_tid (fc_self s)) /\
  exists (r : {ffun 'I_n -> req T}) (b : bool) (ha hr : heap),
    [/\ fc_geth s = lk :-> b \+ ha \+ hr, is_array a r ha & r tid = NoReq T]. 

Notation "[ e ]" := (extend _ e).

Definition help_loop_tp := {h s g}, STsep [W] 
  (fun i => exists hr, 
     [/\ pv_self i = h \+ hr, fc_self i = s, 
         get_mtx (fc_self i) = up own, made_request tid g v i &
         I (getU (fc_self i \+ fc_other i) \+ sum_up (fc_getg i)) hr], 
   fun (_ : unit) j => exists hr,
     [/\ pv_self j = h \+ hr, fc_self j = s, made_request tid g v j &
         I (getU (fc_self j \+ fc_other j) \+ sum_up (fc_getg j)) hr]).

Program Definition help_loop :=
  ffix (fun (help_loop : nat -> help_loop_tp) t => 
    Do (if @decP _ (t < n) idP is left pf return _ then 
          x <-- [read_request tid (Ordinal pf)];
          if x is Req m' v' then 
            w' <-- [framed_interp v'];
            [do_help (Ordinal pf) w'];;
            help_loop t.+1
          else help_loop t.+1
        else ret tt)) 0.    
Next Obligation. exact: injRE (erefl _). Defined.
Next Obligation. exact: injLE (erefl _). Defined.
Next Obligation. exact: injRE (erefl _). Defined.
Next Obligation. 
apply: gh=>i [h][s g][hr][/= Hp Hf Mx [Tx Rx] inv] C. 
case: decP=>[pf|_]; last first.
- apply: val_ret=>i1 M; case: (menvs_coh M)=>_ C1.
  rewrite (menvs_loc M) in Hp Hf. 
  rewrite (sumupCE_ext (injRE _ (erefl _)) M Mx) in inv. 
  exists hr; split=>//. 
  case: (C)=>_ [x][y][?] Cx Cy; case: (C1)=>_ [x1][y1][?] Cx1 Cy1; subst i i1.
  case/(menvs_split (injLE (U:=PA) _ (erefl _)) Cx Cx1): M=>Mp /= Mf.
  rewrite /made_request !(getC Cy C) !(getC Cy1 C1) /= in Tx Rx *.
  by apply: made_reqS Mf (conj Tx Rx).
case: (C)=>_ [s'][xf][?][_][xp][xa][?] Cp _ Cf; subst s' i.
rewrite /made_request !(getC Cf C) !(getC Cp C) /= in Hf Hp Mx inv Tx Rx *.
rewrite joinC in C *; apply: step; apply: val_extend=>//.
set g1 := ((fc_self xf).2 \+ _ \+ _) in inv.
apply: (gh_ex (fc_self xf)); apply: (gh_ex g1); apply: (gh_ex g).
apply: (gh_ex m); apply: (gh_ex v); apply: val_do. 
- case=>_ _ [r][b][ha][hr1][_ Eh][Ha]; exists r, b, ha, hr1; do 2!split=>//.
  case: Rx=>r' [b'][ha'][hr'][Eh' Ha']; move: (fc_cohJ Cf) => Vx.
  by rewrite Eh in Vx Eh'; case: (fc_exE Vx Eh' Ha Ha')=>_ ->. 
move=>x xf1 [r][b][ha][hr1][Eh Ha][Sf Eg Er Tx1 Rx1] Cf1 _ xap1 C1 /= M. 
have Cpa1 : xap1 \In Mod.coh PA.
- by rewrite /= starC in C1; apply/(star_coh_prec C1).
case: (Cpa1)=>_ [xp1][xa1][?] Cp1 _; subst xap1.
case/(menvs_split (injLE _ (erefl _)) Cp Cp1): M=>Mp _.
rewrite {Mp Tx Rx C xa xp Cp Cf} (menvs_loc Mp) !{}Sf in Hp Hf Mx *.
case E: x=>[|m' v'|]; try by 
  [apply: (gh_ex h); apply: (gh_ex s); apply: (gh_ex g); apply: val_do=>// _;
   rewrite !(getC Cp1 C1) !(getC Cf1 C1) /= -Eg; 
   exists hr; do !split=>//; exists r, b, ha, hr1]. 
rewrite E in Er; rewrite joinC in C1 *; apply: step; apply: val_extend=>//.
apply: (gh_ex g1); apply: (gh_ex h); apply: val_do.
- by exists hr; rewrite (getC Cp1 Cpa1).
move=>w xpa2 [hr2][Hp2][g'][V2 inv2 R2] Cpa2 xf2 C2 /= Mf.
case: (menvs_coh Mf)=>_ Cf2; case: (Cpa2)=>_ [xp2][xa2][?] Cp2 _; subst xpa2.
rewrite {Hp} (getC Cp2 Cpa2) /= in Hp2.
rewrite {Cp1 xp1 xa1 Cpa1 C1} (menvs_loc Mf) in Hf.
rewrite (sumupCE Mf Mx) in Eg.
case/(stable_and (@req_helpS T m' v' (Ordinal pf)) 
     (@requested_reqS T tid m v g)): Mf. 
- by exists r, b, ha, hr1; rewrite Er.  
move=>{xf1 Cf1 hr1 ha b r Eh Ha Mx Tx1 Er Rx1} r2 [b2][ha2][hr2'][Eh Ha][]. 
case=>Mx Er [Tx] Rx _; rewrite joinC in C2 *.
apply: step; apply: val_extend=>//. 
apply: (gh_ex s); apply: (gh_ex g); apply: (gh_ex g1); apply: (gh_ex g').
apply: (gh_ex tid); apply: (gh_ex m); apply: (gh_ex v).
apply: val_do; first by exists r2, b2, ha2, hr2'; do !split=>//; exists v'. 
case=>xf3 [r3][b3][ha3][hr3][Eh3 Ha3][Hf3] Rf3 Eg3 Cf3 _ xpa3 C3 /= M.
case: (menvs_coh M)=>_ Cpa3; case: (Cpa3)=>_ [xp3][xa3][?] Cp3 _; subst xpa3.
case/(menvs_split (injLE _ (erefl _)) Cp2 Cp3):  M=>Mp _.
rewrite {xp2 Cp2 Cpa2 C2 Mp} (menvs_loc Mp) in Hp2.
rewrite {xf2 Cf2 Eg Eh Rx xa2 ha2 Ha hr2' b2 r2 Er}Hf Hf3 in Tx Mx. 
apply: (gh_ex h); apply: (gh_ex s); apply: (gh_ex g); apply: val_do=>// _.
rewrite !(getC Cp3 C3) !(getC Cf3 C3) /= -Eg3.
by exists hr2; do !split=>//; exists r3, b3, ha3, hr3. 
Qed.

Program Definition cas_help : {h s g}, STsep [W]
  (fun i => [/\ pv_self i = h, fc_self i = s, get_mtx (fc_self i) = up nown &
                made_request tid g v i], 
   fun (_ : unit) j => 
            [/\ pv_self j = h, fc_self j = s & 
                made_request tid g v j]) := 
  Do (x <-- [cas T pv];
      if x then 
        help_loop;; 
        [unlock T pv] 
      else ret tt).
Next Obligation. exact: injLE (entangleAC _ _ _). Defined.
Next Obligation. exact: injLE (entangleAC _ _ _). Defined.
Next Obligation.
apply: gh=>i [h][s g] /= [Hp Hf Mx][Tx Rx] C.
case: (C)=>_ [s'][xf][?][_][xp][xa][?] Cp Ca Cf; subst s' i.
rewrite /made_request !(getC Cf C) !(getC Cp C) /= in Hf Hp Mx Tx Rx *.
rewrite joinAC starAC in C *. 
apply: step; apply: val_extend; first by apply/(star_coh_prec C).
apply: (gh_ex h); apply: (gh_ex s); apply: (gh_ex tid). 
apply: (gh_ex g); apply: (gh_ex m); apply: (gh_ex v); apply: val_do.
- by move=>C'; rewrite /made_request !(getC Cp C') !(getC Cf C'). 
move=>x xpf1 X Cpf1 xa1 C1 /= M {Tx Rx C Cp Ca Cf}.
case: (Cpf1)=>_ [xp1][xf1][?] Cp1 Cf1; subst xpf1.
rewrite !(getC Cp1 Cpf1) !(getC Cf1 Cpf1) /= in X.
case: x X; last first.
- case=>Hp1 Hf1 Rx1; apply: val_ret=>i3 M3.
  rewrite -{1 2}(menvs_loc M3) (getC Cp1 C1) (getC Cf1 C1). 
  split=>//; apply: made_req_extS (injRE _ (erefl _)) M3 _. 
  by rewrite /made_request !(getC Cf1 C1). 
case=>hr [Hp1 Hf1 inv1] Rx1.
apply: step; apply: (gh_ex h); apply: (gh_ex (fc_self xf1)); apply: (gh_ex g).
apply: val_do; first by rewrite !(getC Cp1 C1) !(getC Cf1 C1) {3}Hf1; exists hr.
case=>i2 [hr2][Hp2 Hf2 Rx2 inv2] C2.
case: (C2)=>_ [xpa2][xf2][?] Cpa2 Cf2; subst i2.
case: (Cpa2)=>_ [xp2][xa2][?] Cp2 Ca2; subst xpa2.
rewrite !(getC Cp2 C2) !(getC Cf2 C2) /= in inv2 Hf2 Hp2 Rx2.
rewrite joinAC /= starAC in C2 *. 
apply: val_extend; first by apply/(star_coh_prec C2).
apply: (gh_ex h); apply: (gh_ex hr2); apply: (gh_ex (fc_self xf2)). 
apply: (gh_ex tid); apply: (gh_ex g); apply: (gh_ex m); apply: (gh_ex v).
apply: val_do.
- by move=>Cfp2; rewrite !(getC Cp2 Cfp2) !(getC Cf2 Cfp2) /= {3}Hf2 Hf1.
case=>xpf3 [Hp3 Hf3 Rx3] Cpf3 xa3 C3 M3.
case: (Cpf3)=>_ [xp3][xf3][?] Cp3 Cf3; subst xpf3.
rewrite !(getC Cp3 Cpf3) !(getC Cf3 Cpf3) in Hp3 Hf3 Rx3.
rewrite !(getC Cp3 C3) !(getC Cf3 C3) /=; split=>//.
by rewrite Hf3 Hf2 {}Hf1 -Mx {}Hf; case: s; case.
Qed.

Definition cas_loop_tp := {h s g}, STsep [W] 
  (fun i => [/\ pv_self i = h, fc_self i = s, get_mtx s = up nown &
                made_request tid g v i],
   fun (w : fc_outT m) j => 
     exists g' g1,
     [/\ pv_self j = h, fc_self j = (get_tid s, up nown, getU s \+ g'), 
         [g1 <<= getU (fc_self j \+ fc_other j) \+ sum_up (fc_getg j)], 
         [g <<= g1], fc_R v w g1 g' & no_reqs j]).

Program Definition cas_loop := 
  ffix (fun (cas_loop : unit -> cas_loop_tp) xx => 
     Do (cas_help;; 
         ow <-- [try_collect tid v];
         if ow is Some w return _ then ret w
         else cas_loop tt)) tt.
Next Obligation. exact: injRE (erefl _). Defined.
Next Obligation. 
apply: gh=>i [_][_ g][/= <-<- Mx Rx] C. 
apply: step; apply: (gh_ex _); apply: (gh_ex _); apply: (gh_ex g).
apply: val_do=>//; case=>i2 [Hp2 Hf2 [Tx2 Rx2]] C2.
case: (C2)=>_ [xpa2][xf2][?] Cpa2 Cf2; subst i2.
case: (Cpa2)=>_ [xp2][xa2][?] Cp2 Ca2; subst xpa2.
rewrite /made_request !(getC Cp2 C2) !(getC Cf2 C2) /= in Hp2 Hf2 Tx2 Rx2.
rewrite {Rx C Cpa2 Ca2} -Hf2 in Mx.
rewrite joinC /= starC in C2 *; apply: step; apply: val_extend=>//.
apply: (gh_ex (fc_self xf2)); apply: (gh_ex g); apply: val_do.
- by case: Rx2=>r [b][ha][hr][]; exists r, b, ha, hr. 
move=>x xf3 X Cf3 xpa3 C3 /= M; case: (menvs_coh M)=>_ Cpa3.
case: (Cpa3)=>_ [xp3][xa3][?] Cp3 Ca3; subst xpa3.
case/(menvs_split (injLE _ (erefl _)) Cp2 Cp3): M=>/= Mp _.
rewrite (menvs_loc Mp) in Hp2.
case: x X; last first.
- case=>r [b][ha][hr][Eh Ha][Sf Rx] _; rewrite Sf in Mx Hf2 Tx2.
  apply: (gh_ex (pv_self i)); apply: (gh_ex (fc_self i)).
  apply: (gh_ex g); apply: val_do=>//.
  rewrite /made_request !(getC Cp3 C3) !(getC Cf3 C3) /= -Hf2 -Hp2. 
  by do !split=>//; exists r, b, ha, hr.  
move=>w [r][b][ha][hr][Eh3 Ha3][g'][g1][Sf3 Er3 Gf3 Eg31 Eg32 R3] _.
rewrite joinC; apply: val_ret=>i4 M; case: (menvs_coh M)=>_ C4.
case: (C4)=>_ [xpa4][xf4][?] Cpa4 Cf4; subst i4.
case: (Cpa4)=>_ [xp4][xa4][?] Cp4 Ca4; subst xpa4.
rewrite /no_reqs !(getC Cp4 C4) !(getC Cf4 C4) /=.
case/(menvs_split (injLE _ (erefl _)) Cpa3 Cpa4): M.
case/(menvs_split (injLE _ (erefl _)) Cp3 Cp4)=>Mp3 _ /= Mf3.
set g2 := (fc_self xf3).2 \+ _ \+ _ in Eg32.
case/(stable_and (@noreq_reqS T tid) (@gmonoS T g2)): (Mf3).
- exists r, b, ha, hr; rewrite {1}Sf3; do !split=>//. 
  by exists Unit; rewrite unitR.  
rewrite (menvs_loc Mp3) (menvs_loc Mf3) in Hp2 Sf3.
move=>r1 [b1][ha1][hr1][Eh4 Ha4][[Tx4 Er4]][g3] Eg4 _.
exists g', g1; do !split=>//.
- by rewrite Sf3 -Mx Hf2. 
- by case: Eg32 Eg4=>g4 <- Eg4; exists (g4 \+ g3); rewrite joinA Eg4. 
by exists r1, b1, ha1, hr1. 
Qed.

Program Definition flat_combine : {h s g}, STsep [W] 
  (fun i => [/\ pv_self i = h, fc_self i = s, 
                [g <<= getU (fc_self i \+ fc_other i) \+ sum_up (fc_getg i)],
                get_mtx s = up nown & no_reqs i],
   fun (w : fc_outT m) j => 
     exists g' g1,
     [/\ pv_self j = h, fc_self j = (get_tid s, up nown, getU s \+ g'), 
         [g1 <<= getU (fc_self j \+ fc_other j) \+ sum_up (fc_getg j)],
         [g <<= g1], fc_R v w g1 g' & no_reqs j]) := 
   Do ([req_help tid v];; 
       cas_loop).
Next Obligation. exact: injRE (erefl _). Defined.
Next Obligation.
apply: gh=>i [_][_ g][/= <-<- Eg Mx [Tx Rx]] C.
case: (C)=>_ [xpa][xf][?] Cpa Cf; subst i.
case: (Cpa)=>_ [xp][xa][?] Cp Ca; subst xpa.
rewrite /no_reqs !(getC Cp C) !(getC Cf C) /= in Eg Mx Tx Rx *.
rewrite joinC /= starC in C *.
apply: step; apply: val_extend=>//.
apply: (gh_ex (fc_self xf)); apply: (gh_ex g); apply: val_do.
- by case: Rx=>r [b][ha][hr][???]; exists r, b, ha, hr. 
case=>xf2 X Cf2 xpa2 C2 /= M; case: (menvs_coh M)=>_ Cpa2.
case: (Cpa2)=>_ [xp2][xa2][?] Cp2 Ca2; subst xpa2.
case/(menvs_split (injLE _ (erefl _)) Cp Cp2): M=>/= M _.
rewrite (menvs_loc M).
apply: (gh_ex (pv_self xp2)); apply: (gh_ex (fc_self xf2)); apply: (gh_ex g).
case: X=>r [b][ha][hr][Eh Ha][Sf2 Eg2 Rx2] _; rewrite Sf2 in Mx Tx *. 
apply: val_do=>//. 
rewrite /made_request !(getC Cp2 C2) !(getC Cf2 C2). 
by do !split=>//; exists r, b, ha, hr.
Qed.

(* a small footprint version of flat_combiner, just to clear up *)
(* the specs as much as possible *)

Program Definition flat_combine' : {g}, STsep [W] 
  (fun i => [/\ pv_self i = Unit, fc_self i = (#tid, up nown, Unit), 
                [g <<= getU (fc_self i \+ fc_other i) \+ sum_up (fc_getg i)] & 
                no_reqs i],
   fun (w : fc_outT m) j => 
     exists g' g1,
     [/\ pv_self j = Unit, fc_self j = (#tid, up nown, g'), 
         [g1 <<= g' \+ getU (fc_other j) \+ sum_up (fc_getg j)],
         [g <<= g1], fc_R v w g1 g' & no_reqs j]) := 
   Do (flat_combine). 
Next Obligation. 
apply: gh=>i /= g [Hp Hf Eg N] _. 
apply: (gh_ex (pv_self i)); apply: (gh_ex (fc_self i)); apply: (gh_ex g).
rewrite Hp Hf; apply: val_do=>// x j [g'][g1]; rewrite unitL. 
by case=>? -> /= ????; exists g', g1. 
Qed.

End MainProgram.
