Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq path.
Require Import pred prelude idynamic ordtype finmap pcm coding unionmap heap.
Require Import feap state zigzag world inject rely hide atomic autosplit.
Require Import privstate footprint model lemmas stsep.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(***********)
(* Getters *)
(***********)

(* general getters, given an encoded pcm/tp, gives the field *)

Structure getter := Getter {
  getL : nat; (* label *)
  getU : encoded_pcm; (* result type of SO getters *)
  getT : encoded_type; (* result type of J getter *)
  getD : getT}. (* default value for J getter *)

Section Definitions.
Variable G : getter. 
Notation l := (getL G).
Notation U := (getU G).
Notation T := (getT G).
Notation d := (getD G). 
Notation sort_pcm := (to_type \o pcm_sort).
Notation idyn_pcm := (@idyn_tp _ sort_pcm). 
Notation sort_tp := (to_type \o tp_sort).
Notation idyn_tp := (@idyn_tp _ sort_tp).

Implicit Types (f p : feap dpcm) (s : state).

Lemma pcm_decodeE : pcm_sort (decode U) = U.
Proof. by case: U=>c [w E s] /=; rewrite /decode /= -E. Qed.

Lemma tp_decodeE : tp_sort (decode T) = T.
Proof. by case: T=>c [w E s] /=; rewrite /decode /= -E. Qed.

Lemma getso_pf v : 
        idyn_pcm v == decode U -> to_type (pcm_sort (idyn_pcm v)) = U.
Proof. by move/eqP=>->; rewrite pcm_decodeE. Qed.

Lemma getj_pf v : 
        idyn_tp v == decode T -> to_type (tp_sort (idyn_tp v)) = T.
Proof. by move/eqP=>->; rewrite tp_decodeE. Qed.

Definition getso' f : U := 
  if flook l f is Some v then 
    if @decP _ (idyn_pcm v == decode U) idP is left pf 
    then icoerce id (idyn_val v) (getso_pf pf) else Unit
  else Unit.

(* for joint, d is the default value *)
(* in current use, always the empty heap, but this may change *)
Definition getj' (f : feap dtp) : T := 
  if flook l f is Some v then 
    if @decP _ (idyn_tp v == decode T) idP is left pf 
    then icoerce id (idyn_val v) (getj_pf pf) else d
  else d.

End Definitions.

Notation getso G s := (getso' G s).
Notation self G s := (getso G (loc s)).
Notation other G s := (getso G (env s)).
Notation getj G s := (getj' G s).
Notation joint G s := (getj G (shr s)). 

Section ManualLemmas.
Variable G : getter. 
Notation getso s := (getso G s).
Notation getj s := (getj G s).
Notation self s := (self G s).
Notation other s := (other G s).
Notation joint s := (joint G s). 
Notation l := (getL G).
Notation U := (getU G).
Notation T := (getT G).

Lemma getsoZ f1 f2 : 
        valid (f1 \* f2) -> getso (f1 \* f2) = getso f1 \+ getso f2.
Proof.
move=>V; rewrite /getso'.
case E : (flook _ _)=>[a|]; last first.
- move: {E} (flook_none E) (flook_none E); rewrite {1}fdomZL fdomZR V.
  by do 2![case: fdom_flook=>// -> /=]; rewrite unitR.
rewrite (pcm_eta a) in E; case/flookZf: E=>{V} u1 [u2][E1] ->->.
case: decP=>/= H; last by rewrite unitR. 
rewrite /pcm_enc/code_enc/encode/pcm_val/code_val /= in u1 u2 E1 *.
do 3![move: (getso_pf _)]; rewrite !ieqc {}E1 /=. 
rewrite /decode /= {H}(eqP H) pcm_decodeE /= in u1 u2 *.
by move=>*; rewrite !ieqc.
Qed.

Lemma getsoPt u : getso (l \-> u) = u.
Proof.
rewrite /getso' flookU eq_refl /=; case: decP=>// pf. 
move: {pf} (getso_pf _) (esym _). 
by rewrite /encode /= pcm_decodeE=>H1 H2; rewrite !ieqc. 
Qed.

Lemma getjPt x : getj (l ::-> x) = x.
Proof.
rewrite /getj' flookU eq_refl /=; case: decP=>// pf.
move: {pf} (getj_pf _) (esym _). 
by rewrite /encode /= tp_decodeE=>H1 H2; rewrite !ieqc. 
Qed.

Lemma getsoPtUn u f : valid (l \-> u \+ f) -> getso (l \-> u \+ f) = u.
Proof. 
by move=>V; move: (getsoPt u); rewrite /getso' flookUnl // fdomPt inE eq_refl.
Qed.

Lemma getjPtUn x f : valid (l ::-> x \+ f) -> getj (l ::-> x \+ f) = x.
Proof.
by move=>V; move: (getjPt x); rewrite /getj' flookUnl // fdomPt inE eq_refl.
Qed.

Lemma getsoPtUnR (U : encoded_pcm) (l' : nat) (x : U) (f : feap dpcm) : 
        valid (l' \-> x \+ f) -> l' != l ->
        getso (l' \-> x \+ f) =  getso f.
Proof. by move=>V N; rewrite /getso' flookUnl // fdomPt inE (negbTE N). Qed.

Lemma getjPtUnR (T : encoded_type) (l' : nat) (x : T) (f : feap dtp) : 
        valid (l' ::-> x \+ f) -> l' != l ->
        getj (l' ::-> x \+ f) =  getj f.
Proof. by move=>V N; rewrite /getj' flookUnl // fdomPt inE (negbTE N). Qed.

Lemma getzE s p : 
        (self (s \> p) = self s) * (other (s \< p) = other s) *
        (joint (s \< p) = joint s) * (joint (s \> p) = joint s).
Proof. by rewrite /getj' !spP. Qed.

Lemma getzV s p : 
        validSO (s \< p) -> (self (s \< p) = self s \+ getso p) * 
                            (other (s \> p) = getso p \+ other s).
Proof.
rewrite /validSO !spP => /allvalidV V; rewrite !getsoZ ?(joinC (getso p)) //.
- by move: V; rewrite zipAC validZE zipC; case/andP.  
by move: V; rewrite validZE; case/andP.
Qed.

Lemma zagzigVE s p : 
        validSO (s \< p) ->
        self (s \< p) \+ other (s \< p) = self (s \> p) \+ other (s \> p).
Proof. by move=>Voh; rewrite !(getzE, getzV) // joinA. Qed.

Lemma zagzigE s p (h : T) : 
        s \< p = l ->> [self (s \< p), h, other (s \< p)] <->
        s \> p = l ->> [self (s \> p), h, other (s \> p)].
Proof.
rewrite !spP -[s]st_eta zigE zagE /= !fsE; split.
- case/stE=>/= /zipPtf /= [u1][t2][H1 ->->] ->->.
  by rewrite zipPt -!fsE !getsoPt. 
case/stE=>/= ->-> /zipPtf [u1][u2][H1 ->->].
by rewrite zipPt -!fsE !getsoPt.
Qed.
 
Lemma joint_transp s : joint (transp s) = joint s.
Proof. by []. Qed.

Lemma getPt (us uo : U) (uj : T) : 
        (self (l ->> [us, uj, uo]) = us) * 
        (other (l ->> [us, uj, uo]) = uo) *
        (joint (l ->> [us, uj, uo]) = uj).
Proof. 
rewrite fsE !getsoPt /getj' flookU eq_refl /=; split=>//.
case: decP=>// pf; move: {pf} (getj_pf _) (esym _). 
by rewrite /encode /= tp_decodeE => H1 H2; rewrite !ieqc. 
Qed.

Lemma getPtUn (us uo : U) (uj : T) s : 
        valid (l ->> [us, uj, uo] \+ s) -> 
        (self (l ->> [us, uj, uo] \+ s) = us) * 
        (other (l ->> [us, uj, uo] \+ s) = uo) *
        (joint (l ->> [us, uj, uo] \+ s) = uj). 
Proof.
move: (getPt us uo uj); rewrite fsE /= => E; case/and3P=>/= V1 V2 V3.
by rewrite /getso' /getj' !flookUnl // fdomPt inE eq_refl in E *. 
Qed.

Definition getE := (getzE, joint_transp, getsoPt, getjPt, getPt, getPtUn).

(* lemma for stability of self, other, joint under injection *)

Lemma getCLV s1 s2 (c1 : cohpred) : 
        s1 \In c1 -> valid (s1 \+ s2) -> 
        (self (s1 \+ s2) = if l \in fV c1 then self s1 else self s2) *
        (other (s1 \+ s2) = if l \in fV c1 then other s1 else other s2) * 
        (joint (s1 \+ s2) = if l \in fV c1 then joint s1 else joint s2).
Proof. 
move=>C /and3P [V1 V2 V3]; case/cohfV: C=>E1 E2 E3.
split; first by split; rewrite /getso' !flookUnl // ?E1 ?E3; case: ifP. 
by rewrite /getj' flookUnl // E2; case: ifP.
Qed.

(* we frequently have a coherence lying around *)
(* so let's use that instead of validity *)

Lemma getCL s1 s2 (c1 c2 : cohpred) : 
        s1 \In c1 -> s1 \+ s2 \In c2 -> 
        (self (s1 \+ s2) = if l \in fV c1 then self s1 else self s2) *
        (other (s1 \+ s2) = if l \in fV c1 then other s1 else other s2) * 
        (joint (s1 \+ s2) = if l \in fV c1 then joint s1 else joint s2).
Proof. by move=>C /cohV; apply: getCLV. Qed.

End ManualLemmas.

Arguments getzV [G][s][p] _.


(*********************************************)
(* Structure combining getters and coherence *)
(*********************************************)

Module GetCoh.

Structure mixin_of (G : getter) (C : cohpred) := Mixin {
  _ : forall s, s \In C -> s = getL G ->> [self G s, joint G s, other G s]}.

Section ClassDef. 


Structure type : Type := 
  Pack {sortg : getter; sortc : cohpred; _ : mixin_of sortg sortc}.

Local Coercion sortg : type >-> getter.
(* Local Coercion sortc : type >-> cohpred. *)

Variables (G : getter) (C : cohpred) (cT : type).

Definition class := let: Pack _ _ c as cT' := cT return mixin_of cT' (sortc cT') in c.
Definition pack c := @Pack G C c.
Definition clone := fun c & G = cT & C = sortc cT & phant_id (pack c) cT => pack c.

End ClassDef.

Module Exports.
Coercion sortg : type >-> getter.
(* Coercion sortc : type >-> cohpred. *)
Notation getcoh := type.
Notation GCMixin := Mixin.
Notation GC G C m := (@pack G C m).
Notation coh_of := sortc.
Notation get_of := sortg.

Notation "[ 'getcoh' 'of' 'getter' G ]" := 
  (@clone G _ _ _ (erefl _) (erefl _) id).
Notation "[ 'getcoh' 'of' 'coh' C ]" := 
  (@clone _ C _ _ (erefl _) (erefl _) id).

Section GenericLemmas. 
Variable gc : getcoh. 
Notation coh := (coh_of gc).
Notation l := (getL (get_of gc)).

Lemma cohE s : s \In coh -> s = l ->> [self gc s, joint gc s, other gc s].
Proof. by case: gc=>g c []; apply. Qed.

Lemma gcfvE s : s \In coh -> fV coh =i [:: l]. 
Proof.
move=>C; case: (cohfV C)=>V _ _; move=>x.
by rewrite -V (cohE C) fsE fdomPt inE.
Qed.

Lemma cohVSO s : s \In coh -> valid (self gc s \+ other gc s).
Proof. 
move=>C; move: (cohSO C).
by rewrite {1}(cohE C) /validSO fsE zipPt allvalidPt.
Qed.

Lemma zigSOE s p : s \< p \In coh -> p = l \-> getso gc p.
Proof.
case/cohE/stE; rewrite !spP fsE /=.
by case/zipPtf=>u1 [u2][_ ->]; rewrite getsoPt.
Qed.

Lemma zagSOE s p : s \> p \In coh -> p = l \-> getso gc p.
Proof. by move/coh_zigzag/zigSOE. Qed.

Lemma getzC s p :
        s \< p \In coh -> (self gc (s \< p) = self gc s \+ getso gc p) *
                          (other gc (s \> p) = getso gc p \+ other gc s).
Proof. by move/cohSO; apply: getzV. Qed.

End GenericLemmas.

Arguments getzC [gc][s][p] _. 

End Exports.

End GetCoh.

Export GetCoh.Exports. 


(*********************************************************)
(* lemma for searching in an expression, given coherence *)
(*********************************************************)

Module GetAuto1.
Section GetAuto1.

Structure tagged_state := Tag {untag : state}.
Definition right_tag := Tag.
Definition left_tag := right_tag.
Canonical found_tag s := left_tag s.

Definition search_ax s (sx : tagged_state) := 
  exists s', untag sx = s \+ s'.

Structure search_state s := SearchState {
  state_of : tagged_state;
  _ : search_ax s state_of}.

Local Coercion state_of : search_state >-> tagged_state.

Lemma found_pf s : search_ax s (found_tag s).
Proof. by exists Unit; rewrite unitR. Qed.

Canonical found_state s := SearchState (found_pf s).

Lemma left_pf s s' (f : search_state s) : 
        search_ax s (left_tag (untag f \+ s')).
Proof. by case: f=>/= f [s2] ->; exists (s2 \+ s'); rewrite joinA. Qed.

Canonical left_recurse s s' f := SearchState (@left_pf s s' f).

Lemma right_pf s s' (f : search_state s) :
        search_ax s (right_tag (s' \+ untag f)). 
Proof. by case: f=>/= f [s2] ->; exists (s' \+ s2); rewrite joinCA. Qed.

Canonical right_recurse s s' f := SearchState (@right_pf s s' f).

End GetAuto1.

Module Exports.
Coercion state_of : search_state >-> tagged_state.
Canonical found_tag.
Canonical found_state.
Canonical left_recurse.
Canonical right_recurse.

Lemma getCV gc s1 (s2 : search_state s1) :
        s1 \In coh_of gc -> valid (untag s2) -> 
        (self gc (untag s2) = self gc s1) * 
        (other gc (untag s2) = other gc s1) *
        (joint gc (untag s2) = joint gc s1).
Proof. 
case: s2=>_ [/= s2 ->] C V; do !split=>//; 
by rewrite (getCLV gc C V) (gcfvE C) inE eq_refl.
Qed.

Lemma getC gc s1 (s2 : search_state s1) (c2 : cohpred) :
        s1 \In coh_of gc -> untag s2 \In c2 -> 
        (self gc (untag s2) = self gc s1) * 
        (other gc (untag s2) = other gc s1) *
        (joint gc (untag s2) = joint gc s1).
Proof. by move=>C /cohV; apply: getCV. Qed.

Arguments getCV [gc][s1][s2] _ _.
Arguments getC [gc][s1][s2][c2] _ _.

(* should we use the notation below? 
Notation "'blast' x1 & x" := (getC x1 x) (at level 0). 
Notation "'blast' x1 , x2 , .. , x3 & x" := 
  (pair .. (pair (getC x1 x) (getC x2 x)) .. (getC x3 x)) (at level 0).
*)
End Exports.

End GetAuto1.

Export GetAuto1.Exports.


(******************************************************************)       
(* lemmas for searching in a *feap* expresison, without coherence *)
(* but needs the rewrite target to be eta-expanded wrt. label     *)
(******************************************************************)

Module GetAuto2.
Section GetAuto2.
Variable gc : getcoh. 
Notation U := (getU (get_of gc)).
Notation x := (getL (get_of gc)).

Structure tagged_feap := Tag {untag : feap dpcm}.
Definition right_tag := Tag.
Definition left_tag := right_tag.
Canonical found_tag f := left_tag f.

Definition search_ax (f : tagged_feap) (v : U) := 
  exists f', untag f =  x \-> v \+ f'. 

Structure search_label := SearchLabel {
  feap_of : tagged_feap;
  v_of : getU gc;
  _ : search_ax feap_of v_of}. 

Local Coercion feap_of : search_label >-> tagged_feap.

Lemma found_pf (v : getU gc) : search_ax (found_tag (x \-> v)) v. 
Proof. by exists Unit; rewrite unitR. Qed.

Canonical found_label v := SearchLabel (found_pf v).

Lemma left_pf f' (f : search_label) : 
        search_ax (left_tag (untag f \+ f')) (v_of f).
Proof. by case: f=>/= f v [f2 ->]; exists (f2 \+ f'); rewrite joinA. Qed.

Canonical left_recurse f' f := SearchLabel (@left_pf f' f).

Lemma right_pf f' (f : search_label) :
        search_ax (right_tag (f' \+ untag f)) (v_of f). 
Proof. by case: f=>/= f v [f2 ->]; exists (f' \+ f2); rewrite joinCA. Qed.

Canonical right_recurse f' f := SearchLabel (@right_pf f' f).

End GetAuto2. 

Module Exports.
Coercion feap_of : search_label >-> tagged_feap.
Canonical found_tag.
Canonical found_label.
Canonical left_recurse.
Canonical right_recurse.

Lemma fgetV gc (f : search_label gc) : 
        valid (untag f) -> getso gc (untag f) = v_of f.
Proof. by case: f=>f v [/= f2 ->]; apply: getsoPtUn. Qed.

Arguments fgetV [gc][f] _.

End Exports.
End GetAuto2.

Export GetAuto2.Exports.


(****************************************************************)       
(* now the same as above, but searching in a *state* expresison *)
(* but needs the rewrite target to be eta-expanded wrt. label   *)
(****************************************************************)       

Module GetAuto3.
Section GetAuto3.
Variable gc : getcoh. 
Notation U := (getU (get_of gc)).
Notation T := (getT (get_of gc)).
Notation x := (getL (get_of gc)).

Structure tagged_state := Tag {untag : state}.
Definition right_tag := Tag.
Definition left_tag := right_tag.
Canonical found_tag f := left_tag f.

Definition search_ax s (xs os : U) (js : T) := 
  exists s', untag s = x ->> [xs, js, os] \+ s'. 

Structure search_label := SearchLabel {
  state_of : tagged_state;
  xs_of : U; os_of : U; js_of : T; 
  _ : search_ax state_of xs_of os_of js_of}. 

Local Coercion state_of : search_label >-> tagged_state.

Lemma found_pf xs os js : 
        search_ax (found_tag (x ->> [xs, js, os])) xs os js. 
Proof. by exists Unit; rewrite unitR. Qed.

Canonical found_label xs os js := SearchLabel (found_pf xs os js).

Lemma left_pf s' (s : search_label) : 
        search_ax (left_tag (untag s \+ s')) (xs_of s) (os_of s) (js_of s).
Proof. by case: s=>/= s v1 v2 v3 [s2 ->]; exists (s2 \+ s'); rewrite joinA. Qed.

Canonical left_recurse s' s := SearchLabel (@left_pf s' s).

Lemma right_pf s' (s : search_label) :
        search_ax (right_tag (s' \+ untag s)) (xs_of s) (os_of s) (js_of s). 
Proof. by case: s=>/= s v1 v2 v3 [s2 ->]; exists (s' \+ s2); rewrite joinCA. Qed.

Canonical right_recurse s' s := SearchLabel (@right_pf s' s).

End GetAuto3. 

Module Exports.
Coercion state_of : search_label >-> tagged_state.
Canonical found_tag.
Canonical found_label.
Canonical left_recurse.
Canonical right_recurse.

Lemma getV gc (s : search_label gc) :
        valid (untag s) -> 
        (self gc (untag s) = xs_of s) * 
        (other gc (untag s) = os_of s) * 
        (joint gc (untag s) = js_of s).
Proof. by case: s=>/= s v1 v2 v3 [s' ->] V; split=>*; rewrite !getPtUn. Qed.

Arguments getV [gc][s] _. 

End Exports.
End GetAuto3.

Export GetAuto3.Exports.


(************************)
(* getters and star_coh *)
(************************)

Section SplitLemmas.
Variables (gc1 gc2 : getcoh). 
Notation coh1 := (coh_of gc1).
Notation coh2 := (coh_of gc2).
Notation l1 := (getL (get_of gc1)).
Notation l2 := (getL (get_of gc2)). 

Lemma star_cohP s : 
        s \In star_coh coh1 coh2 -> 
        [/\ s = l1 ->> [self gc1 s, joint gc1 s, other gc1 s] \+ 
                l2 ->> [self gc2 s, joint gc2 s, other gc2 s],
            validH s,
            l1 ->> [self gc1 s, joint gc1 s, other gc1 s] \In coh1 &
            l2 ->> [self gc2 s, joint gc2 s, other gc2 s] \In coh2].
Proof.
move=>C; case: (C)=>Vh [s1][s2][E] C1 C2; rewrite {s}E in C Vh *. 
move: (cohV C) (cohE C1) (cohE C2) => V E1 E2.
by rewrite E1 E2 in C C1 C2 V Vh *; rewrite !getV. 
Qed.

Lemma star_cohE s : 
        s \In star_coh coh1 coh2 ->
        s = l1 ->> [self gc1 s, joint gc1 s, other gc1 s] \+ 
            l2 ->> [self gc2 s, joint gc2 s, other gc2 s].
Proof. by case/star_cohP. Qed.

Lemma star_cohVSO s : 
        s \In star_coh coh1 coh2 -> 
        valid (self gc1 s \+ other gc1 s) /\ valid (self gc2 s \+ other gc2 s).
Proof.
case/star_cohP=>E /validHV V /cohVSO C1 /cohVSO C2.
by rewrite !getE in C1 C2.
Qed.

Lemma star_zigSOE s p : 
        s \< p \In star_coh coh1 coh2 -> 
        p = l1 \-> getso gc1 p \+ l2 \-> getso gc2 p.
Proof.
case=>Vh C; case: C Vh=>s1 [s2][E] C1 C2 /validHV V.
case: (valid_zig V)=>_ Vp; case/and3P: V=>[Vl _ _].
case/(zig_plusf Vl): E Vp C1 C2=>s1' [s2'][p1][p2][_ ->->->] Vp C1 C2. 
by rewrite (zigSOE C1) (zigSOE C2) in Vp *; rewrite !fgetV. 
Qed.

Lemma star_zagSOE s p : 
        s \> p \In star_coh coh1 coh2 -> 
        p = l1 \-> getso gc1 p \+ l2 \-> getso gc2 p.
Proof. by move/coh_zigzag/star_zigSOE. Qed.

End SplitLemmas.




(**************************************)
(* The getter lemmas specific to Priv *)
(* should be moved to privstate.v     *)
(**************************************)

Definition pvg pv := Getter pv heap_pcm_Encoded tt.

Lemma pv_cohP pv s : 
        s \In priv_coh pv <->
        [/\ joint (pvg pv) s = tt, 
            s = pv ->> [self (pvg pv) s, joint (pvg pv) s, other (pvg pv) s] &
            valid (self (pvg pv) s \+ other (pvg pv) s)].
Proof.
split; first by case=>V [?][?] E; rewrite E /validH priv_flatsE !getE in V *.
case=>J E V; split; first by rewrite E /validH J priv_flatsE.
by exists (self (pvg pv) s), (other (pvg pv) s); rewrite -J. 
Qed.

Lemma pv_cohE pv s : 
       s \In priv_coh pv -> 
       s = pv ->> [self (pvg pv) s, joint (pvg pv) s, other (pvg pv) s].
Proof. by case/pv_cohP. Qed.

Definition privGCMixin pv := GCMixin (@pv_cohE pv).
Canonical priv_getcoh pv := 
  Eval hnf in GC (pvg pv) (priv_coh pv) (privGCMixin pv).

Lemma pv_cohJ pv s : joint (pvg pv) s = tt.
Proof. by case: (joint _ _). Qed.

(* I need a new priv_write, which uses getters *)

Program Definition priv_write pv A x (e : A) : 
  {h}, STsep [Priv pv] 
    (fun i => exists B (v : B), self (pvg pv) i = x :-> v \+ h,
     fun (_ : unit) m => self (pvg pv) m = x :-> e \+ h) := 
  Do (lemmas.priv_write pv x e). 
Next Obligation.
apply: gh=>i h /= [B][v] P C; apply: (gh_ex h). 
apply: val_do; last by case=>m [_][hO'] -> _; rewrite getE.
by exists (other (pvg pv) i), B, v; rewrite -P -(pv_cohJ pv i) -(cohE C).
Qed.

Program Definition priv_read pv A x : 
  {v h}, STsep [Priv pv] 
    (fun i => self (pvg pv) i = x :-> v \+ h,
     fun r m => r = v /\ self (pvg pv) m = x :-> v \+ h) := 
  Do (lemmas.priv_read pv A x). 
Next Obligation.
apply: gh=>i [v h] /= P C; apply: (gh_ex v); apply: (gh_ex h).
apply: val_do=>[_|]; last by move=>m r[->{m}][hO']->{r}; rewrite getE. 
by exists (other (pvg pv) i); rewrite -P -(pv_cohJ pv i) -(cohE C).
Qed.

(**********************************************)
(* Iterating getters over sequences of labels *)
(**********************************************)

(* We need arity polymorphism wrt. sequences of getters *)
(* in order to state the generic par lemma *)

Definition get_pcm' (gc : getcoh) (gcs : seq getcoh) U :=
  if gcs is [::] then getU gc else prod_pcm_Encoded (getU gc) U.

Fixpoint get_pcm (gcs : seq getcoh) : encoded_pcm := 
  if gcs is gc :: gcs' then get_pcm' gc gcs' (get_pcm gcs') 
  else unit_pcm_Encoded.

Definition get_tp' (gc : getcoh) (gcs : seq getcoh) T := 
  if gcs is [::] then getT gc else prod_tp_Encoded (getT gc) T.

Fixpoint get_tp (gcs : seq getcoh) : encoded_type := 
  if gcs is gc :: gcs' then get_tp' gc gcs' (get_tp gcs')
  else unit_tp_Encoded.

Definition get_coh' (gc : getcoh) (gcs : seq getcoh) C := 
  if gcs is [::] then coh_of gc else star_coh (coh_of gc) C.

Fixpoint get_coh (gcs : seq getcoh) : cohpred := 
  if gcs is gc :: gcs' then get_coh' gc gcs' (get_coh gcs')
  else emp_coh.

Fixpoint get_fV (gcs : seq getcoh) : seq nat := 
  if gcs is gc :: gcs' then getL gc :: get_fV gcs' else [::].

(* the programs for iterating getters *)
(* each program comes with a "primed" helper which tests *)
(* the end case -- to remove the need for trailing units *)

Definition sgetso' gc gcs (g : feap dpcm -> get_pcm gcs) :=
  match gcs as gcs return _ -> feap dpcm -> get_pcm (gc :: gcs) with
    [::] => fun g => getso' gc
  | _ => fun g f => (getso' gc f, g f)
  end g.

Fixpoint sgetso gcs := 
  match gcs as gcs return feap dpcm -> get_pcm gcs with 
    gc :: gcs' => sgetso' gc (sgetso gcs')
  | [::] => fun f => tt
  end. 

Definition sgetj' gc gcs (g : feap dtp -> get_tp gcs) :=
  match gcs as gcs return _ -> feap dtp -> get_tp (gc :: gcs) with
    [::] => fun g => getj' gc
  | _ => fun g f => (getj' gc f, g f)
  end g.

Fixpoint sgetj gcs := 
  match gcs as gcs return feap dtp -> get_tp gcs with
    gc :: gcs' => sgetj' gc (sgetj gcs')
  | [::] => fun f => tt
  end.

Definition sself' gc gcs (g : state -> get_pcm gcs) :=
  match gcs as gcs return _ -> state -> get_pcm (gc :: gcs) with 
    [::] => fun g s => self gc s
  | _ => fun g s => (self gc s, g s)
  end g.

Fixpoint sself gcs := 
  match gcs as gcs return state -> get_pcm gcs with 
    gc :: gcs' => sself' gc (sself gcs')
  | [::] => fun s => tt
  end.   

Definition sjoint' gc gcs (g : state -> get_tp gcs) :=
  match gcs as gcs return _ -> state -> get_tp (gc :: gcs) with 
    [::] => fun g s => joint gc s
  | _ => fun g s => (joint gc s, g s)
   end g.
  
Fixpoint sjoint gcs := 
  match gcs as gcs return state -> get_tp gcs with 
    gc :: gcs' => sjoint' gc (sjoint gcs')
  | [::] => fun s => tt
  end.     

Definition sother' gc gcs (g : state -> get_pcm gcs) := 
  match gcs as gcs return _ -> state -> get_pcm (gc :: gcs) with 
    [::] => fun g s => other gc s
  | _ => fun g s => (other gc s, g s)
  end g.

Fixpoint sother gcs := 
  match gcs as gcs return state -> get_pcm gcs with 
    gc :: gcs' => sother' gc (sother gcs')
  | [::] => fun s => tt
  end.   

Definition sstate' gc gcs g := 
  match gcs as gcs 
    return (get_pcm gcs -> get_tp gcs -> 
            get_pcm gcs -> state) ->
           (get_pcm (gc :: gcs) -> get_tp (gc :: gcs) -> 
            get_pcm (gc :: gcs) -> state) with
    [::] => fun g s j o => getL gc ->> [s, j, o]
  | _ => fun g s j o => getL gc ->> [s.1, j.1, o.1] \+ g s.2 j.2 o.2
  end g.

Fixpoint sstate gcs :=
  match gcs as gcs 
    return get_pcm gcs -> get_tp gcs -> get_pcm gcs -> state with 
    gc :: gcs' => sstate' (sstate gcs')
  | [::] => fun s o j => Unit
  end.

Notation "[ 'st' s , j , o ']'" := (sstate s j o)
  (at level 0, format "[ 'st'  s ,  j ,  o ]").


(* the lemmas about iterated getters *)

Lemma sselfE' gcs gc s :  
        sself' gc (sself gcs) s = sgetso' gc (sgetso gcs) (loc s).
Proof. by elim: gcs gc=>//= gc' gcs IH gc; rewrite IH. Qed.

Lemma sselfE gcs s : sself gcs s = sgetso gcs (loc s). 
Proof. by case: gcs=>//= gc gcs; rewrite sselfE'. Qed.

Lemma sotherE' gcs gc s : 
        sother' gc (sother gcs) s = sgetso' gc (sgetso gcs) (env s). 
Proof. by elim: gcs gc=>//= gc' gcs IH gc; rewrite IH. Qed.

Lemma sotherE gcs s : sother gcs s = sgetso gcs (env s).
Proof. by case: gcs=>//= gc gcs; rewrite sotherE'. Qed.

Lemma sjointE' gcs gc s : 
        sjoint' gc (sjoint gcs) s = sgetj' gc (sgetj gcs) (shr s).
Proof. by elim: gcs gc=>//= gc' gcs IH gc; rewrite IH. Qed.

Lemma sjointE gcs s : sjoint gcs s = sgetj gcs (shr s). 
Proof. by case: gcs=>// gc gcs; apply: sjointE'. Qed.

Lemma sgetsoZ' gc gcs f1 f2 : 
        valid (f1 \* f2) ->
        sgetso' gc (sgetso gcs) (f1 \* f2) =
        sgetso' gc (sgetso gcs) f1 \+ sgetso' gc (sgetso gcs) f2.
Proof. by elim: gcs gc=>[|gc' gcs IH] gc /= V; rewrite getsoZ // IH. Qed.

Lemma sgetsoZ gcs f1 f2 : 
        valid (f1 \* f2) -> 
        sgetso gcs (f1 \* f2) = sgetso gcs f1 \+ sgetso gcs f2.
Proof. by case: gcs=>// gc gcs; apply: sgetsoZ'. Qed.

Lemma sgetsoPtUn' (gc : getcoh) gcs (U : encoded_pcm) l (x : U) f : 
        valid (l \-> x \+ f) ->
        l \notin getL gc :: get_fV gcs ->
        sgetso' gc (sgetso gcs) (l \-> x \+ f) = sgetso' gc (sgetso gcs) f.
Proof.
elim: gcs gc=>[|gc' gcs IH] gc /= V.
- by rewrite inE => N; rewrite getsoPtUnR.
by rewrite inE negb_or; case/andP=>H1 H2; rewrite IH // getsoPtUnR.
Qed.

Lemma sgetsoPtUn gcs (U : encoded_pcm) (l : nat) (x : U) (f : feap dpcm) :
        valid (l \-> x \+ f) -> 
        l \notin get_fV gcs ->
        sgetso gcs (l \-> x \+ f) = sgetso gcs f.
Proof. by case: gcs=>// gc gcs; apply: sgetsoPtUn'. Qed.

Lemma sgcfvE' gc gcs s : 
        s \In get_coh' gc gcs (get_coh gcs) ->
        fV (get_coh' gc gcs (get_coh gcs)) =i getL gc :: get_fV gcs.
Proof.
elim: gcs gc s=>[|gc' gcs IH] gc s C z; first by rewrite (gcfvE C z). 
case: C (C)=>_ [s1][s2][->] C1 C2 C.
by rewrite inE /= (gcfvE C1 z) (IH _ _ C2) !inE.
Qed.

Lemma sgcfvE gcs s : s \In get_coh gcs -> fV (get_coh gcs) =i get_fV gcs.
Proof. by case: gcs=>// gc gcs; apply: sgcfvE'. Qed.

Lemma sselfUn' gc gcs c s1 s2 : 
        s1 \In c ->
        s1 \+ s2 \In star_coh c (get_coh' gc gcs (get_coh gcs)) ->
        (sself' gc (sself gcs) (s1 \+ s2) = sself' gc (sself gcs) s2) *
        (sjoint' gc (sjoint gcs) (s1 \+ s2) = sjoint' gc (sjoint gcs) s2) *
        (sother' gc (sother gcs) (s1 \+ s2) = sother' gc (sother gcs) s2).
Proof.
elim: gcs gc c s1 s2=>[|gc' gcs IH] gc c x s' /= C Cggs;
move/(star_coh_prec Cggs): (C)=>Cgg. 
- rewrite (joinC x) (joinC (loc x)) (joinC (shr x)) (joinC (env x)) in Cggs *.
  by rewrite !(getCL _ Cgg Cggs) (gcfvE Cgg) inE eq_refl. 
case: Cgg (Cgg) Cggs=>_ [xgc][xgcs][->{s'}] Cg Cgs Cgg Cggs. 
rewrite !(IH _ _ _ _ Cg Cgg).
rewrite !joinA starA in Cggs *.
move/(star_coh_prec Cggs): (Cgs)=>Cs.
rewrite !(IH _ _ _ _ Cs Cggs).
rewrite -!joinA (joinC x) (joinC (loc x)) 
         (joinC (shr x)) (joinC (env x)) in Cggs *.
by rewrite !(getCL _ Cgg Cggs) inE /= (gcfvE Cg) inE eq_refl.
Qed.

Lemma sselfUn s1 s2 c gcs : 
        s1 \In c ->
        s1 \+ s2 \In star_coh c (get_coh gcs) ->
        (sself gcs (s1 \+ s2) = sself gcs s2) *
        (sjoint gcs (s1 \+ s2) = sjoint gcs s2) *
        (sother gcs (s1 \+ s2) = sother gcs s2).
Proof. by case: gcs=>// gc gcs; apply: sselfUn'. Qed.

(* this lemma annoying; should be stated in terms of getso, *)
(* but i don't see how to do that *)

Lemma sselfZf' gc gcs x (x1 x2 : get_pcm' gc gcs (get_pcm gcs)) :
        x \In get_coh' gc gcs (get_coh gcs) ->
        sself' gc (sself gcs) x = x1 \+ x2 ->
        exists f1 f2 : feap dpcm,
          [/\ loc x = f1 \* f2, sgetso' gc (sgetso gcs) f1 = x1 & 
              sgetso' gc (sgetso gcs) f2 = x2].
Proof.
elim: gcs gc x x1 x2=>[|gc' gcs IH] /= gc x x1 x2.
- move=>C E1; exists (getL gc \-> x1), (getL gc \-> x2).  
  by rewrite zipPt !getE {1}(cohE C) fsE /= E1. 
case: x1=>a1 x1; case: x2=>a2 x2 C. 
case: C (C)=>_ [s1][s2][->{x}] C1 C2 C. 
rewrite (getCL _ C1 C) (gcfvE C1) inE eq_refl (sselfUn' C1 C). 
case=>E1; have {E1} E1 : loc s1 = getL gc \-> (a1 \+ a2).
- by rewrite {1}(cohE C1) fsE /= -E1.  
case/(IH _ _ _ _ C2)=>f1 [f2][E2 X1 X2].
have [V1 V2] : valid (getL gc \-> a1 \+ f1) /\ valid (getL gc \-> a2 \+ f2).
- by case/and3P: (cohV C); rewrite /= E1 E2 -zipUnPt validZE; case/and3P.
suff N : getL gc \notin get_fV (gc' :: gcs).
- exists (getL gc \-> a1 \+ f1), (getL gc \-> a2 \+ f2).
  by rewrite zipUnPt -E1 -E2 !fgetV // !sgetsoPtUn' // X1 X2.
rewrite -(sgcfvE' C2); case: (cohfV C2)=><- _ _.
rewrite E2 fdomZL -E2; case/and3P: (cohV C)=> /validR -> _ _.
by rewrite fvalidPtUn in V1; case/andP: V1.
Qed.

Lemma sselfZf gcs x x1 x2 : 
        x \In get_coh gcs ->
        sself gcs x = x1 \+ x2 -> 
        exists f1 f2, 
          [/\ loc x = f1 \* f2, sgetso gcs f1 = x1 & sgetso gcs f2 = x2].
Proof. 
case: gcs x1 x2 => [|gc gcs] x1 x2; last by apply: sselfZf'.
by case: x1 x2=>[[]] ->; exists Unit, [zunit env Unit]; rewrite zunitR. 
Qed.

Lemma scohE' gc gcs x : 
        x \In get_coh' gc gcs (get_coh gcs) ->
        x = sstate' (@sstate gcs)
                    (sself' gc (sself gcs) x)
                    (sjoint' gc (sjoint gcs) x)
                    (sother' gc (sother gcs) x).
Proof.
elim: gcs gc x =>[|gc' gcs IH] /= gc x C.
- by rewrite -(cohE C).
case: C (C)=>_ [s1][s2][->{x}] C1 C2 C.
by rewrite !(sselfUn' C1 C) -{1}(IH _ _ C2) !(getC C1 C) -(cohE C1).
Qed.

Lemma scohE gcs x : 
        x \In get_coh gcs -> 
        x = [st sself gcs x, sjoint gcs x, sother gcs x].
Proof. by case: gcs=>// gc gcs; apply: scohE'. Qed.


(*************************************************)
(* Generic parallel composition and frame lemmas *)
(* expressed through getters                     *)
(*************************************************)

(* two helper lemmas *)
Lemma zagstE gcs lp1 lp2 sh ep :
       [lp1 \* lp2, sh, ep] \In get_coh gcs ->
       [lp1, sh, ep] \> lp2 = 
       [st sgetso gcs lp1, sgetj gcs sh, sgetso gcs lp2 \+ sgetso gcs ep].
Proof.
move=>C.
have C1 : [lp1, sh, ep] \> lp2 \In get_coh gcs.
- by apply/coh_zigzag; rewrite zigE zipC.
rewrite (scohE C1) zagE sselfE sotherE sjointE /= sgetsoZ; last first.
- by move/allvalidV: (cohSO C); rewrite zipAC -zipA; case/validZP. 
by rewrite joinC; congr sstate; apply: sjointE.
Qed.

Lemma zigstE gcs lp1 lp2 sh ep :
       [lp1 \* lp2, sh, ep] \In get_coh gcs ->
       [lp1, sh, ep] \< lp2 = 
       [st sgetso gcs lp1 \+ sgetso gcs lp2, sgetj gcs sh, sgetso gcs ep].
Proof.
move=>C; rewrite zigE zipC {1}(scohE C) sselfE sotherE sjointE /= sgetsoZ //.
by case/cohV/and3P: C. 
Qed.

(* par is not algebraic, and doesn't allow nesting of verifies for e1 and e2 *)
(* a nesting allows that if e1 diverges, then e2 is not checked for safety *)
(* so we have to stick with writing explicitly r1 and r2 *)

Lemma par_do (gcs : seq getcoh) (W : mod) (A1 A2 : Type) (x : state) 
             (x1 x2 : get_pcm gcs) (e1 : ST W A1) (e2 : ST W A2)             
             (r1 : cont A1) (r2 : cont A2) (r : cont (A1 * A2)) : 
        Mod.coh W = get_coh gcs ->
        sself gcs x = x1 \+ x2 ->
        verify [st x1, sjoint gcs x, x2 \+ sother gcs x] e1 r1 ->
        verify [st x2, sjoint gcs x, x1 \+ sother gcs x] e2 r2 ->
        (forall v x x1 x2, 
           x \In get_coh gcs ->
           [st x1, sjoint gcs x, x2 \+ sother gcs x] \In get_coh gcs ->
           [st x2, sjoint gcs x, x1 \+ sother gcs x] \In get_coh gcs ->
           sself gcs x = x1 \+ x2 ->
           r1 v.1 [st x1, sjoint gcs x, x2 \+ sother gcs x] ->
           r2 v.2 [st x2, sjoint gcs x, x1 \+ sother gcs x] ->
           r v x) ->
        verify x (par e1 e2) r.
Proof.
move=>E S H1 H2 H; apply: vrf_coh; rewrite E => C.
case: (sselfZf C S)=>lp1 [lp2][/= Ex S1 S2]. 
rewrite -[x]st_eta Ex -S1 -S2 sotherE sjointE /= in C H1 H2 *.
rewrite -!zagstE ?(zipC lp2) // in H1 H2.
apply: {Ex S S1 S2 C lp1 lp2 x x1 x2} par_do H1 H2 _.
rewrite E => v lp1 lp2 sh ep C H1 H2.
set x := [lp1 \* lp2, sh, ep].
have S : sself gcs x = sgetso gcs lp1 \+ sgetso gcs lp2.
- by rewrite /x sselfE sgetsoZ //; case/and3P: (cohV C). 
apply: H (C) _ _ S _ _; rewrite sotherE /= sjointE /= -zagstE ?(zipC lp2) //;
by rewrite -coh_zigzag zigE ?(zipC lp2). 
Qed.

Lemma frame_do (gcs : seq getcoh) (W : mod) (A : Type) (x : state) 
               (x1 x2 : get_pcm gcs) (e : ST W A) (r : cont A) : 
        Mod.coh W = get_coh gcs ->
        sself gcs x = x1 \+ x2 ->
        verify [st x1, sjoint gcs x, x2 \+ sother gcs x] e 
          (fun v x => forall o, sother gcs x = x2 \+ o -> 
             r v [st sself gcs x \+ x2, sjoint gcs x, o]) ->
        verify x e r.
Proof.
move=>E S H; apply: vrf_coh=>C; rewrite E in C.
case: (sselfZf C S)=>/= lp1 [lp2][/= Ex S1 S2].
rewrite -[x]st_eta Ex in C H *.
rewrite -S1 -S2 sotherE sjointE -zagstE // in H.  
rewrite zipC -zigE; apply: frame_do3.
apply: vrf_coh_post; apply: {x S Ex C lp1 S1 S2} vrf_mono H.
move=>v x H C [lp1 sh ep] X; rewrite {x}X in C H.
rewrite zagE sselfE sjointE sotherE /= in H.
move/coh_zigzag: C => C; rewrite zigE zipC E in C.
rewrite (zigstE C); apply: H; rewrite zipC sgetsoZ //.
by move/cohSO/allvalidV: C; rewrite -zipA; case/validZP. 
Qed.

(*
A different form of framing, where a chunk of state doesn't have a
getter associated with it. Only supports framing wrt one label, and
ignores the rest eventually. I think this is an obsolete lemma, as we
should have every structure have a getter. But, until we migrate all
the structure to getters, some proofs use this lemma, so I keep it.
*)

Lemma frame' (gc : getcoh) W A c (x y : state) (x1 x2 : getU gc) 
             (e : ST W A) (r : cont A) : 
        x \In coh_of gc -> 
        self gc x = x1 \+ x2 ->
        Mod.coh W = star_coh (coh_of gc) c ->
        verify (getL gc ->> [x1, joint gc x, x2 \+ other gc x] \+ y) e 
          (fun v s' => forall x' y' o, 
             s' = x' \+ y' -> x' \In coh_of gc -> 
             y' \In c -> other gc x' = x2 \+ o ->
             r v (getL gc ->> [self gc x' \+ x2, joint gc x', o] \+ y')) ->
        verify (x \+ y) e r.
Proof.
set l := getL gc.
move=>Cx Efr Ecoh H; apply: vrf_coh=>Cxy. 
have {Cxy} Cy : y \In c by rewrite Ecoh in Cxy; apply/(star_coh_prec Cxy).
have Ze : forall z, (z \< [zunit loc z] = z) * (z \> [zunit env z] = z).
- by move=>z; rewrite -[z]st_eta zigE zagE zunitL zunitR.
have Zunit : forall s, s \> [zunit env y] \In c -> 
  ([zunit loc s] = [zunit loc y]) * ([zunit env s] = [zunit env y]). 
- move=>s Cs.
  rewrite -(sig_loc Cy) (sig_loc Cs) -(sig_env Cy) (sig_env Cs).
  rewrite -[s]st_eta /= zunitZL.   
  by case/cohV/and3P: Cs=>_ _; rewrite -{1}[s]st_eta /= => ->.
have Zle : forall s, s \In c -> [zunit loc s] = [zunit env s].
- by move=>s Cs; rewrite -(sig_loc Cs) -(sig_env Cs).
have -> : x \+ y = (l ->> [x1, joint gc x, other gc x] \+ y) \< 
                   (l \-> x2 \+ [zunit loc y]) 
  by rewrite zigUnPt fsE zigE zipPt (joinC x2) -Efr -fsE Ze -cohE.
apply: frame_do3; rewrite zagUnPt Zle // Ze fsE zagE zipPt -fsE -(joinC x2).
move=>C' t' T'; apply: {H C' t' T'} always.aft_imp (H C' t' T'). 
move=>v x' Cxy H j Ex'; rewrite {x'}Ex' {}Ecoh in Cxy H *.
have Vxy : valid (env (j \> (l \-> x2 \+ [zunit env y]))).
- by case/cohV/and3P: Cxy.
have V : valid (l \-> x2 \+ [zunit env y]) by case/cohV/valid_zag: Cxy.  
case: Cxy (Cxy)=>_ [sx][sy][E]; case/(zag_plusf Vxy): E (E) H.
move=>x' [y'][px][py][->] E ->->{sx sy j Vxy} -> H Cx' Cy' Cxy. 
have Epx : px = l \-> x2.
- suff: fdom (l \-> x2) =i fdom px by move/(fdomE V E).
  move=>k; case/and3P: (cohV Cx') (cohfV Cx')=>_ _ /fdom_zagp <- [_ _ ->].
  by rewrite fdomPt (gcfvE Cx') !inE eq_sym.
move: Epx V E H Cxy Cx' Cy' => -> V /(funhKl V) <-{px py V} H Cxy Cx' Cy'.
rewrite zig_plus; last first.
- by case/coh_zigzag/cohV/and3P: Cy'. 
- by case/coh_zigzag/cohV/and3P: Cx'. 
have : other gc (x' \> l \-> x2) = x2 \+ other gc x'.
- by rewrite getzV ?getsoPt //; move/coh_zigzag/cohSO: Cx'.
move/(H _ _ _ (erefl _) Cx' Cy'); move/coh_zigzag: Cx'=>Cx'.
rewrite !getE -{1}(Zunit _ Cy') -[y']st_eta /= zagE zigE /=.
rewrite zunitR -Zle // -(Zunit _ Cy') zunitL st_eta.
by rewrite (cohE Cx') !getE (getzV (cohSO Cx')) getsoPt.
Qed.



