Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq path.
Require Import pred prelude idynamic finmap pcm coding unionmap heap feap.
Require Import state zigzag world inject rely atomic autosplit.
Require Import privstate footprint model lemmas.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.
(*
(* we introduce a bit stronger definition for precision *)
(* than usual; in that we demand that the ghosts are *)
(* determined by the heaps *)
(* strictly speaking, this is not necessary for much of the *)
(* development *)
(* however, it leads to a nice rule for hiding/retraction, whereby *)
(* we don't need the user to provide the ghosts by hand as *)
(* they can be uniquely determined from the heap *)
(* so, if this is not found good at some point, we can just *)
(* revert to use forall g, precise (inv_of g) *)
(* instead of aux_precise inv_of *)

Definition aux_precise (U : pcm) (P : U -> Pred heap) := 
  forall g1 g2 h1 h2 t1 t2, 
    valid (h1 \+ t1) ->
    h1 \+ t1 = h2 \+ t2 ->
    h1 \In P g1 -> h2 \In P g2 -> g1 = g2 /\ h1 = h2.
*)

Structure lock_info := LockInfo {
  label_of: nat;
  pcm_of : encoded_pcm; 
  inv_of : pcm_of -> Pred heap; 
  precise_of : forall g, precise (inv_of g)
}.

Section LockSpec.
Variables (hp : nat) (lkI : lock_info).
Notation U := (@pcm_of lkI).          (* pcm *) 
Notation I := (@inv_of lkI).          (* resource invariant *)
Notation pf_prec := (@precise_of lkI).
Variables (f : U -> U) (pf : local f).

(* Locking procedure *)
Definition lock_spec is_lock locked : spec unit := 
  (fun i => exists (hl he : heap) (gl ge: U) i', 
     i = hp ->> [hl, tt, he] \+ i' /\ is_lock gl ge i', 
   fun i y m => forall hl he (gl ge: U) i', 
     i = hp ->> [hl, tt, he] \+ i' -> is_lock gl ge i' ->
     exists h ge' he' m',
       [/\ m = hp ->> [hl \+ h, tt, he'] \+ m', 
           locked gl ge' m',
           valid (gl \+ ge') & I (gl \+ ge') h]).

(* Unlocking procedure *)
Definition unlock_spec is_lock locked : spec unit := 
  (fun i => exists (gL gE: U)(hl he h: heap) i', 
     (fun i => 
       [/\ valid (f gL \+ gE), I (f gL \+ gE) h, 
           locked gL gE i' & i = hp ->> [hl \+ h, tt, he] \+ i']) i,
   fun i y m => forall gL gE hl he h i', 
     (fun i => 
       [/\ valid (f gL \+ gE), I (f gL \+ gE) h, 
           locked gL gE i' & i = hp ->> [hl \+ h, tt, he] \+ i']) i ->
     exists he' (gE': U) m',
       [/\ m = hp ->> [hl, tt, he'] \+ m' &
           is_lock (f gL) gE' m']).

End LockSpec.

(*
Section LockStruct'.
Variables (hp : nat).
Structure lock_struct' (lock: lockLike) := Lockish {
  lock_mod : mod;
  is_lock : pcm_of lock -> pcm_of lock -> state -> Prop;
  is_lock_coh : forall gl ge s, is_lock gl ge s -> s \In Mod.coh lock_mod;
  coh_is_lock : forall s, s \In Mod.coh lock_mod -> (exists gl ge, is_lock gl ge s);
  _ : forall gl ge s, is_lock gl ge s <-> is_lock ge gl (transp s);
  locked : pcm_of lock -> pcm_of lock -> state -> Prop;
  locked_islock : forall s gl ge, locked gl ge s -> is_lock gl ge s; 
  is_lock_stable : forall gl ge s1 s2, is_lock gl ge s1 -> menv_steps lock_mod s1 s2 
                                    -> exists ge', is_lock gl ge' s2;
  locked_stable : forall gl ge s1 s2, locked gl ge s1 -> menv_steps lock_mod s1 s2 
                                   -> locked gl ge s2;
  _ : forall s gl ge, locked gl ge s -> locked ge gl (transp s)-> False;
  lock_proc : @STbin (entangle (Priv hp) lock_mod) unit (lock_spec hp is_lock locked);
  unlock_proc : forall (f : pcm_of lock -> pcm_of lock)(pf: local f), 
    @STbin (entangle (Priv hp) lock_mod) unit (unlock_spec hp f is_lock locked)
}.

End LockStruct'.
*)


Module LockStruct.
Section LockStruct.
Variable lkI : lock_info. 

Structure mixin_of (W : mod) := Mixin {
  is_lock_op : pcm_of lkI -> pcm_of lkI -> state -> Prop;
  locked_op : pcm_of lkI -> pcm_of lkI -> state -> Prop;

  lock_op : forall hp, 
    STbin (entangle (Priv hp) W) (lock_spec hp is_lock_op locked_op);
  unlock_op : forall hp f, local f -> 
    STbin (entangle (Priv hp) W) (unlock_spec hp f is_lock_op locked_op);

  _ : forall gl ge s, is_lock_op gl ge s -> s \In Mod.coh W;
  _ : forall s, s \In Mod.coh W -> exists gl ge, is_lock_op gl ge s;
  _ : forall gl ge s, is_lock_op gl ge s <-> is_lock_op ge gl (transp s);

  _ : forall s gl ge, locked_op gl ge s -> is_lock_op gl ge s; 
  _ : forall gl ge s1 s2, is_lock_op gl ge s1 -> menv_steps W s1 s2 -> 
             exists ge', is_lock_op gl ge' s2;
  _ : forall gl ge s1 s2, locked_op gl ge s1 -> menv_steps W s1 s2 -> 
             locked_op gl ge s2;
  _ : forall s gl ge, locked_op gl ge s -> locked_op ge gl (transp s)-> False}.

Section ClassDef.
Notation class_of := mixin_of (only parsing). 

Structure type : Type := Pack {sort : mod; _ : class_of sort; _ : mod}.
Local Coercion sort : type >-> mod. 

Variables (W : mod) (cT : type).
Definition class := let: Pack _ c _ as cT' := cT return class_of cT' in c.

Definition pack c := @Pack W c W.
Definition clone := fun c & W = sort cT & phant_id (pack c) cT => pack c. 

Definition is_lock := @is_lock_op cT class : 
  pcm_of lkI -> pcm_of lkI -> state -> Prop.
Definition locked := @locked_op cT class : 
  pcm_of lkI -> pcm_of lkI -> state -> Prop.
Definition lock hp := lock_op class hp : 
    STbin (entangle (Priv hp) cT) (lock_spec hp is_lock locked).
Definition unlock hp f pf := @unlock_op cT class hp f pf : 
  STbin (entangle (Priv hp) cT) (unlock_spec hp f is_lock locked).

End ClassDef.
End LockStruct.

Module Exports.
Coercion sort : type >-> mod.
Notation lockmod := type.
Notation LMMixin := Mixin.
Notation LM lkI W m := (@pack lkI W m).

Notation "[ 'lmMixin' 'of' lkI , W ]" := (class _ _ : mixin_of lkI W)
  (at level 0, format "[ 'lmMixin'  'of'  lkI ,  W ]") : form_scope.
Notation "[ 'lockmod' 'of' lkI , W 'for' C ]" := (@clone lkI W C _ (erefl _) id)
  (at level 0, format "[ 'lockmod'  'of'  lkI ,  W  'for'  C ]") : form_scope.
Notation "[ 'lockmod' 'of' lkI , W ]" := (@clone lkI W _ _ (erefl _) id)
  (at level 0, format "[ 'lockmod'  'of'  lkI ,  W ]") : form_scope.

Notation is_lock := is_lock.
Notation locked := locked.
Notation lock := lock.
Notation unlock := unlock.

Section Repack.
Variables (lkI : lock_info) (W : lockmod lkI). 

Lemma is_lock_coh gL gE s : is_lock W gL gE s -> s \In Mod.coh W. 
Proof. 
by case: W gL gE=>U [il ld lk ul H1 H2 H3 H4 H5 H6 H7 m] gL gE; apply: H1.
Qed.

Lemma coh_is_lock s : s \In Mod.coh W -> exists gL gE, is_lock W gL gE s.
Proof. by case: W=>U [il ld lk ul H1 H2 H3 H4 H5 H6 H7 m]; apply: H2. Qed.

Lemma is_lock_transp gL gE s : is_lock W gL gE s <-> is_lock W gE gL (transp s).
Proof. 
by case: W gL gE=>U [il ld lk ul H1 H2 H3 H4 H5 H6 H7 m] gL gE; apply: H3.
Qed.

Lemma locked_islock gL gE s : locked W gL gE s -> is_lock W gL gE s.
Proof.
by case: W gL gE=>U [il ld lk ul H1 H2 H3 H4 H5 H6 H7 m] gL gE; apply: H4.
Qed.

Lemma is_lock_stable gL gE s1 s2 : 
        is_lock W gL gE s1 -> menv_steps W s1 s2 -> 
        exists gE', is_lock W gL gE' s2.
Proof.
by case: W gL gE=>U [il ld lk ul H1 H2 H3 H4 H5 H6 H7 m] gL gE; apply: H5.
Qed.

Lemma locked_stable gL gE s1 s2 : 
        locked W gL gE s1 -> menv_steps W s1 s2 -> locked W gL gE s2.
Proof.
by case: W gL gE=>U [il ld lk ul H1 H2 H3 H4 H5 H6 H7 m] gL gE; apply: H6.
Qed.

Lemma locked_transp gL gE s : 
        locked W gL gE s -> locked W gE gL (transp s)-> False.
Proof.
by case: W gL gE=>U [il ld lk ul H1 H2 H3 H4 H5 H6 H7 m] gL gE; apply: H7.
Qed.

End Repack.

End Exports.

End LockStruct.

Export LockStruct.Exports.


