Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq.
Require Import pred prelude idynamic domain pcm unionmap heap coding feap.
Require Import state zigzag world rely inject atomic schedule refine hide. 
Require Import process always model. 
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(****************************************************)
(* Notation to move from binary posts to unary ones *)
(****************************************************)

Definition logvar {B A} (s : A -> spec B) : spec B := 
  (fun i => exists x : A, let 'pair p _ := s x in p i, 
   fun y i m => forall x : A, let 'pair _ q := s x in q y i m).

Definition binarify {A} (p : pre) (q : A -> pre) : spec A := 
  (p, fun i y m => p i -> q y m).

(* how to remove the need for brackets around W? *)
Notation "'STsep' [ W ] ( p , q ) " := 
  (STbin W (binarify p q)) (at level 0, 
   format "'[hv ' STsep  [ W ]  ( '[' p , '/' q ']' ) ']'").  

(* this notation can be made for spec, but I like it better for STsep *)
Notation "{ x .. y }, 'STsep' [ W ] ( p , q )" :=
  (STbin W (logvar (fun x => .. (logvar (fun y => binarify p q)) .. )))
   (at level 0, x binder, y binder, right associativity,
    format "'[hv ' { x .. y }, '/ ' STsep  [ W ]  ( '[' p , '/' q ']' ) ']'").

(*********************)
(* Structural lemmas *)
(*********************)

Lemma vrf_coh W A (e : ST W A) i r : 
        (i \In Mod.coh W -> verify i e r) -> verify i e r.
Proof. by move=>H C; apply: H. Qed.

Lemma vrf_coh_post W A (e : ST W A) i (r : cont A) : 
        verify i e (fun v m => m \In Mod.coh W -> r v m) ->
        verify i e r.
Proof. by move=>H C t /(H C); apply: aft_imp=>v s Cs; apply. Qed.

(* verify is a weakest-pre computation *)
(* could have written i \in verify e r to emphasize *)
Lemma vrf_mono W A (e : ST W A) i (r1 r2 : cont A) : 
        r1 <== r2 -> verify i e r1 -> verify i e r2. 
Proof. by move=>T H1 C t; move/(H1 C); apply: aft_imp=>v m _; apply: T. Qed.

(* stability of preconditions *)
Lemma vrf_pre W A (e : ST W A) i i' (k : cont A) : 
        verify i e k -> menv_steps W i i' -> verify i' e k. 
Proof.
move=>H M Ci' t H'; case: (menvs_coh M)=>Ci _.
by apply: aft_imp (alw_envs (H Ci t H') M).
Qed.

(* stability of postconditions *)
Lemma vrf_post W A (e : ST W A) i (k : cont A) : 
        verify i e k ->
        verify i e (fun x m => forall m', menv_steps W m m' -> k x m').
Proof.
move=>H Ci t H'; move: (alw_envsq (H Ci t H')).
apply: alw_imp=>s p Cs H2 s3 M v E; apply: H2 E _ M.
Qed.

(* Hoare lemma for sequential composition *)
Lemma hoare_seq W A B (e1 : ST W A) (e2 : A -> ST W B) i q k : 
        verify i e1 q -> 
        (forall y j, q y j -> verify j (e2 y) k) ->
        verify i (bind e1 e2) k.
Proof.
move=>H1 H2 Ci t [->|[t'][H3 H4]]. 
- by apply: alw_unfin=>//; move/alw_coh: (H1 Unfinished (prog_unfin e1)). 
by apply: aft_bnd H3 _; move/(H1 Ci): H4; apply: aft_imp=>y j Cj H; apply: H2.
Qed.

(* stepping form of sequential composition lemma, infers q automatically *)
(* or: weakest pre of sequential composition is a sequence of weakest pre's *)
Lemma step W A B (e1 : ST W A) (e2 : A -> ST W B) i k : 
        verify i e1 (fun y m => verify m (e2 y) k) ->
        verify i (bind e1 e2) k.
Proof. by move=>H; apply: (hoare_seq (q:=(fun y m => verify m (e2 y) k))). Qed.


(********************************)
(* Do lemmas for function calls *)
(********************************)

Section DoLemmas.
Variables (W : mod) (A B : Type).

(* the basic do lemma for binary post *)
Lemma val_do' i (e : ST W A) (k : cont A) : 
        (i \In Mod.coh W -> pre_of e i) -> 
        (forall x m, post_of e i x m -> m \In Mod.coh W -> k x m) ->
        verify i e k.
Proof.
case: e=>s [e] /= H H1 H2 Ci t H3.
apply: aft_imp (H i t (H1 Ci) Ci H3). 
by move=>v m Cm H4; apply: H2.
Qed.

(* the do lemma for unary posts produces simpler second obligation *)
Lemma val_do (p : Pred state) (q : A -> Pred state) i {e} (k : cont A) : 
        (i \In Mod.coh W -> p i) -> 
        (forall x m, q x m -> m \In Mod.coh W -> k x m) ->
        verify i (@with_spec W A (binarify p q) e) k.
Proof. 
move=>H1 H2; apply: vrf_coh=>C; apply: val_do'=>//. 
by move=>x m /(_ (H1 C)); apply: H2.
Qed.

End DoLemmas.


(********************************************)
(* Lemmas for instantiating ghost variables *)
(********************************************)

Section GhostInstLemmas.
Variables (W : mod) (A C : Type) (s : C -> spec A) (e : STbin W (logvar s)).

(* helper lemma, to express the instantiation *)
Lemma gh_conseq t : conseq e (s t).
Proof.
case E: (s t)=>[a b] h /= H; apply: val_do'=>[|x m]. 
- by exists t; rewrite E. 
by move/(_ t); rewrite E. 
Qed.

(* the main lemma for instantiating a ghost *)
(* should gh_ex and val_do be combined into one? *)
Lemma gh_ex t i (k : cont A) : 
        verify i (do' (@gh_conseq t)) k ->
        verify i (@with_spec W A (logvar s) e) k.
Proof. by []. Qed.

End GhostInstLemmas.

Implicit Arguments gh_ex [W A C s e i k].


(******************************************)
(* Lemmas for pulling out ghost variables *)
(******************************************)

Section BaseGhostPullLemmas.
Variables (W : mod) (A B C : Type) (e : ST W A).

Lemma ghE (s : B -> C -> spec A) : 
        conseq e (logvar (fun x => logvar (s x))) <->
        conseq e (logvar (fun xy => s xy.1 xy.2)).
Proof.
split.
- move=>/= H1 i [[x y]] H2.
  have: exists x1 y1, let '(p, _) := s x1 y1 in p i by exists x, y. 
  by move/H1; apply: vrf_mono=>y1 m1 T1 [x2 y2]; apply: (T1 x2 y2). 
move=>/= H1 i [x][y] H2.  
have: exists x, let '(p, _) := s x.1 x.2 in p i by exists (x, y). 
by move/H1; apply: vrf_mono=>y1 m1 T1 x2 y2; apply: (T1 (x2, y2)).
Qed.

Lemma ghC (p : B -> pre) (q : B -> A -> pre) :
        (forall i x, p x i -> i \In Mod.coh W -> verify i e (q x)) ->
        conseq e (logvar (fun x => binarify (p x) (q x))).
Proof.
move=>H i /= [x Hp] Ci t Ht. 
have S : alwsafe i t by apply: alw_imp (H i x Hp Ci Ci t Ht). 
by apply/aftA=>// y; apply/aftI=>// /H; apply.
Qed.

End BaseGhostPullLemmas.


(* Canonical structure program to automate applications of ghE and ghC.   *)
(*                                                                        *)
(* The form pivots on a spec, and computes as output the following.       *)
(* - rT is a product of types of all encounted ghost vars.                *)
(* - p and q are pre and post parametrized by rT that should be output by *)
(*   the main lemma.                                                      *)
(*                                                                        *)
(* In the future, rT should be a list of types, rather than a product,    *)
(* but that leads to arity polimorphism, and dependent programming, which *)
(* I want to avoid for now.                                               *)

Module GhostAuto. 
Section GhostAuto. 

Structure tagged_spec A := spec_tag {untag : spec A}.
Definition dstep_tag := spec_tag.
Definition step_tag := dstep_tag.
Canonical base_tag A (s : spec A) := step_tag s.

Local Coercion untag : tagged_spec >-> spec.

Definition axiom A rT  p q (pivot : tagged_spec A) := 
  untag pivot = logvar (fun x : rT => binarify (p x) (q x)).

Structure form A rT (p : rT -> pre) (q : rT -> cont A) := Form {
   pivot : tagged_spec A;
   _ : axiom p q pivot}. 

Local Coercion pivot : form >-> tagged_spec.

(* the main lemma that automates the applications of ghE and gh *)

Lemma gh W A (e : ST W A) rT p q (f : @form A rT p q) : 
        (forall i x, p x i -> i \In Mod.coh W -> verify i e (q x)) ->
        conseq e f. 
Proof. by case: f=>p' ->; apply: ghC. Qed.

(* base case; check if we reached binarify *)

Lemma base_pf A rT (p : rT -> pre) (q : rT -> cont A) :
        axiom p q (base_tag (logvar (fun x => binarify (p x) (q x)))).
Proof. by []. Qed.

Canonical base_form A rT p q := Form (@base_pf A rT p q).
                 
(* inductive case; merge adjacent logvars and continue *)

Lemma step_pf A B rT p q (f : forall x : A, @form B rT (p x) (q x)) : 
        axiom (fun xy => p xy.1 xy.2) (fun xy => q xy.1 xy.2)
              (step_tag (logvar (fun x => f x))).
Proof. 
congr (_, _).
- apply: fext=>i; apply: pext; split. 
  - by case=>x; case: (f x)=>[_ ->] /= [y H]; exists (x, y). 
  by case; case=>x y; exists x; case: (f x)=>[_ ->]; exists y.  
apply: fext=>y; apply: fext=>i; apply: fext=>m; apply: pext; split.
- by move=>H [x z] /= H1; case: (f x) (H x)=>[_ ->]; apply.  
by move=>H x; case: (f x)=>[_ ->] z; apply: (H (x, z)).
Qed.

Canonical step_form A B rT p q f := Form (@step_pf A B rT p q f). 

Lemma dstep_pf A B rT p q (f : forall x : A, @form B (rT x) (p x) (q x)) : 
        axiom (fun xy => let: existT xy1 xy2 := xy in p xy1 xy2) 
              (fun xy => let: existT xy1 xy2 := xy in q xy1 xy2)
              (dstep_tag (logvar (fun x => f x))).
Proof. 
congr (_, _).
- apply: fext=>i; apply: pext; split. 
  - by case=>x; case: (f x)=>[_ ->] /= [y H]; exists (existT _ _ y). 
  by case; case=>x y; exists x; case: (f x)=>[_ ->]; exists y.  
apply: fext=>y; apply: fext=>i; apply: fext=>m; apply: pext; split.
- by move=>H [x z] /= H1; case: (f x) (H x)=>[_ ->]; apply.  
by move=>H x; case: (f x)=>[_ ->] z; apply: (H (existT _ _ z)).
Qed.

Canonical dstep_form A B rT p q f := Form (@dstep_pf A B rT p q f). 

End GhostAuto.

Module Exports.
Coercion untag : tagged_spec >-> spec.
Coercion pivot : form >-> tagged_spec.
Canonical base_tag.
Canonical base_form.
Canonical step_form.
Canonical dstep_form.
Notation gh := gh.
End Exports.

End GhostAuto. 

Export GhostAuto.Exports.


